/**
 * KIELER - Kiel Integrated Environment for Layout Eclipse RichClient
 * 
 * http://www.informatik.uni-kiel.de/rtsys/kieler/
 * 
 * Copyright 2012 by
 * + Christian-Albrechts-University of Kiel
 *   + Department of Computer Science
 *     + Real-Time and Embedded Systems Group
 * 
 * This code is provided under the terms of the Eclipse Public License (EPL).
 * See the file epl-v10.html for the license text.
 */
package de.cau.cs.kieler.core.krendering.extensions;

import com.google.common.collect.Iterators;
import de.cau.cs.kieler.core.kgraph.KGraphData;
import de.cau.cs.kieler.core.kgraph.KLabel;
import de.cau.cs.kieler.core.kgraph.KLabeledGraphElement;
import de.cau.cs.kieler.core.kgraph.KNode;
import de.cau.cs.kieler.core.kgraph.KPort;
import de.cau.cs.kieler.core.krendering.KColor;
import de.cau.cs.kieler.core.krendering.KRendering;
import de.cau.cs.kieler.core.krendering.KRenderingFactory;
import de.cau.cs.kieler.core.krendering.KText;
import de.cau.cs.kieler.core.krendering.ViewSynthesisShared;
import de.cau.cs.kieler.core.krendering.extensions.KNodeExtensions;
import de.cau.cs.kieler.core.krendering.extensions.KRenderingExtensions;
import de.cau.cs.kieler.core.properties.IProperty;
import de.cau.cs.kieler.kiml.klayoutdata.KShapeLayout;
import de.cau.cs.kieler.kiml.options.EdgeLabelPlacement;
import de.cau.cs.kieler.kiml.options.LayoutOptions;
import de.cau.cs.kieler.kiml.options.NodeLabelPlacement;
import de.cau.cs.kieler.kiml.options.PortLabelPlacement;
import de.cau.cs.kieler.kiml.util.KimlUtil;
import java.util.ArrayList;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import javax.inject.Inject;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.xtext.xbase.lib.CollectionLiterals;
import org.eclipse.xtext.xbase.lib.Extension;
import org.eclipse.xtext.xbase.lib.IteratorExtensions;
import org.eclipse.xtext.xbase.lib.ObjectExtensions;
import org.eclipse.xtext.xbase.lib.Procedures.Procedure1;

/**
 * Provides some helpful extension methods for simplifying the composition of KGraph/KRendering-based view models.<br>
 * <br>
 * In order to employ them beyond KLighD diagram syntheses you best declare a field of type
 * {@link KNodeExtensions} in your class and annotate it with {@link javax.inject.Inject Inject}.<br>
 * <br>
 * Make sure to bind the {@link ViewSynthesisShared} annotation in the employed
 * {@link com.google.inject.Injector Injector} to a {@link com.google.inject.Scope}, e.g. by calling
 * {@code Guice.createInjector(KRenderingExtensionsPlugin.createSingletonScopeBindingModule());} or
 * {@code Guice.createInjector(KRenderingExtensionsPlugin.createNoScopeBindingModule());}.<br>
 * <br>
 * By means of that {@link com.google.inject.Injector Injector} you may get a new instance of your class,
 * or you may inject the above mentioned attribute within instances of your class, e.g. by calling
 * {@code injector.injectMembers(this)} in the constructor of your class.
 * 
 * @author chsch
 * @author ssm
 * 
 * @containsExtensions
 */
@ViewSynthesisShared
@SuppressWarnings("all")
public class KLabelExtensions {
  @Extension
  private KRenderingFactory _kRenderingFactory = KRenderingFactory.eINSTANCE;
  
  @Inject
  @Extension
  private KNodeExtensions _kNodeExtensions;
  
  @Inject
  @Extension
  private KRenderingExtensions _kRenderingExtensions;
  
  /**
   * A convenient getter preserving the element image relation by a create extension.
   */
  private KLabel getLabel(final List<?> o, final KLabeledGraphElement labeledElement) {
    final ArrayList<?> _cacheKey = CollectionLiterals.newArrayList(o, labeledElement);
    final KLabel _result;
    synchronized (_createCache_getLabel) {
      if (_createCache_getLabel.containsKey(_cacheKey)) {
        return _createCache_getLabel.get(_cacheKey);
      }
      KLabel _createInitializedLabel = KimlUtil.createInitializedLabel(labeledElement);
      _result = _createInitializedLabel;
      _createCache_getLabel.put(_cacheKey, _result);
    }
    _init_getLabel(_result, o, labeledElement);
    return _result;
  }
  
  private final HashMap<ArrayList<?>, KLabel> _createCache_getLabel = CollectionLiterals.newHashMap();
  
  private void _init_getLabel(final KLabel node, final List<?> o, final KLabeledGraphElement labeledElement) {
  }
  
  /**
   * A convenient getter preserving the element image relation by a create extension.
   */
  public KLabel getLabel(final Object o, final KLabeledGraphElement labeledElement) {
    ArrayList<Object> _newArrayList = CollectionLiterals.<Object>newArrayList(o);
    return this.getLabel(_newArrayList, labeledElement);
  }
  
  /**
   * Convenience creator of KLabels that are not related to semantic elements.
   * It is just syntactic sugar.
   */
  public KLabel createLabel(final KLabeledGraphElement labeledElement) {
    return KimlUtil.createInitializedLabel(labeledElement);
  }
  
  /**
   * An alias of {@link #getLabel(Object, KLabeledGraphElement)} allowing to express in business
   *  that the KLabel will be created at this place. It is just syntactic sugar.
   */
  public KLabel createLabel(final Object o, final KLabeledGraphElement labeledElement) {
    ArrayList<Object> _newArrayList = CollectionLiterals.<Object>newArrayList(o);
    return this.getLabel(_newArrayList, labeledElement);
  }
  
  /**
   * A convenient getter associating the label with to elements
   *  preserving the element image relation by a create extension.
   */
  public KLabel getLabel(final Object o1, final Object o2, final KLabeledGraphElement labeledElement) {
    ArrayList<Object> _newArrayList = CollectionLiterals.<Object>newArrayList(o1, o2);
    return this.getLabel(_newArrayList, labeledElement);
  }
  
  /**
   * An alias of {@link #getLabel(Object, Object, KLabeledGraphElement)} allowing to express in business
   *  that the KLabel will be created at this place. It is just syntactic sugar.
   */
  public KLabel createLabel(final Object o1, final Object o2, final KLabeledGraphElement labeledElement) {
    ArrayList<Object> _newArrayList = CollectionLiterals.<Object>newArrayList(o1, o2);
    return this.getLabel(_newArrayList, labeledElement);
  }
  
  public KLabel setLabelSize(final KLabel label, final float width, final float height) {
    final Procedure1<KLabel> _function = new Procedure1<KLabel>() {
      public void apply(final KLabel it) {
        KShapeLayout _data = it.<KShapeLayout>getData(KShapeLayout.class);
        _data.setSize(width, height);
      }
    };
    return ObjectExtensions.<KLabel>operator_doubleArrow(label, _function);
  }
  
  public KLabel setLabelPos(final KLabel label, final float x, final float y) {
    final Procedure1<KLabel> _function = new Procedure1<KLabel>() {
      public void apply(final KLabel it) {
        KShapeLayout _data = it.<KShapeLayout>getData(KShapeLayout.class);
        _data.setPos(x, y);
      }
    };
    return ObjectExtensions.<KLabel>operator_doubleArrow(label, _function);
  }
  
  public <T extends Object> KLabel addLayoutParam(final KLabel label, final IProperty<? super T> property, final T value) {
    final Procedure1<KLabel> _function = new Procedure1<KLabel>() {
      public void apply(final KLabel it) {
        KShapeLayout _data = it.<KShapeLayout>getData(KShapeLayout.class);
        _data.<T>setProperty(property, value);
      }
    };
    return ObjectExtensions.<KLabel>operator_doubleArrow(label, _function);
  }
  
  /**
   * Configures an inside centrally-aligned node label!
   */
  public KLabel configureInsideCenteredNodeLabel(final KLabel label, final String labelText, final int fontSize, final String fontName) {
    final Procedure1<KLabel> _function = new Procedure1<KLabel>() {
      public void apply(final KLabel it) {
        KLabelExtensions.this.basicConfigureLabel(it, labelText, fontSize, fontName);
        final KLabeledGraphElement node = it.getParent();
        boolean _matched = false;
        if (!_matched) {
          if (node instanceof KNode) {
            _matched=true;
            EnumSet<NodeLabelPlacement> _insideCenter = NodeLabelPlacement.insideCenter();
            KLabelExtensions.this._kNodeExtensions.<EnumSet<NodeLabelPlacement>>addLayoutParam(((KNode)node), LayoutOptions.NODE_LABEL_PLACEMENT, _insideCenter);
          }
        }
      }
    };
    return ObjectExtensions.<KLabel>operator_doubleArrow(label, _function);
  }
  
  /**
   * Adds an inside bottom right-aligned node label!
   */
  public KLabel addInsideCenteredNodeLabel(final KNode node, final String labelText, final int fontSize, final String fontName) {
    KLabel _createLabel = this.createLabel(node);
    return this.configureInsideCenteredNodeLabel(_createLabel, labelText, fontSize, fontName);
  }
  
  /**
   * Configures an inside bottom centrally-aligned node label!
   */
  public KLabel configureInsideBottomCenteredNodeLabel(final KLabel label, final String labelText, final int fontSize, final String fontName) {
    final Procedure1<KLabel> _function = new Procedure1<KLabel>() {
      public void apply(final KLabel it) {
        KLabelExtensions.this.basicConfigureLabel(it, labelText, fontSize, fontName);
        final KLabeledGraphElement node = it.getParent();
        boolean _matched = false;
        if (!_matched) {
          if (node instanceof KNode) {
            _matched=true;
            EnumSet<NodeLabelPlacement> _insideBottomCenter = NodeLabelPlacement.insideBottomCenter();
            KLabelExtensions.this._kNodeExtensions.<EnumSet<NodeLabelPlacement>>addLayoutParam(((KNode)node), LayoutOptions.NODE_LABEL_PLACEMENT, _insideBottomCenter);
          }
        }
      }
    };
    return ObjectExtensions.<KLabel>operator_doubleArrow(label, _function);
  }
  
  /**
   * Adds an inside bottom centrally-aligned node label!
   */
  public KLabel addInsideBottomCenteredNodeLabel(final KNode node, final String labelText, final int fontSize, final String fontName) {
    KLabel _createLabel = this.createLabel(node);
    return this.configureInsideBottomCenteredNodeLabel(_createLabel, labelText, fontSize, fontName);
  }
  
  /**
   * Configures an inside bottom left-aligned node label!
   */
  public KLabel configureInsideBottomLeftNodeLabel(final KLabel label, final String labelText, final int fontSize, final String fontName) {
    final Procedure1<KLabel> _function = new Procedure1<KLabel>() {
      public void apply(final KLabel it) {
        KLabelExtensions.this.basicConfigureLabel(it, labelText, fontSize, fontName);
        final KLabeledGraphElement node = it.getParent();
        boolean _matched = false;
        if (!_matched) {
          if (node instanceof KNode) {
            _matched=true;
            EnumSet<NodeLabelPlacement> _insideBottomLeft = NodeLabelPlacement.insideBottomLeft();
            KLabelExtensions.this._kNodeExtensions.<EnumSet<NodeLabelPlacement>>addLayoutParam(((KNode)node), LayoutOptions.NODE_LABEL_PLACEMENT, _insideBottomLeft);
          }
        }
      }
    };
    return ObjectExtensions.<KLabel>operator_doubleArrow(label, _function);
  }
  
  /**
   * Adds an inside bottom left-aligned node label!
   */
  public KLabel addInsideBottomLeftNodeLabel(final KNode node, final String labelText, final int fontSize, final String fontName) {
    KLabel _createLabel = this.createLabel(node);
    return this.configureInsideBottomLeftNodeLabel(_createLabel, labelText, fontSize, fontName);
  }
  
  /**
   * Configures an inside bottom right-aligned node label!
   */
  public KLabel configureInsideBottomRightNodeLabel(final KLabel label, final String labelText, final int fontSize, final String fontName) {
    final Procedure1<KLabel> _function = new Procedure1<KLabel>() {
      public void apply(final KLabel it) {
        KLabelExtensions.this.basicConfigureLabel(it, labelText, fontSize, fontName);
        final KLabeledGraphElement node = it.getParent();
        boolean _matched = false;
        if (!_matched) {
          if (node instanceof KNode) {
            _matched=true;
            EnumSet<NodeLabelPlacement> _insideBottomRight = NodeLabelPlacement.insideBottomRight();
            KLabelExtensions.this._kNodeExtensions.<EnumSet<NodeLabelPlacement>>addLayoutParam(((KNode)node), LayoutOptions.NODE_LABEL_PLACEMENT, _insideBottomRight);
          }
        }
      }
    };
    return ObjectExtensions.<KLabel>operator_doubleArrow(label, _function);
  }
  
  /**
   * Adds an inside bottom right-aligned node label!
   */
  public KLabel addInsideBottomRightNodeLabel(final KNode node, final String labelText, final int fontSize, final String fontName) {
    KLabel _createLabel = this.createLabel(node);
    return this.configureInsideBottomRightNodeLabel(_createLabel, labelText, fontSize, fontName);
  }
  
  /**
   * Configures an inside top centrally-aligned node label!
   */
  public KLabel configureInsideTopCenteredNodeLabel(final KLabel label, final String labelText, final int fontSize, final String fontName) {
    final Procedure1<KLabel> _function = new Procedure1<KLabel>() {
      public void apply(final KLabel it) {
        KLabelExtensions.this.basicConfigureLabel(it, labelText, fontSize, fontName);
        final KLabeledGraphElement node = it.getParent();
        boolean _matched = false;
        if (!_matched) {
          if (node instanceof KNode) {
            _matched=true;
            EnumSet<NodeLabelPlacement> _insideTopCenter = NodeLabelPlacement.insideTopCenter();
            KLabelExtensions.this._kNodeExtensions.<EnumSet<NodeLabelPlacement>>addLayoutParam(((KNode)node), LayoutOptions.NODE_LABEL_PLACEMENT, _insideTopCenter);
          }
        }
      }
    };
    return ObjectExtensions.<KLabel>operator_doubleArrow(label, _function);
  }
  
  /**
   * Adds an inside top centrally-aligned node label!
   */
  public KLabel addInsideTopCenteredNodeLabel(final KNode node, final String labelText, final int fontSize, final String fontName) {
    KLabel _createLabel = this.createLabel(node);
    return this.configureInsideTopCenteredNodeLabel(_createLabel, labelText, fontSize, fontName);
  }
  
  /**
   * Configures an inside top left-aligned node label!
   */
  public KLabel configureInsideTopLeftNodeLabel(final KLabel label, final String labelText, final int fontSize, final String fontName) {
    final Procedure1<KLabel> _function = new Procedure1<KLabel>() {
      public void apply(final KLabel it) {
        KLabelExtensions.this.basicConfigureLabel(it, labelText, fontSize, fontName);
        final KLabeledGraphElement node = it.getParent();
        boolean _matched = false;
        if (!_matched) {
          if (node instanceof KNode) {
            _matched=true;
            EnumSet<NodeLabelPlacement> _insideTopLeft = NodeLabelPlacement.insideTopLeft();
            KLabelExtensions.this._kNodeExtensions.<EnumSet<NodeLabelPlacement>>addLayoutParam(((KNode)node), LayoutOptions.NODE_LABEL_PLACEMENT, _insideTopLeft);
          }
        }
      }
    };
    return ObjectExtensions.<KLabel>operator_doubleArrow(label, _function);
  }
  
  /**
   * Adds an inside top left-aligned node label!
   */
  public KLabel addInsideTopLeftNodeLabel(final KNode node, final String labelText, final int fontSize, final String fontName) {
    KLabel _createLabel = this.createLabel(node);
    return this.configureInsideTopLeftNodeLabel(_createLabel, labelText, fontSize, fontName);
  }
  
  /**
   * Configures an inside top right-aligned node label!
   */
  public KLabel configureInsideTopRightNodeLabel(final KLabel label, final String labelText, final int fontSize, final String fontName) {
    final Procedure1<KLabel> _function = new Procedure1<KLabel>() {
      public void apply(final KLabel it) {
        KLabelExtensions.this.basicConfigureLabel(it, labelText, fontSize, fontName);
        final KLabeledGraphElement node = it.getParent();
        boolean _matched = false;
        if (!_matched) {
          if (node instanceof KNode) {
            _matched=true;
            EnumSet<NodeLabelPlacement> _insideTopRight = NodeLabelPlacement.insideTopRight();
            KLabelExtensions.this._kNodeExtensions.<EnumSet<NodeLabelPlacement>>addLayoutParam(((KNode)node), LayoutOptions.NODE_LABEL_PLACEMENT, _insideTopRight);
          }
        }
      }
    };
    return ObjectExtensions.<KLabel>operator_doubleArrow(label, _function);
  }
  
  /**
   * Adds an inside top right-aligned node label!
   */
  public KLabel addInsideTopRightNodeLabel(final KNode node, final String labelText, final int fontSize, final String fontName) {
    KLabel _createLabel = this.createLabel(node);
    return this.configureInsideTopRightNodeLabel(_createLabel, labelText, fontSize, fontName);
  }
  
  /**
   * Configures an outside bottom centrally-aligned node label!
   */
  public KLabel configureOutsideBottomCenteredNodeLabel(final KLabel label, final String labelText, final int fontSize, final String fontName) {
    final Procedure1<KLabel> _function = new Procedure1<KLabel>() {
      public void apply(final KLabel it) {
        KLabelExtensions.this.basicConfigureLabel(it, labelText, fontSize, fontName);
        final KLabeledGraphElement node = it.getParent();
        boolean _matched = false;
        if (!_matched) {
          if (node instanceof KNode) {
            _matched=true;
            EnumSet<NodeLabelPlacement> _outsideBottomCenter = NodeLabelPlacement.outsideBottomCenter();
            KLabelExtensions.this._kNodeExtensions.<EnumSet<NodeLabelPlacement>>addLayoutParam(((KNode)node), LayoutOptions.NODE_LABEL_PLACEMENT, _outsideBottomCenter);
          }
        }
      }
    };
    return ObjectExtensions.<KLabel>operator_doubleArrow(label, _function);
  }
  
  /**
   * Adds an outside bottom centrally-aligned node label!
   */
  public KLabel addOutsideBottomCenteredNodeLabel(final KNode node, final String labelText, final int fontSize, final String fontName) {
    KLabel _createLabel = this.createLabel(node);
    return this.configureOutsideBottomCenteredNodeLabel(_createLabel, labelText, fontSize, fontName);
  }
  
  /**
   * Configures an outside bottom left-aligned node label!
   */
  public KLabel configureOutsideBottomLeftNodeLabel(final KLabel label, final String labelText, final int fontSize, final String fontName) {
    final Procedure1<KLabel> _function = new Procedure1<KLabel>() {
      public void apply(final KLabel it) {
        KLabelExtensions.this.basicConfigureLabel(it, labelText, fontSize, fontName);
        final KLabeledGraphElement node = it.getParent();
        boolean _matched = false;
        if (!_matched) {
          if (node instanceof KNode) {
            _matched=true;
            EnumSet<NodeLabelPlacement> _outsideBottomLeft = NodeLabelPlacement.outsideBottomLeft();
            KLabelExtensions.this._kNodeExtensions.<EnumSet<NodeLabelPlacement>>addLayoutParam(((KNode)node), LayoutOptions.NODE_LABEL_PLACEMENT, _outsideBottomLeft);
          }
        }
      }
    };
    return ObjectExtensions.<KLabel>operator_doubleArrow(label, _function);
  }
  
  /**
   * Adds an outside bottom left-aligned node label!
   */
  public KLabel addOutsideBottomLeftNodeLabel(final KNode node, final String labelText, final int fontSize, final String fontName) {
    KLabel _createLabel = this.createLabel(node);
    return this.configureOutsideBottomLeftNodeLabel(_createLabel, labelText, fontSize, fontName);
  }
  
  /**
   * Configures an outside bottom right-aligned node label!
   */
  public KLabel configureOutsideBottomRightNodeLabel(final KLabel label, final String labelText, final int fontSize, final String fontName) {
    final Procedure1<KLabel> _function = new Procedure1<KLabel>() {
      public void apply(final KLabel it) {
        KLabelExtensions.this.basicConfigureLabel(it, labelText, fontSize, fontName);
        final KLabeledGraphElement node = it.getParent();
        boolean _matched = false;
        if (!_matched) {
          if (node instanceof KNode) {
            _matched=true;
            EnumSet<NodeLabelPlacement> _outsideBottomRight = NodeLabelPlacement.outsideBottomRight();
            KLabelExtensions.this._kNodeExtensions.<EnumSet<NodeLabelPlacement>>addLayoutParam(((KNode)node), LayoutOptions.NODE_LABEL_PLACEMENT, _outsideBottomRight);
          }
        }
      }
    };
    return ObjectExtensions.<KLabel>operator_doubleArrow(label, _function);
  }
  
  /**
   * Adds an outside bottom right-aligned node label!
   */
  public KLabel addOutsideBottomRightNodeLabel(final KNode node, final String labelText, final int fontSize, final String fontName) {
    KLabel _createLabel = this.createLabel(node);
    return this.configureOutsideBottomRightNodeLabel(_createLabel, labelText, fontSize, fontName);
  }
  
  /**
   * Configures an outside top centrally-aligned node label!
   */
  public KLabel configureOutsideTopCenteredNodeLabel(final KLabel label, final String labelText, final int fontSize, final String fontName) {
    final Procedure1<KLabel> _function = new Procedure1<KLabel>() {
      public void apply(final KLabel it) {
        KLabelExtensions.this.basicConfigureLabel(it, labelText, fontSize, fontName);
        final KLabeledGraphElement node = it.getParent();
        boolean _matched = false;
        if (!_matched) {
          if (node instanceof KNode) {
            _matched=true;
            EnumSet<NodeLabelPlacement> _outsideTopCenter = NodeLabelPlacement.outsideTopCenter();
            KLabelExtensions.this._kNodeExtensions.<EnumSet<NodeLabelPlacement>>addLayoutParam(((KNode)node), LayoutOptions.NODE_LABEL_PLACEMENT, _outsideTopCenter);
          }
        }
      }
    };
    return ObjectExtensions.<KLabel>operator_doubleArrow(label, _function);
  }
  
  /**
   * Adds an outside top centrally-aligned node label!
   */
  public KLabel addOutsideTopCenteredNodeLabel(final KNode node, final String labelText, final int fontSize, final String fontName) {
    KLabel _createLabel = this.createLabel(node);
    return this.configureOutsideTopCenteredNodeLabel(_createLabel, labelText, fontSize, fontName);
  }
  
  /**
   * Configures an outside top left-aligned node label!
   */
  public KLabel configureOutsideTopLeftNodeLabel(final KLabel label, final String labelText, final int fontSize, final String fontName) {
    final Procedure1<KLabel> _function = new Procedure1<KLabel>() {
      public void apply(final KLabel it) {
        KLabelExtensions.this.basicConfigureLabel(it, labelText, fontSize, fontName);
        final KLabeledGraphElement node = it.getParent();
        boolean _matched = false;
        if (!_matched) {
          if (node instanceof KNode) {
            _matched=true;
            EnumSet<NodeLabelPlacement> _outsideTopLeft = NodeLabelPlacement.outsideTopLeft();
            KLabelExtensions.this._kNodeExtensions.<EnumSet<NodeLabelPlacement>>addLayoutParam(((KNode)node), LayoutOptions.NODE_LABEL_PLACEMENT, _outsideTopLeft);
          }
        }
      }
    };
    return ObjectExtensions.<KLabel>operator_doubleArrow(label, _function);
  }
  
  /**
   * Adds an outside top left-aligned node label!
   */
  public KLabel addOutsideTopLeftNodeLabel(final KNode node, final String labelText, final int fontSize, final String fontName) {
    KLabel _createLabel = this.createLabel(node);
    return this.configureOutsideTopLeftNodeLabel(_createLabel, labelText, fontSize, fontName);
  }
  
  /**
   * Configures an outside top right-aligned node label!
   */
  public KLabel configureOutsideTopRightNodeLabel(final KLabel label, final String labelText, final int fontSize, final String fontName) {
    final Procedure1<KLabel> _function = new Procedure1<KLabel>() {
      public void apply(final KLabel it) {
        KLabelExtensions.this.basicConfigureLabel(it, labelText, fontSize, fontName);
        final KLabeledGraphElement node = it.getParent();
        boolean _matched = false;
        if (!_matched) {
          if (node instanceof KNode) {
            _matched=true;
            EnumSet<NodeLabelPlacement> _outsideTopRight = NodeLabelPlacement.outsideTopRight();
            KLabelExtensions.this._kNodeExtensions.<EnumSet<NodeLabelPlacement>>addLayoutParam(((KNode)node), LayoutOptions.NODE_LABEL_PLACEMENT, _outsideTopRight);
          }
        }
      }
    };
    return ObjectExtensions.<KLabel>operator_doubleArrow(label, _function);
  }
  
  /**
   * Adds an outside top right-aligned node label!
   */
  public KLabel addOutsideTopRightNodeLabel(final KNode node, final String labelText, final int fontSize, final String fontName) {
    KLabel _createLabel = this.createLabel(node);
    return this.configureOutsideTopRightNodeLabel(_createLabel, labelText, fontSize, fontName);
  }
  
  /**
   * Configures an inside port label!<br>
   * Note that <code>label</code> must be already contained in a {@link KPort} that in turn is contained
   * in a {@link KNode} in order to let the configuration be performed properly.
   */
  public KLabel configureInsidePortLabel(final KLabel label, final String labelText, final int fontSize, final String fontName) {
    final Procedure1<KLabel> _function = new Procedure1<KLabel>() {
      public void apply(final KLabel it) {
        KLabelExtensions.this.basicConfigureLabel(it, labelText, fontSize, fontName);
        KLabeledGraphElement _parent = it.getParent();
        EObject _eContainer = null;
        if (_parent!=null) {
          _eContainer=_parent.eContainer();
        }
        final EObject node = _eContainer;
        boolean _matched = false;
        if (!_matched) {
          if (node instanceof KNode) {
            _matched=true;
            KLabelExtensions.this._kNodeExtensions.<PortLabelPlacement>addLayoutParam(((KNode)node), LayoutOptions.PORT_LABEL_PLACEMENT, PortLabelPlacement.INSIDE);
          }
        }
      }
    };
    return ObjectExtensions.<KLabel>operator_doubleArrow(label, _function);
  }
  
  /**
   * Adds an inside label to {@link KPort} <code>port</code>!<br>
   * Note that <code>port</code> must be already contained in a {@link KNode} in order to let the
   * configuration be performed properly.
   */
  public KLabel addInsidePortLabel(final KPort port, final String labelText, final int fontSize, final String fontName) {
    KLabel _createLabel = this.createLabel(port);
    return this.configureInsidePortLabel(_createLabel, labelText, fontSize, fontName);
  }
  
  /**
   * Configures a outside port label!<br>
   * Note that <code>label</code> must be already contained in a {@link KPort} that in turn is contained
   * in a {@link KNode} in order to let the configuration be performed properly.
   */
  public KLabel configureOutsidePortLabel(final KLabel label, final String labelText, final int fontSize, final String fontName) {
    final Procedure1<KLabel> _function = new Procedure1<KLabel>() {
      public void apply(final KLabel it) {
        KLabelExtensions.this.basicConfigureLabel(it, labelText, fontSize, fontName);
        KLabeledGraphElement _parent = it.getParent();
        EObject _eContainer = null;
        if (_parent!=null) {
          _eContainer=_parent.eContainer();
        }
        final EObject node = _eContainer;
        boolean _matched = false;
        if (!_matched) {
          if (node instanceof KNode) {
            _matched=true;
            KLabelExtensions.this._kNodeExtensions.<PortLabelPlacement>addLayoutParam(((KNode)node), LayoutOptions.PORT_LABEL_PLACEMENT, PortLabelPlacement.OUTSIDE);
          }
        }
      }
    };
    return ObjectExtensions.<KLabel>operator_doubleArrow(label, _function);
  }
  
  /**
   * Adds an outside label to {@link KPort} <code>port</code>!<br>
   * Note that <code>port</code> must be already contained in a {@link KNode} in order to let the
   * configuration be performed properly.
   */
  public KLabel addOutsidePortLabel(final KPort port, final String labelText, final int fontSize, final String fontName) {
    KLabel _createLabel = this.createLabel(port);
    return this.configureOutsidePortLabel(_createLabel, labelText, fontSize, fontName);
  }
  
  /**
   * Configures a central (main) edge label, e.g. a state transition guard/effect label!
   */
  public KLabel configureCenterEdgeLabel(final KLabel label, final String labelText, final int fontSize, final String fontName) {
    final Procedure1<KLabel> _function = new Procedure1<KLabel>() {
      public void apply(final KLabel it) {
        KLabelExtensions.this.basicConfigureLabel(it, labelText, fontSize, fontName);
        KLabelExtensions.this.<EdgeLabelPlacement>addLayoutParam(it, LayoutOptions.EDGE_LABEL_PLACEMENT, EdgeLabelPlacement.CENTER);
      }
    };
    return ObjectExtensions.<KLabel>operator_doubleArrow(label, _function);
  }
  
  /**
   * Configures a head edge label, e.g. the cardinality of a relation in an class diagram!
   */
  public KLabel configureHeadEdgeLabel(final KLabel label, final String labelText, final int fontSize, final String fontName) {
    final Procedure1<KLabel> _function = new Procedure1<KLabel>() {
      public void apply(final KLabel it) {
        KLabelExtensions.this.basicConfigureLabel(it, labelText, fontSize, fontName);
        KLabelExtensions.this.<EdgeLabelPlacement>addLayoutParam(it, LayoutOptions.EDGE_LABEL_PLACEMENT, EdgeLabelPlacement.HEAD);
      }
    };
    return ObjectExtensions.<KLabel>operator_doubleArrow(label, _function);
  }
  
  /**
   * Configures a tail edge label, e.g. the cardinality of a relation in an class diagram!
   */
  public KLabel configureTailEdgeLabel(final KLabel label, final String labelText, final int fontSize, final String fontName) {
    final Procedure1<KLabel> _function = new Procedure1<KLabel>() {
      public void apply(final KLabel it) {
        KLabelExtensions.this.basicConfigureLabel(it, labelText, fontSize, fontName);
        KLabelExtensions.this.<EdgeLabelPlacement>addLayoutParam(it, LayoutOptions.EDGE_LABEL_PLACEMENT, EdgeLabelPlacement.TAIL);
      }
    };
    return ObjectExtensions.<KLabel>operator_doubleArrow(label, _function);
  }
  
  /**
   * The least common denominator of all the 'configure...Label' methods.<br>
   * Is private as it's to be used internally only!
   */
  public void basicConfigureLabel(final KLabel label, final String labelText, final int fontSize, final String fontName) {
    label.setText(labelText);
    EList<KGraphData> _data = label.getData();
    KText _createKText = this._kRenderingFactory.createKText();
    KText _setFontName = this._kRenderingExtensions.<KText>setFontName(_createKText, fontName);
    KText _setFontSize = this._kRenderingExtensions.<KText>setFontSize(_setFontName, fontSize);
    _data.add(_setFontSize);
    this.<String>addLayoutParam(label, LayoutOptions.FONT_NAME, fontName);
    this.<Integer>addLayoutParam(label, LayoutOptions.FONT_SIZE, Integer.valueOf(fontSize));
  }
  
  /**
   * Reveals the first KText element of a label KRendering, which is assumed to be the label text configuration.
   * This is useful for additionally linking it with the business element represented by the label.
   * 
   * Note: KLabelNodes of the Piccolo2D binding are configured to ignore the KText element while selecting them.
   * Thus, only the KLabel needs to be linked to the source element.
   */
  public KText getFirstText(final KLabel label) {
    KText _elvis = null;
    KText _data = null;
    if (label!=null) {
      _data=label.<KText>getData(KText.class);
    }
    if (_data != null) {
      _elvis = _data;
    } else {
      KRendering _data_1 = null;
      if (label!=null) {
        _data_1=label.<KRendering>getData(KRendering.class);
      }
      TreeIterator<EObject> _eAllContents = null;
      if (_data_1!=null) {
        _eAllContents=_data_1.eAllContents();
      }
      Iterator<KText> _filter = null;
      if (_eAllContents!=null) {
        _filter=Iterators.<KText>filter(_eAllContents, KText.class);
      }
      KText _head = null;
      if (_filter!=null) {
        _head=IteratorExtensions.<KText>head(_filter);
      }
      _elvis = _head;
    }
    return _elvis;
  }
  
  /**
   * Shortcut for setting the label's foreground color directly.
   */
  public KLabel foreground(final KLabel label, final KColor color) {
    final Procedure1<KLabel> _function = new Procedure1<KLabel>() {
      public void apply(final KLabel it) {
        KRendering _kRendering = KLabelExtensions.this._kRenderingExtensions.getKRendering(it);
        KLabelExtensions.this._kRenderingExtensions.setForeground(_kRendering, color);
      }
    };
    return ObjectExtensions.<KLabel>operator_doubleArrow(label, _function);
  }
  
  /**
   * Shortcut for setting the label's background color directly.
   */
  public KLabel background(final KLabel label, final KColor color) {
    final Procedure1<KLabel> _function = new Procedure1<KLabel>() {
      public void apply(final KLabel it) {
        KRendering _kRendering = KLabelExtensions.this._kRenderingExtensions.getKRendering(it);
        KLabelExtensions.this._kRenderingExtensions.setBackground(_kRendering, color);
      }
    };
    return ObjectExtensions.<KLabel>operator_doubleArrow(label, _function);
  }
}
