/**
 * KIELER - Kiel Integrated Environment for Layout Eclipse RichClient
 * 
 * http://www.informatik.uni-kiel.de/rtsys/kieler/
 * 
 * Copyright 2013 by
 * + Christian-Albrechts-University of Kiel
 *   + Department of Computer Science
 *     + Real-Time and Embedded Systems Group
 * 
 * This code is provided under the terms of the Eclipse Public License (EPL).
 * See the file epl-v10.html for the license text.
 */
package de.cau.cs.kieler.kiml.formats.json;

import com.google.common.base.Objects;
import com.google.common.collect.Maps;
import de.cau.cs.kieler.core.kgraph.KEdge;
import de.cau.cs.kieler.core.kgraph.KLabel;
import de.cau.cs.kieler.core.kgraph.KLabeledGraphElement;
import de.cau.cs.kieler.core.kgraph.KNode;
import de.cau.cs.kieler.core.kgraph.KPort;
import de.cau.cs.kieler.kiml.LayoutMetaDataService;
import de.cau.cs.kieler.kiml.formats.IGraphTransformer;
import de.cau.cs.kieler.kiml.formats.TransformationData;
import de.cau.cs.kieler.kiml.formats.TransformationException;
import de.cau.cs.kieler.kiml.formats.json.JsonExtensions;
import de.cau.cs.kieler.kiml.klayoutdata.KEdgeLayout;
import de.cau.cs.kieler.kiml.klayoutdata.KInsets;
import de.cau.cs.kieler.kiml.klayoutdata.KLayoutData;
import de.cau.cs.kieler.kiml.klayoutdata.KLayoutDataFactory;
import de.cau.cs.kieler.kiml.klayoutdata.KPoint;
import de.cau.cs.kieler.kiml.klayoutdata.KShapeLayout;
import de.cau.cs.kieler.kiml.options.EdgeLabelPlacement;
import de.cau.cs.kieler.kiml.options.LayoutOptions;
import de.cau.cs.kieler.kiml.util.KimlUtil;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.xtext.xbase.lib.Exceptions;
import org.eclipse.xtext.xbase.lib.ExclusiveRange;
import org.eclipse.xtext.xbase.lib.Extension;
import org.eclipse.xtext.xbase.lib.InputOutput;
import org.eclipse.xtext.xbase.lib.IterableExtensions;
import org.eclipse.xtext.xbase.lib.IteratorExtensions;
import org.eclipse.xtext.xbase.lib.ObjectExtensions;
import org.eclipse.xtext.xbase.lib.Procedures.Procedure1;
import org.json.JSONArray;
import org.json.JSONObject;

/**
 * Importer for graphs in the json format.
 * 
 * @see {@link JsonHandler} for more information.
 * 
 * @author uru
 */
@SuppressWarnings("all")
public class JsonImporter implements IGraphTransformer<JSONObject, KNode> {
  @Extension
  private JsonExtensions _jsonExtensions = new JsonExtensions();
  
  private final Map<String, KNode> nodeIdMap = Maps.<String, KNode>newHashMap();
  
  private final Map<String, KPort> portIdMap = Maps.<String, KPort>newHashMap();
  
  private final Map<String, KEdge> edgeIdMap = Maps.<String, KEdge>newHashMap();
  
  private final Map<KNode, JSONObject> nodeJsonMap = Maps.<KNode, JSONObject>newHashMap();
  
  private final Map<KPort, JSONObject> portJsonMap = Maps.<KPort, JSONObject>newHashMap();
  
  private final Map<KEdge, JSONObject> edgeJsonMap = Maps.<KEdge, JSONObject>newHashMap();
  
  private final Map<KLabel, JSONObject> labelJsonMap = Maps.<KLabel, JSONObject>newHashMap();
  
  /**
   * Main entry point for the json to kgraph transformation. Runs through all elements
   * of the graph (nodes, ports, edges) and creates correlating KGraph elements.
   */
  public void transform(final TransformationData<JSONObject, KNode> data) {
    final JSONObject graph = data.getSourceGraph();
    this.clearMaps();
    final KNode root = this.transformNode(graph, null);
    this.transformEdges(graph);
    List<KNode> _targetGraphs = data.getTargetGraphs();
    _targetGraphs.add(root);
  }
  
  private void clearMaps() {
    this.nodeIdMap.clear();
    this.portIdMap.clear();
    this.edgeIdMap.clear();
    this.nodeJsonMap.clear();
    this.portJsonMap.clear();
    this.edgeJsonMap.clear();
  }
  
  private JSONArray transformChildNodes(final JSONObject jsonNode, final KNode parent) {
    JSONArray _optJSONArray = jsonNode.optJSONArray("children");
    final Procedure1<JSONArray> _function = new Procedure1<JSONArray>() {
      public void apply(final JSONArray children) {
        boolean _notEquals = (!Objects.equal(children, null));
        if (_notEquals) {
          int _length = children.length();
          ExclusiveRange _doubleDotLessThan = new ExclusiveRange(0, _length, true);
          for (final Integer i : _doubleDotLessThan) {
            JSONObject _optJSONObject = children.optJSONObject((i).intValue());
            if (_optJSONObject!=null) {
              JsonImporter.this.transformNode(_optJSONObject, parent);
            }
          }
        }
      }
    };
    return ObjectExtensions.<JSONArray>operator_doubleArrow(_optJSONArray, _function);
  }
  
  private KNode transformNode(final JSONObject jsonNode, final KNode parent) {
    KNode _createInitializedNode = KimlUtil.createInitializedNode();
    final KNode node = this.register(_createInitializedNode, jsonNode);
    boolean _notEquals = (!Objects.equal(parent, null));
    if (_notEquals) {
      EList<KNode> _children = parent.getChildren();
      _children.add(node);
    }
    KShapeLayout _layout = this._jsonExtensions.layout(node);
    this.transformShapeLayout(jsonNode, _layout);
    this.transformPorts(jsonNode, node);
    this.transformLabels(jsonNode, node);
    KShapeLayout _layout_1 = this._jsonExtensions.layout(node);
    this.transformProperties(jsonNode, _layout_1);
    this.transformChildNodes(jsonNode, node);
    return node;
  }
  
  private void transformEdges(final JSONObject jsonObj) {
    JSONArray _optJSONArray = jsonObj.optJSONArray("edges");
    final Procedure1<JSONArray> _function = new Procedure1<JSONArray>() {
      public void apply(final JSONArray edges) {
        boolean _notEquals = (!Objects.equal(edges, null));
        if (_notEquals) {
          int _length = edges.length();
          ExclusiveRange _doubleDotLessThan = new ExclusiveRange(0, _length, true);
          for (final Integer i : _doubleDotLessThan) {
            JSONObject _optJSONObject = edges.optJSONObject((i).intValue());
            if (_optJSONObject!=null) {
              JsonImporter.this.transformEdge(_optJSONObject);
            }
          }
        }
      }
    };
    ObjectExtensions.<JSONArray>operator_doubleArrow(_optJSONArray, _function);
    JSONArray _optJSONArray_1 = jsonObj.optJSONArray("children");
    final Procedure1<JSONArray> _function_1 = new Procedure1<JSONArray>() {
      public void apply(final JSONArray children) {
        boolean _notEquals = (!Objects.equal(children, null));
        if (_notEquals) {
          int _length = children.length();
          ExclusiveRange _doubleDotLessThan = new ExclusiveRange(0, _length, true);
          for (final Integer i : _doubleDotLessThan) {
            JSONObject _optJSONObject = children.optJSONObject((i).intValue());
            if (_optJSONObject!=null) {
              JsonImporter.this.transformEdges(_optJSONObject);
            }
          }
        }
      }
    };
    ObjectExtensions.<JSONArray>operator_doubleArrow(_optJSONArray_1, _function_1);
  }
  
  private JSONArray transformEdge(final JSONObject jsonObj) {
    JSONArray _xblockexpression = null;
    {
      KEdge _createInitializedEdge = KimlUtil.createInitializedEdge();
      final KEdge edge = this.register(_createInitializedEdge, jsonObj);
      String _optString = jsonObj.optString("source");
      KNode _get = this.nodeIdMap.get(_optString);
      edge.setSource(_get);
      String _optString_1 = jsonObj.optString("sourcePort");
      KPort _get_1 = this.portIdMap.get(_optString_1);
      edge.setSourcePort(_get_1);
      String _optString_2 = jsonObj.optString("target");
      KNode _get_2 = this.nodeIdMap.get(_optString_2);
      edge.setTarget(_get_2);
      String _optString_3 = jsonObj.optString("targetPort");
      KPort _get_3 = this.portIdMap.get(_optString_3);
      edge.setTargetPort(_get_3);
      boolean _or = false;
      KNode _source = edge.getSource();
      boolean _equals = Objects.equal(_source, null);
      if (_equals) {
        _or = true;
      } else {
        KNode _target = edge.getTarget();
        boolean _equals_1 = Objects.equal(_target, null);
        _or = _equals_1;
      }
      if (_or) {
        String _optString_4 = jsonObj.optString("id");
        String _plus = (("Invalid JSON graph format, the target" + " and source of a node may not be null (edge ") + _optString_4);
        String _plus_1 = (_plus + ").");
        throw new TransformationException(_plus_1);
      }
      this.transformLabels(jsonObj, edge);
      KEdgeLayout _layout = this._jsonExtensions.layout(edge);
      _xblockexpression = this.transformEdgeLayout(jsonObj, _layout);
    }
    return _xblockexpression;
  }
  
  private JSONArray transformEdgeLayout(final JSONObject jsonObj, final KEdgeLayout edgeLayout) {
    JSONArray _xblockexpression = null;
    {
      JSONObject _optJSONObject = jsonObj.optJSONObject("sourcePoint");
      final Procedure1<JSONObject> _function = new Procedure1<JSONObject>() {
        public void apply(final JSONObject srcPnt) {
          boolean _notEquals = (!Objects.equal(srcPnt, null));
          if (_notEquals) {
            double _optDouble = srcPnt.optDouble("x");
            final Procedure1<Double> _function = new Procedure1<Double>() {
              public void apply(final Double it) {
                KPoint _sourcePoint = edgeLayout.getSourcePoint();
                float _floatValueValid = JsonImporter.this._jsonExtensions.floatValueValid(it);
                _sourcePoint.setX(_floatValueValid);
              }
            };
            ObjectExtensions.<Double>operator_doubleArrow(Double.valueOf(_optDouble), _function);
            double _optDouble_1 = srcPnt.optDouble("y");
            final Procedure1<Double> _function_1 = new Procedure1<Double>() {
              public void apply(final Double it) {
                KPoint _sourcePoint = edgeLayout.getSourcePoint();
                float _floatValueValid = JsonImporter.this._jsonExtensions.floatValueValid(it);
                _sourcePoint.setY(_floatValueValid);
              }
            };
            ObjectExtensions.<Double>operator_doubleArrow(Double.valueOf(_optDouble_1), _function_1);
          }
        }
      };
      ObjectExtensions.<JSONObject>operator_doubleArrow(_optJSONObject, _function);
      JSONObject _optJSONObject_1 = jsonObj.optJSONObject("targetPoint");
      final Procedure1<JSONObject> _function_1 = new Procedure1<JSONObject>() {
        public void apply(final JSONObject tgtPnt) {
          boolean _notEquals = (!Objects.equal(tgtPnt, null));
          if (_notEquals) {
            double _optDouble = tgtPnt.optDouble("x");
            final Procedure1<Double> _function = new Procedure1<Double>() {
              public void apply(final Double it) {
                KPoint _targetPoint = edgeLayout.getTargetPoint();
                float _floatValueValid = JsonImporter.this._jsonExtensions.floatValueValid(it);
                _targetPoint.setX(_floatValueValid);
              }
            };
            ObjectExtensions.<Double>operator_doubleArrow(Double.valueOf(_optDouble), _function);
            double _optDouble_1 = tgtPnt.optDouble("y");
            final Procedure1<Double> _function_1 = new Procedure1<Double>() {
              public void apply(final Double it) {
                KPoint _targetPoint = edgeLayout.getTargetPoint();
                float _floatValueValid = JsonImporter.this._jsonExtensions.floatValueValid(it);
                _targetPoint.setY(_floatValueValid);
              }
            };
            ObjectExtensions.<Double>operator_doubleArrow(Double.valueOf(_optDouble_1), _function_1);
          }
        }
      };
      ObjectExtensions.<JSONObject>operator_doubleArrow(_optJSONObject_1, _function_1);
      JSONArray _optJSONArray = jsonObj.optJSONArray("bendPoints");
      final Procedure1<JSONArray> _function_2 = new Procedure1<JSONArray>() {
        public void apply(final JSONArray bends) {
          boolean _notEquals = (!Objects.equal(bends, null));
          if (_notEquals) {
            int _length = bends.length();
            ExclusiveRange _doubleDotLessThan = new ExclusiveRange(0, _length, true);
            for (final Integer i : _doubleDotLessThan) {
              {
                final JSONObject jsonBend = bends.optJSONObject((i).intValue());
                final KPoint bend = KLayoutDataFactory.eINSTANCE.createKPoint();
                double _optDouble = jsonBend.optDouble("x");
                final Procedure1<Double> _function = new Procedure1<Double>() {
                  public void apply(final Double it) {
                    float _floatValueValid = JsonImporter.this._jsonExtensions.floatValueValid(it);
                    bend.setX(_floatValueValid);
                  }
                };
                ObjectExtensions.<Double>operator_doubleArrow(Double.valueOf(_optDouble), _function);
                double _optDouble_1 = jsonBend.optDouble("y");
                final Procedure1<Double> _function_1 = new Procedure1<Double>() {
                  public void apply(final Double it) {
                    float _floatValueValid = JsonImporter.this._jsonExtensions.floatValueValid(it);
                    bend.setY(_floatValueValid);
                  }
                };
                ObjectExtensions.<Double>operator_doubleArrow(Double.valueOf(_optDouble_1), _function_1);
                EList<KPoint> _bendPoints = edgeLayout.getBendPoints();
                _bendPoints.add(bend);
              }
            }
          }
        }
      };
      _xblockexpression = ObjectExtensions.<JSONArray>operator_doubleArrow(_optJSONArray, _function_2);
    }
    return _xblockexpression;
  }
  
  private JSONObject transformProperties(final JSONObject jsonObject, final KLayoutData layoutData) {
    JSONObject _xblockexpression = null;
    {
      final LayoutMetaDataService metaService = LayoutMetaDataService.getInstance();
      JSONObject _optJSONObject = jsonObject.optJSONObject("properties");
      final Procedure1<JSONObject> _function = new Procedure1<JSONObject>() {
        public void apply(final JSONObject props) {
          Iterator _keys = null;
          if (props!=null) {
            _keys=props.keys();
          }
          Iterator<Object> _emptyIfNull = JsonImporter.this._jsonExtensions.<Object>emptyIfNull(_keys);
          final Procedure1<Object> _function = new Procedure1<Object>() {
            public void apply(final Object key) {
              final String value = props.optString(((String) key));
              KimlUtil.loadDataElement(metaService, layoutData, ((String) key), value);
            }
          };
          IteratorExtensions.<Object>forEach(_emptyIfNull, _function);
        }
      };
      _xblockexpression = ObjectExtensions.<JSONObject>operator_doubleArrow(_optJSONObject, _function);
    }
    return _xblockexpression;
  }
  
  private JSONArray transformLabels(final JSONObject jsonObj, final KLabeledGraphElement element) {
    JSONArray _optJSONArray = jsonObj.optJSONArray("labels");
    final Procedure1<JSONArray> _function = new Procedure1<JSONArray>() {
      public void apply(final JSONArray labels) {
        boolean _notEquals = (!Objects.equal(labels, null));
        if (_notEquals) {
          int _length = labels.length();
          ExclusiveRange _doubleDotLessThan = new ExclusiveRange(0, _length, true);
          for (final Integer i : _doubleDotLessThan) {
            {
              final JSONObject jsonLabel = labels.optJSONObject((i).intValue());
              boolean _notEquals_1 = (!Objects.equal(jsonLabel, null));
              if (_notEquals_1) {
                final KLabel label = KimlUtil.createInitializedLabel(element);
                JsonImporter.this.labelJsonMap.put(label, jsonLabel);
                String _optString = jsonLabel.optString("text");
                label.setText(_optString);
                KShapeLayout _layout = JsonImporter.this._jsonExtensions.layout(label);
                JsonImporter.this.transformProperties(jsonLabel, _layout);
                KShapeLayout _layout_1 = JsonImporter.this._jsonExtensions.layout(label);
                EdgeLabelPlacement _property = _layout_1.<EdgeLabelPlacement>getProperty(LayoutOptions.EDGE_LABEL_PLACEMENT);
                boolean _equals = Objects.equal(_property, EdgeLabelPlacement.UNDEFINED);
                if (_equals) {
                  KShapeLayout _layout_2 = JsonImporter.this._jsonExtensions.layout(label);
                  _layout_2.<EdgeLabelPlacement>setProperty(LayoutOptions.EDGE_LABEL_PLACEMENT, 
                    EdgeLabelPlacement.CENTER);
                }
              }
            }
          }
        }
      }
    };
    return ObjectExtensions.<JSONArray>operator_doubleArrow(_optJSONArray, _function);
  }
  
  private JSONArray transformPorts(final JSONObject jsonObj, final KNode parent) {
    JSONArray _optJSONArray = jsonObj.optJSONArray("ports");
    final Procedure1<JSONArray> _function = new Procedure1<JSONArray>() {
      public void apply(final JSONArray ports) {
        boolean _notEquals = (!Objects.equal(ports, null));
        if (_notEquals) {
          int _length = ports.length();
          ExclusiveRange _doubleDotLessThan = new ExclusiveRange(0, _length, true);
          for (final Integer i : _doubleDotLessThan) {
            JSONObject _optJSONObject = ports.optJSONObject((i).intValue());
            if (_optJSONObject!=null) {
              JsonImporter.this.transformPort(_optJSONObject, parent);
            }
          }
        }
      }
    };
    return ObjectExtensions.<JSONArray>operator_doubleArrow(_optJSONArray, _function);
  }
  
  private JSONObject transformPort(final JSONObject jsonPort, final KNode parent) {
    JSONObject _xblockexpression = null;
    {
      KPort _createInitializedPort = KimlUtil.createInitializedPort();
      final KPort port = this.register(_createInitializedPort, jsonPort);
      EList<KPort> _ports = parent.getPorts();
      _ports.add(port);
      KShapeLayout _layout = this._jsonExtensions.layout(port);
      this.transformShapeLayout(jsonPort, _layout);
      this.transformLabels(jsonPort, port);
      KShapeLayout _layout_1 = this._jsonExtensions.layout(port);
      _xblockexpression = this.transformProperties(jsonPort, _layout_1);
    }
    return _xblockexpression;
  }
  
  private JSONObject transformShapeLayout(final JSONObject jsonObj, final KShapeLayout shapeLayout) {
    JSONObject _xblockexpression = null;
    {
      double _optDouble = jsonObj.optDouble("x");
      final Procedure1<Double> _function = new Procedure1<Double>() {
        public void apply(final Double it) {
          float _floatValueValid = JsonImporter.this._jsonExtensions.floatValueValid(it);
          shapeLayout.setXpos(_floatValueValid);
        }
      };
      ObjectExtensions.<Double>operator_doubleArrow(Double.valueOf(_optDouble), _function);
      double _optDouble_1 = jsonObj.optDouble("y");
      final Procedure1<Double> _function_1 = new Procedure1<Double>() {
        public void apply(final Double it) {
          float _floatValueValid = JsonImporter.this._jsonExtensions.floatValueValid(it);
          shapeLayout.setYpos(_floatValueValid);
        }
      };
      ObjectExtensions.<Double>operator_doubleArrow(Double.valueOf(_optDouble_1), _function_1);
      double _optDouble_2 = jsonObj.optDouble("width");
      final Procedure1<Double> _function_2 = new Procedure1<Double>() {
        public void apply(final Double it) {
          float _floatValueValid = JsonImporter.this._jsonExtensions.floatValueValid(it);
          shapeLayout.setWidth(_floatValueValid);
        }
      };
      ObjectExtensions.<Double>operator_doubleArrow(Double.valueOf(_optDouble_2), _function_2);
      double _optDouble_3 = jsonObj.optDouble("height");
      final Procedure1<Double> _function_3 = new Procedure1<Double>() {
        public void apply(final Double it) {
          float _floatValueValid = JsonImporter.this._jsonExtensions.floatValueValid(it);
          shapeLayout.setHeight(_floatValueValid);
        }
      };
      ObjectExtensions.<Double>operator_doubleArrow(Double.valueOf(_optDouble_3), _function_3);
      JSONObject _optJSONObject = jsonObj.optJSONObject("padding");
      final Procedure1<JSONObject> _function_4 = new Procedure1<JSONObject>() {
        public void apply(final JSONObject padding) {
          boolean _notEquals = (!Objects.equal(padding, null));
          if (_notEquals) {
            final KInsets insets = shapeLayout.getInsets();
            double _optDouble = padding.optDouble("left");
            final Procedure1<Double> _function = new Procedure1<Double>() {
              public void apply(final Double it) {
                float _floatValueValid = JsonImporter.this._jsonExtensions.floatValueValid(it);
                insets.setLeft(_floatValueValid);
              }
            };
            ObjectExtensions.<Double>operator_doubleArrow(Double.valueOf(_optDouble), _function);
            double _optDouble_1 = padding.optDouble("right");
            final Procedure1<Double> _function_1 = new Procedure1<Double>() {
              public void apply(final Double it) {
                float _floatValueValid = JsonImporter.this._jsonExtensions.floatValueValid(it);
                insets.setRight(_floatValueValid);
              }
            };
            ObjectExtensions.<Double>operator_doubleArrow(Double.valueOf(_optDouble_1), _function_1);
            double _optDouble_2 = padding.optDouble("top");
            final Procedure1<Double> _function_2 = new Procedure1<Double>() {
              public void apply(final Double it) {
                float _floatValueValid = JsonImporter.this._jsonExtensions.floatValueValid(it);
                insets.setTop(_floatValueValid);
              }
            };
            ObjectExtensions.<Double>operator_doubleArrow(Double.valueOf(_optDouble_2), _function_2);
            double _optDouble_3 = padding.optDouble("bottom");
            final Procedure1<Double> _function_3 = new Procedure1<Double>() {
              public void apply(final Double it) {
                float _floatValueValid = JsonImporter.this._jsonExtensions.floatValueValid(it);
                insets.setBottom(_floatValueValid);
              }
            };
            ObjectExtensions.<Double>operator_doubleArrow(Double.valueOf(_optDouble_3), _function_3);
            InputOutput.<KInsets>println(insets);
          }
        }
      };
      _xblockexpression = ObjectExtensions.<JSONObject>operator_doubleArrow(_optJSONObject, _function_4);
    }
    return _xblockexpression;
  }
  
  /**
   * Main entry point if the passed input format has been json and the requested
   * output format is json as well.
   */
  public void transferLayout(final TransformationData<JSONObject, KNode> data) {
    List<KNode> _targetGraphs = data.getTargetGraphs();
    final Procedure1<KNode> _function = new Procedure1<KNode>() {
      public void apply(final KNode graph) {
        JsonImporter.this.transferLayoutInt(graph);
        TreeIterator<EObject> _eAllContents = graph.eAllContents();
        final Procedure1<EObject> _function = new Procedure1<EObject>() {
          public void apply(final EObject element) {
            JsonImporter.this.transferLayoutInt(element);
          }
        };
        IteratorExtensions.<EObject>forEach(_eAllContents, _function);
      }
    };
    IterableExtensions.<KNode>forEach(_targetGraphs, _function);
  }
  
  private JSONObject _transferLayoutInt(final KNode node) {
    JSONObject _xblockexpression = null;
    {
      final JSONObject jsonObj = this.nodeJsonMap.get(node);
      KShapeLayout _layout = this._jsonExtensions.layout(node);
      _xblockexpression = this.transferShapeLayout(_layout, jsonObj);
    }
    return _xblockexpression;
  }
  
  private JSONObject _transferLayoutInt(final KPort port) {
    JSONObject _xblockexpression = null;
    {
      final JSONObject jsonObj = this.portJsonMap.get(port);
      KShapeLayout _layout = this._jsonExtensions.layout(port);
      _xblockexpression = this.transferShapeLayout(_layout, jsonObj);
    }
    return _xblockexpression;
  }
  
  private JSONObject _transferLayoutInt(final KEdge edge) {
    try {
      JSONObject _xblockexpression = null;
      {
        final JSONObject jsonObj = this.edgeJsonMap.get(edge);
        final JSONObject sourcePoint = new JSONObject();
        KEdgeLayout _layout = this._jsonExtensions.layout(edge);
        KPoint _sourcePoint = _layout.getSourcePoint();
        float _x = _sourcePoint.getX();
        sourcePoint.put("x", _x);
        KEdgeLayout _layout_1 = this._jsonExtensions.layout(edge);
        KPoint _sourcePoint_1 = _layout_1.getSourcePoint();
        float _y = _sourcePoint_1.getY();
        sourcePoint.put("y", _y);
        if (jsonObj!=null) {
          jsonObj.put("sourcePoint", sourcePoint);
        }
        final JSONObject targetPoint = new JSONObject();
        KEdgeLayout _layout_2 = this._jsonExtensions.layout(edge);
        KPoint _targetPoint = _layout_2.getTargetPoint();
        float _x_1 = _targetPoint.getX();
        targetPoint.put("x", _x_1);
        KEdgeLayout _layout_3 = this._jsonExtensions.layout(edge);
        KPoint _targetPoint_1 = _layout_3.getTargetPoint();
        float _y_1 = _targetPoint_1.getY();
        targetPoint.put("y", _y_1);
        if (jsonObj!=null) {
          jsonObj.put("targetPoint", targetPoint);
        }
        final JSONArray bends = new JSONArray();
        KEdgeLayout _layout_4 = this._jsonExtensions.layout(edge);
        EList<KPoint> _bendPoints = _layout_4.getBendPoints();
        final Procedure1<KPoint> _function = new Procedure1<KPoint>() {
          public void apply(final KPoint pnt) {
            try {
              final JSONObject jsonPnt = new JSONObject();
              float _x = pnt.getX();
              jsonPnt.put("x", _x);
              float _y = pnt.getY();
              jsonPnt.put("y", _y);
              bends.put(jsonPnt);
            } catch (Throwable _e) {
              throw Exceptions.sneakyThrow(_e);
            }
          }
        };
        IterableExtensions.<KPoint>forEach(_bendPoints, _function);
        JSONObject _put = null;
        if (jsonObj!=null) {
          _put=jsonObj.put("bendPoints", bends);
        }
        _xblockexpression = _put;
      }
      return _xblockexpression;
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  private JSONObject _transferLayoutInt(final KLabel label) {
    JSONObject _xblockexpression = null;
    {
      final JSONObject jsonObj = this.labelJsonMap.get(label);
      KShapeLayout _layout = this._jsonExtensions.layout(label);
      _xblockexpression = this.transferShapeLayout(_layout, jsonObj);
    }
    return _xblockexpression;
  }
  
  private JSONObject _transferLayoutInt(final Object obj) {
    return null;
  }
  
  private JSONObject transferShapeLayout(final KShapeLayout layout, final JSONObject jsonObj) {
    try {
      JSONObject _xblockexpression = null;
      {
        if (jsonObj!=null) {
          float _xpos = layout.getXpos();
          jsonObj.put("x", _xpos);
        }
        if (jsonObj!=null) {
          float _ypos = layout.getYpos();
          jsonObj.put("y", _ypos);
        }
        if (jsonObj!=null) {
          float _width = layout.getWidth();
          jsonObj.put("width", _width);
        }
        if (jsonObj!=null) {
          float _height = layout.getHeight();
          jsonObj.put("height", _height);
        }
        final KInsets insets = layout.getInsets();
        JSONObject _xifexpression = null;
        boolean _notEquals = (!Objects.equal(insets, null));
        if (_notEquals) {
          JSONObject _xblockexpression_1 = null;
          {
            JSONObject _optJSONObject = null;
            if (jsonObj!=null) {
              _optJSONObject=jsonObj.optJSONObject("padding");
            }
            JSONObject padding = _optJSONObject;
            boolean _equals = Objects.equal(padding, null);
            if (_equals) {
              JSONObject _jSONObject = new JSONObject();
              padding = _jSONObject;
              if (jsonObj!=null) {
                jsonObj.put("padding", padding);
              }
            }
            if (padding!=null) {
              float _left = insets.getLeft();
              padding.put("left", _left);
            }
            if (padding!=null) {
              float _top = insets.getTop();
              padding.put("top", _top);
            }
            if (padding!=null) {
              float _right = insets.getRight();
              padding.put("right", _right);
            }
            JSONObject _put = null;
            if (padding!=null) {
              float _bottom = insets.getBottom();
              _put=padding.put("bottom", _bottom);
            }
            _xblockexpression_1 = _put;
          }
          _xifexpression = _xblockexpression_1;
        }
        _xblockexpression = _xifexpression;
      }
      return _xblockexpression;
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  /**
   * ---------------------------------------------------------------------------
   *   Convenience methods
   */
  public KNode register(final KNode node, final JSONObject obj) {
    final String id = obj.optString("id");
    boolean _equals = Objects.equal(id, null);
    if (_equals) {
      throw new TransformationException("Invalid json graph format. Node is missing \'id\' key.");
    }
    this.nodeIdMap.put(id, node);
    this.nodeJsonMap.put(node, obj);
    return node;
  }
  
  public KPort register(final KPort port, final JSONObject obj) {
    final String id = obj.optString("id");
    boolean _equals = Objects.equal(id, null);
    if (_equals) {
      throw new TransformationException("Invalid json graph format. Port is missing \'id\' key.");
    }
    this.portIdMap.put(id, port);
    this.portJsonMap.put(port, obj);
    return port;
  }
  
  public KEdge register(final KEdge edge, final JSONObject obj) {
    final String id = obj.optString("id");
    boolean _equals = Objects.equal(id, null);
    if (_equals) {
      throw new TransformationException("Invalid json graph format. Edge is missing \'id\' key.");
    }
    this.edgeIdMap.put(id, edge);
    this.edgeJsonMap.put(edge, obj);
    return edge;
  }
  
  private JSONObject transferLayoutInt(final Object edge) {
    if (edge instanceof KEdge) {
      return _transferLayoutInt((KEdge)edge);
    } else if (edge instanceof KNode) {
      return _transferLayoutInt((KNode)edge);
    } else if (edge instanceof KPort) {
      return _transferLayoutInt((KPort)edge);
    } else if (edge instanceof KLabel) {
      return _transferLayoutInt((KLabel)edge);
    } else if (edge != null) {
      return _transferLayoutInt(edge);
    } else {
      throw new IllegalArgumentException("Unhandled parameter types: " +
        Arrays.<Object>asList(edge).toString());
    }
  }
}
