/**
 * KIELER - Kiel Integrated Environment for Layout Eclipse RichClient
 * 
 * http://www.informatik.uni-kiel.de/rtsys/kieler/
 * 
 * Copyright 2014 by
 * + Christian-Albrechts-University of Kiel
 *   + Department of Computer Science
 *     + Real-Time and Embedded Systems Group
 * 
 * This code is provided under the terms of the Eclipse Public License (EPL).
 * See the file epl-v10.html for the license text.
 */
package de.cau.cs.kieler.kiml.grana.text;

import com.google.common.base.Objects;
import com.google.common.collect.Iterables;
import com.google.common.collect.Lists;
import de.cau.cs.kieler.core.alg.IKielerProgressMonitor;
import de.cau.cs.kieler.kiml.config.CompoundLayoutConfig;
import de.cau.cs.kieler.kiml.config.VolatileLayoutConfig;
import de.cau.cs.kieler.kiml.config.text.LayoutConfigTransformer;
import de.cau.cs.kieler.kiml.grana.AnalysisData;
import de.cau.cs.kieler.kiml.grana.AnalysisService;
import de.cau.cs.kieler.kiml.grana.text.grana.Analysis;
import de.cau.cs.kieler.kiml.grana.text.grana.GlobalResourceRef;
import de.cau.cs.kieler.kiml.grana.text.grana.Grana;
import de.cau.cs.kieler.kiml.grana.text.grana.Job;
import de.cau.cs.kieler.kiml.grana.text.grana.LocalResource;
import de.cau.cs.kieler.kiml.grana.text.grana.Resource;
import de.cau.cs.kieler.kiml.grana.text.grana.ResourceReference;
import de.cau.cs.kieler.kiml.grana.ui.batch.Batch;
import de.cau.cs.kieler.kiml.grana.ui.batch.BatchJob;
import de.cau.cs.kieler.kiml.grana.ui.batch.BatchResult;
import de.cau.cs.kieler.kiml.grana.ui.batch.CSVResultSerializer;
import de.cau.cs.kieler.kiml.grana.ui.batch.FileKGraphProvider;
import de.cau.cs.kieler.kiml.klayoutdata.KIdentifier;
import de.cau.cs.kieler.kiml.service.util.ProgressMonitorAdapter;
import java.io.File;
import java.io.IOException;
import java.io.OutputStream;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspace;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.Path;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.resource.impl.ExtensibleURIConverterImpl;
import org.eclipse.xtext.xbase.lib.CollectionLiterals;
import org.eclipse.xtext.xbase.lib.Conversions;
import org.eclipse.xtext.xbase.lib.Exceptions;
import org.eclipse.xtext.xbase.lib.Functions.Function1;
import org.eclipse.xtext.xbase.lib.IterableExtensions;
import org.eclipse.xtext.xbase.lib.ListExtensions;
import org.eclipse.xtext.xbase.lib.StringExtensions;

/**
 * Utility class to convert textually specified grana executions to
 * {@link BatchJob}s and execute them.
 * 
 * @author uru
 * @kieler.ignore (excluded from review process)
 */
@SuppressWarnings("all")
public final class GranaTextToBatchJob {
  /**
   * @return an iterable with the results of all executed batch jobs.
   *         These can be used to print problems that might have occurred during the execution.
   */
  public static Iterable<BatchResult> execute(final Grana grana, final IProgressMonitor progressMonitor) {
    try {
      final ProgressMonitorAdapter pm = new ProgressMonitorAdapter(progressMonitor);
      IWorkspace _workspace = ResourcesPlugin.getWorkspace();
      final IWorkspaceRoot wsRoot = _workspace.getRoot();
      final LinkedList<BatchResult> results = Lists.<BatchResult>newLinkedList();
      EList<Job> _jobs = grana.getJobs();
      for (final Job job : _jobs) {
        String _name = job.getName();
        boolean _startsWith = _name.startsWith("_");
        boolean _not = (!_startsWith);
        if (_not) {
          EList<Analysis> _analyses = job.getAnalyses();
          final Function1<Analysis, AnalysisData> _function = new Function1<Analysis, AnalysisData>() {
            public AnalysisData apply(final Analysis it) {
              AnalysisService _instance = AnalysisService.getInstance();
              String _name = it.getName();
              return _instance.getAnalysis(_name);
            }
          };
          final List<AnalysisData> analyses = ListExtensions.<Analysis, AnalysisData>map(_analyses, _function);
          final Batch batch = new Batch(analyses);
          EList<Resource> _resources = job.getResources();
          final Function1<Resource, Iterable<LocalResource>> _function_1 = new Function1<Resource, Iterable<LocalResource>>() {
            public Iterable<LocalResource> apply(final Resource r) {
              Iterable<LocalResource> _switchResult = null;
              boolean _matched = false;
              if (!_matched) {
                if (r instanceof ResourceReference) {
                  _matched=true;
                  EList<GlobalResourceRef> _resourceRefs = ((ResourceReference) r).getResourceRefs();
                  final Function1<GlobalResourceRef, EList<LocalResource>> _function = new Function1<GlobalResourceRef, EList<LocalResource>>() {
                    public EList<LocalResource> apply(final GlobalResourceRef it) {
                      return it.getResources();
                    }
                  };
                  List<EList<LocalResource>> _map = ListExtensions.<GlobalResourceRef, EList<LocalResource>>map(_resourceRefs, _function);
                  _switchResult = Iterables.<LocalResource>concat(_map);
                }
              }
              if (!_matched) {
                _switchResult = Collections.<LocalResource>unmodifiableList(CollectionLiterals.<LocalResource>newArrayList(((LocalResource) r)));
              }
              return _switchResult;
            }
          };
          List<Iterable<LocalResource>> _map = ListExtensions.<Resource, Iterable<LocalResource>>map(_resources, _function_1);
          final Iterable<LocalResource> resources = Iterables.<LocalResource>concat(_map);
          for (final LocalResource resource : resources) {
            {
              Pattern _xifexpression = null;
              String _filter = resource.getFilter();
              boolean _isNullOrEmpty = StringExtensions.isNullOrEmpty(_filter);
              boolean _not_1 = (!_isNullOrEmpty);
              if (_not_1) {
                String _filter_1 = resource.getFilter();
                _xifexpression = Pattern.compile(_filter_1);
              } else {
                _xifexpression = null;
              }
              final Pattern filter = _xifexpression;
              try {
                IProject[] _projects = wsRoot.getProjects();
                final Function1<IProject, Boolean> _function_2 = new Function1<IProject, Boolean>() {
                  public Boolean apply(final IProject p) {
                    String _path = resource.getPath();
                    String _name = p.getName();
                    return Boolean.valueOf(_path.contains(_name));
                  }
                };
                final IProject p = IterableExtensions.<IProject>findFirst(((Iterable<IProject>)Conversions.doWrapArray(_projects)), _function_2);
                boolean _equals = Objects.equal(p, null);
                if (_equals) {
                  String _path = resource.getPath();
                  String _plus = ("Could not find resource " + _path);
                  throw new IllegalArgumentException(_plus);
                }
                String _path_1 = resource.getPath();
                String _name_1 = p.getName();
                String _replace = _path_1.replace(_name_1, "");
                final IResource wsloc = p.findMember(_replace);
                IResource[] _members = ((IContainer) wsloc).members();
                for (final IResource file : _members) {
                  boolean _or = false;
                  boolean _equals_1 = Objects.equal(filter, null);
                  if (_equals_1) {
                    _or = true;
                  } else {
                    String _name_2 = file.getName();
                    Matcher _matcher = filter.matcher(_name_2);
                    boolean _matches = _matcher.matches();
                    _or = _matches;
                  }
                  if (_or) {
                    String _path_2 = resource.getPath();
                    String _plus_1 = (_path_2 + "/");
                    String _name_3 = file.getName();
                    String _plus_2 = (_plus_1 + _name_3);
                    Path _path_3 = new Path(_plus_2);
                    GranaTextToBatchJob.addBatchJob(batch, job, _path_3);
                  }
                }
              } catch (final Throwable _t) {
                if (_t instanceof Exception) {
                  final Exception e = (Exception)_t;
                  String _path_4 = resource.getPath();
                  final URI fileURI = URI.createURI(_path_4, true);
                  String _fileString = fileURI.toFileString();
                  final File dir = new File(_fileString);
                  boolean _exists = dir.exists();
                  boolean _not_2 = (!_exists);
                  if (_not_2) {
                    String _path_5 = resource.getPath();
                    String _plus_3 = ("Could not find resource location: \'" + _path_5);
                    String _plus_4 = (_plus_3 + "\'");
                    throw new IllegalArgumentException(_plus_4);
                  }
                  File[] _listFiles = dir.listFiles();
                  for (final File file_1 : _listFiles) {
                    boolean _or_1 = false;
                    boolean _equals_2 = Objects.equal(filter, null);
                    if (_equals_2) {
                      _or_1 = true;
                    } else {
                      String _name_4 = file_1.getName();
                      Matcher _matcher_1 = filter.matcher(_name_4);
                      boolean _matches_1 = _matcher_1.matches();
                      _or_1 = _matches_1;
                    }
                    if (_or_1) {
                      String _absolutePath = file_1.getAbsolutePath();
                      Path _path_6 = new Path(_absolutePath);
                      GranaTextToBatchJob.addBatchJob(batch, job, _path_6);
                    }
                  }
                } else {
                  throw Exceptions.sneakyThrow(_t);
                }
              }
            }
          }
          IKielerProgressMonitor _subTask = pm.subTask(1);
          final BatchResult result = batch.execute(_subTask);
          boolean _isCanceled = pm.isCanceled();
          if (_isCanceled) {
            return Collections.<BatchResult>unmodifiableList(CollectionLiterals.<BatchResult>newArrayList());
          }
          results.add(result);
          try {
            OutputStream os = null;
            String _output = job.getOutput();
            final URI fileURI = URI.createURI(_output, true);
            final ExtensibleURIConverterImpl uriConv = new ExtensibleURIConverterImpl();
            OutputStream _createOutputStream = uriConv.createOutputStream(fileURI);
            os = _createOutputStream;
            final CSVResultSerializer serializer = new CSVResultSerializer();
            IKielerProgressMonitor _subTask_1 = pm.subTask(1);
            serializer.serialize(os, result, _subTask_1);
            os.close();
          } catch (final Throwable _t) {
            if (_t instanceof IOException) {
              final IOException e = (IOException)_t;
              String _output_1 = job.getOutput();
              String _plus = ("Could not store results to file \'" + _output_1);
              String _plus_1 = (_plus + "\'");
              throw new IllegalArgumentException(_plus_1);
            } else {
              throw Exceptions.sneakyThrow(_t);
            }
          }
        }
      }
      return results;
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  private static void addBatchJob(final Batch batch, final Job job, final IPath path) {
    final FileKGraphProvider provider = new FileKGraphProvider();
    boolean _isLayoutBeforeAnalysis = job.isLayoutBeforeAnalysis();
    provider.setLayoutBeforeAnalysis(_isLayoutBeforeAnalysis);
    boolean _isMeasureExecutionTime = job.isMeasureExecutionTime();
    provider.setExecutionTimeAnalysis(_isMeasureExecutionTime);
    final CompoundLayoutConfig clc = new CompoundLayoutConfig();
    EList<KIdentifier> _layoutOptions = job.getLayoutOptions();
    for (final KIdentifier cfg : _layoutOptions) {
      VolatileLayoutConfig _transformConfig = LayoutConfigTransformer.transformConfig(cfg);
      clc.add(_transformConfig);
    }
    provider.setLayoutConfigurator(clc);
    final BatchJob<IPath> batchJob = new BatchJob<IPath>(path, provider);
    batch.appendJob(batchJob);
  }
}
