/**
 * KIELER - Kiel Integrated Environment for Layout Eclipse RichClient
 * 
 * http://www.informatik.uni-kiel.de/rtsys/kieler/
 * 
 * Copyright 2013 by
 * + Christian-Albrechts-University of Kiel
 *   + Department of Computer Science
 *     + Real-Time and Embedded Systems Group
 * 
 * This code is provided under the terms of the Eclipse Public License (EPL).
 * See the file epl-v10.html for the license text.
 */
package de.cau.cs.kieler.ptolemy.klighd.transformation;

import com.google.common.base.Objects;
import com.google.inject.Inject;
import de.cau.cs.kieler.core.kgraph.KEdge;
import de.cau.cs.kieler.core.kgraph.KGraphData;
import de.cau.cs.kieler.core.kgraph.KNode;
import de.cau.cs.kieler.core.kgraph.KPort;
import de.cau.cs.kieler.core.krendering.HorizontalAlignment;
import de.cau.cs.kieler.core.krendering.KBackground;
import de.cau.cs.kieler.core.krendering.KColor;
import de.cau.cs.kieler.core.krendering.KContainerRendering;
import de.cau.cs.kieler.core.krendering.KDecoratorPlacementData;
import de.cau.cs.kieler.core.krendering.KEllipse;
import de.cau.cs.kieler.core.krendering.KGridPlacement;
import de.cau.cs.kieler.core.krendering.KPlacementData;
import de.cau.cs.kieler.core.krendering.KPointPlacementData;
import de.cau.cs.kieler.core.krendering.KPolygon;
import de.cau.cs.kieler.core.krendering.KPolyline;
import de.cau.cs.kieler.core.krendering.KPosition;
import de.cau.cs.kieler.core.krendering.KRectangle;
import de.cau.cs.kieler.core.krendering.KRendering;
import de.cau.cs.kieler.core.krendering.KRenderingFactory;
import de.cau.cs.kieler.core.krendering.KRenderingLibrary;
import de.cau.cs.kieler.core.krendering.KRenderingRef;
import de.cau.cs.kieler.core.krendering.KRotation;
import de.cau.cs.kieler.core.krendering.KRoundedBendsPolyline;
import de.cau.cs.kieler.core.krendering.KRoundedRectangle;
import de.cau.cs.kieler.core.krendering.KSpline;
import de.cau.cs.kieler.core.krendering.KStyle;
import de.cau.cs.kieler.core.krendering.KStyleHolder;
import de.cau.cs.kieler.core.krendering.KText;
import de.cau.cs.kieler.core.krendering.LineStyle;
import de.cau.cs.kieler.core.krendering.VerticalAlignment;
import de.cau.cs.kieler.core.krendering.extensions.KColorExtensions;
import de.cau.cs.kieler.core.krendering.extensions.KContainerRenderingExtensions;
import de.cau.cs.kieler.core.krendering.extensions.KPolylineExtensions;
import de.cau.cs.kieler.core.krendering.extensions.KRenderingExtensions;
import de.cau.cs.kieler.core.util.Pair;
import de.cau.cs.kieler.kiml.klayoutdata.KLayoutData;
import de.cau.cs.kieler.klighd.KlighdConstants;
import de.cau.cs.kieler.ptolemy.klighd.PtolemyProperties;
import de.cau.cs.kieler.ptolemy.klighd.transformation.PtolemyFigureInterface;
import de.cau.cs.kieler.ptolemy.klighd.transformation.extensions.AnnotationExtensions;
import de.cau.cs.kieler.ptolemy.klighd.transformation.extensions.LabelExtensions;
import de.cau.cs.kieler.ptolemy.klighd.transformation.extensions.MarkerExtensions;
import de.cau.cs.kieler.ptolemy.klighd.transformation.extensions.MiscellaneousExtensions;
import de.cau.cs.kieler.ptolemy.klighd.transformation.util.GraphicsUtils;
import de.cau.cs.kieler.ptolemy.klighd.transformation.util.TransformationConstants;
import java.util.List;
import org.eclipse.emf.common.util.EList;
import org.eclipse.xtext.xbase.lib.Extension;
import org.eclipse.xtext.xbase.lib.Functions.Function1;
import org.eclipse.xtext.xbase.lib.IterableExtensions;
import org.eclipse.xtext.xbase.lib.ObjectExtensions;
import org.eclipse.xtext.xbase.lib.Procedures.Procedure1;

/**
 * Creates concrete KRendering information for Ptolemy diagram elements.
 * 
 * @author ckru
 * @author cds
 * @author uru
 */
@SuppressWarnings("all")
public class KRenderingFigureProvider {
  /**
   * Accessing annotations.
   */
  @Inject
  @Extension
  private AnnotationExtensions _annotationExtensions;
  
  /**
   * Handling labels.
   */
  @Inject
  @Extension
  private LabelExtensions _labelExtensions;
  
  /**
   * Marking nodes.
   */
  @Inject
  @Extension
  private MarkerExtensions _markerExtensions;
  
  /**
   * Extensions used during the transformation. To make things easier. And stuff.
   */
  @Inject
  @Extension
  private MiscellaneousExtensions _miscellaneousExtensions;
  
  /**
   * Create KRenderings from Ptolemy figures.
   */
  @Inject
  @Extension
  private PtolemyFigureInterface _ptolemyFigureInterface;
  
  /**
   * Color stuff.
   */
  @Inject
  @Extension
  private KColorExtensions _kColorExtensions;
  
  /**
   * Rendering stuff.
   */
  @Inject
  @Extension
  private KRenderingExtensions _kRenderingExtensions;
  
  /**
   * Rendering stuff.
   */
  @Inject
  @Extension
  private KContainerRenderingExtensions _kContainerRenderingExtensions;
  
  /**
   * Rendering stuff.
   */
  @Inject
  @Extension
  private KPolylineExtensions _kPolylineExtensions;
  
  /**
   * Rendering factory used to instantiate KRendering instances.
   */
  private final KRenderingFactory renderingFactory = KRenderingFactory.eINSTANCE;
  
  /**
   * Returns the rendering library associated with the given node. The associated rendering library is
   * the library of the root ancestor of the node. If no library exists yet, one is created and
   * attached to the root ancestor.
   * 
   * @param node the node whose's rendering library to return.
   * @return the rendering library associated with the given node.
   */
  private KRenderingLibrary getLibrary(final KNode node) {
    KNode parent = node;
    while ((!Objects.equal(parent.getParent(), null))) {
      KNode _parent = parent.getParent();
      parent = _parent;
    }
    KRenderingLibrary library = parent.<KRenderingLibrary>getData(KRenderingLibrary.class);
    boolean _equals = Objects.equal(library, null);
    if (_equals) {
      KRenderingLibrary _createKRenderingLibrary = this.renderingFactory.createKRenderingLibrary();
      library = _createKRenderingLibrary;
      EList<KGraphData> _data = parent.getData();
      _data.add(library);
    }
    return library;
  }
  
  /**
   * Returns a reference to rendering with the given ID in the given library.
   * 
   * @param id identifier of the rendering to look up.
   * @param library the rendering library to search for the rendering.
   * @return new rendering reference to the rendering, or {@code null} if no rendering with the given
   *         identifier exists in the library.
   */
  private KRenderingRef getFromLibrary(final String id, final KRenderingLibrary library) {
    EList<KStyleHolder> _renderings = library.getRenderings();
    final Function1<KStyleHolder, Boolean> _function = new Function1<KStyleHolder, Boolean>() {
      public Boolean apply(final KStyleHolder r) {
        String _id = r.getId();
        return Boolean.valueOf(Objects.equal(_id, id));
      }
    };
    KStyleHolder _findFirst = IterableExtensions.<KStyleHolder>findFirst(_renderings, _function);
    final KRendering rendering = ((KRendering) _findFirst);
    boolean _notEquals = (!Objects.equal(rendering, null));
    if (_notEquals) {
      final KRenderingRef ref = this.renderingFactory.createKRenderingRef();
      ref.setRendering(rendering);
      return ref;
    } else {
      return null;
    }
  }
  
  /**
   * Adds the given rendering to the given rendering library with the given id.
   * 
   * @param rendering the rendering to add to the library.
   * @param id the id that will identify the rendering in the library.
   * @param library the library to add the rendering to.
   */
  private KRenderingRef addToLibrary(final KRendering rendering, final String id, final KRenderingLibrary library) {
    rendering.setId(id);
    EList<KStyleHolder> _renderings = library.getRenderings();
    _renderings.add(rendering);
    final KRenderingRef ref = this.renderingFactory.createKRenderingRef();
    ref.setRendering(rendering);
    return ref;
  }
  
  /**
   * Renders the given node in a default way, that is, as a simple rectangle.
   * 
   * @return the created rendering.
   */
  public KRendering createDefaultRendering(final KNode node) {
    KRectangle _createKRectangle = this.renderingFactory.createKRectangle();
    final Procedure1<KRectangle> _function = new Procedure1<KRectangle>() {
      public void apply(final KRectangle rect) {
        KRenderingFigureProvider.this._kRenderingExtensions.<KRectangle>setBackgroundColor(rect, 255, 255, 255);
      }
    };
    final KRectangle rendering = ObjectExtensions.<KRectangle>operator_doubleArrow(_createKRectangle, _function);
    return rendering;
  }
  
  /**
   * Creates a rendering for an expanded compound node.
   * 
   * @param node the node to create the rendering information for.
   * @param alpha the alpha value of the compound node background.
   * @return the rendering.
   */
  public KRendering createExpandedCompoundNodeRendering(final KNode node, final int alpha) {
    KColor _xifexpression = null;
    boolean _isMarkedAsState = this._markerExtensions.isMarkedAsState(node);
    if (_isMarkedAsState) {
      KColor _createKColor = this.renderingFactory.createKColor();
      final Procedure1<KColor> _function = new Procedure1<KColor>() {
        public void apply(final KColor col) {
          col.setRed(11);
          col.setGreen(188);
          col.setBlue(11);
        }
      };
      _xifexpression = ObjectExtensions.<KColor>operator_doubleArrow(_createKColor, _function);
    } else {
      KColor _createKColor_1 = this.renderingFactory.createKColor();
      final Procedure1<KColor> _function_1 = new Procedure1<KColor>() {
        public void apply(final KColor col) {
          col.setRed(16);
          col.setGreen(78);
          col.setBlue(139);
        }
      };
      _xifexpression = ObjectExtensions.<KColor>operator_doubleArrow(_createKColor_1, _function_1);
    }
    final KColor bgColor = _xifexpression;
    KRoundedRectangle _createKRoundedRectangle = this.renderingFactory.createKRoundedRectangle();
    final Procedure1<KRoundedRectangle> _function_2 = new Procedure1<KRoundedRectangle>() {
      public void apply(final KRoundedRectangle rect) {
        rect.setCornerHeight(15);
        rect.setCornerWidth(15);
        int _xifexpression = (int) 0;
        boolean _isMarkedAsState = KRenderingFigureProvider.this._markerExtensions.isMarkedAsState(node);
        if (_isMarkedAsState) {
          _xifexpression = 1;
        } else {
          _xifexpression = 0;
        }
        KRenderingFigureProvider.this._kRenderingExtensions.<KRoundedRectangle>setLineWidth(rect, _xifexpression);
        EList<KStyle> _styles = rect.getStyles();
        KBackground _createKBackground = KRenderingFigureProvider.this.renderingFactory.createKBackground();
        final Procedure1<KBackground> _function = new Procedure1<KBackground>() {
          public void apply(final KBackground bg) {
            bg.setAlpha(alpha);
            bg.setColor(bgColor);
          }
        };
        KBackground _doubleArrow = ObjectExtensions.<KBackground>operator_doubleArrow(_createKBackground, _function);
        _styles.add(_doubleArrow);
      }
    };
    final KRoundedRectangle rendering = ObjectExtensions.<KRoundedRectangle>operator_doubleArrow(_createKRoundedRectangle, _function_2);
    return rendering;
  }
  
  /**
   * Creates a rendering for a relation node.
   * 
   * @param node the node to create the rendering information for.
   * @return the rendering.
   */
  public KRendering createRelationNodeRendering(final KNode node) {
    final KRenderingLibrary library = this.getLibrary(node);
    final KRenderingRef rendering = this.getFromLibrary("ren_relation", library);
    boolean _notEquals = (!Objects.equal(rendering, null));
    if (_notEquals) {
      return rendering;
    }
    KPolygon _createKPolygon = this.renderingFactory.createKPolygon();
    final Procedure1<KPolygon> _function = new Procedure1<KPolygon>() {
      public void apply(final KPolygon poly) {
        EList<KPosition> _points = poly.getPoints();
        KPosition _createKPosition = KRenderingFigureProvider.this._miscellaneousExtensions.createKPosition(4, 0);
        _points.add(_createKPosition);
        EList<KPosition> _points_1 = poly.getPoints();
        KPosition _createKPosition_1 = KRenderingFigureProvider.this._miscellaneousExtensions.createKPosition(8, 4);
        _points_1.add(_createKPosition_1);
        EList<KPosition> _points_2 = poly.getPoints();
        KPosition _createKPosition_2 = KRenderingFigureProvider.this._miscellaneousExtensions.createKPosition(4, 8);
        _points_2.add(_createKPosition_2);
        EList<KPosition> _points_3 = poly.getPoints();
        KPosition _createKPosition_3 = KRenderingFigureProvider.this._miscellaneousExtensions.createKPosition(0, 4);
        _points_3.add(_createKPosition_3);
        EList<KPosition> _points_4 = poly.getPoints();
        KPosition _createKPosition_4 = KRenderingFigureProvider.this._miscellaneousExtensions.createKPosition(4, 0);
        _points_4.add(_createKPosition_4);
        KRenderingFigureProvider.this._kRenderingExtensions.<KPolygon>setBackgroundColor(poly, 0, 0, 0);
      }
    };
    KPolygon _doubleArrow = ObjectExtensions.<KPolygon>operator_doubleArrow(_createKPolygon, _function);
    return this.addToLibrary(_doubleArrow, "ren_relation", library);
  }
  
  /**
   * Creates a rendering for a director node.
   * 
   * @param node the node to create the rendering information for.
   * @return the rendering.
   */
  public KRendering createDirectorNodeRendering(final KNode node) {
    final KRenderingLibrary library = this.getLibrary(node);
    final KRenderingRef rendering = this.getFromLibrary("ren_director", library);
    boolean _notEquals = (!Objects.equal(rendering, null));
    if (_notEquals) {
      return rendering;
    }
    KRectangle _createKRectangle = this.renderingFactory.createKRectangle();
    final Procedure1<KRectangle> _function = new Procedure1<KRectangle>() {
      public void apply(final KRectangle rec) {
        KColor _color = KRenderingFigureProvider.this._kColorExtensions.getColor("green");
        KRenderingFigureProvider.this._kRenderingExtensions.setBackground(rec, _color);
        KColor _color_1 = KRenderingFigureProvider.this._kColorExtensions.getColor("black");
        KRenderingFigureProvider.this._kRenderingExtensions.setForeground(rec, _color_1);
        KPosition _createKPosition = KRenderingFigureProvider.this._kRenderingExtensions.createKPosition(KRenderingFigureProvider.this._kRenderingExtensions.LEFT, 0, 0, KRenderingFigureProvider.this._kRenderingExtensions.TOP, 0, 0);
        KPosition _createKPosition_1 = KRenderingFigureProvider.this._kRenderingExtensions.createKPosition(KRenderingFigureProvider.this._kRenderingExtensions.LEFT, 100, 0, KRenderingFigureProvider.this._kRenderingExtensions.TOP, 30, 0);
        KRenderingFigureProvider.this._kRenderingExtensions.<KRectangle>setAreaPlacementData(rec, _createKPosition, _createKPosition_1);
      }
    };
    KRectangle _doubleArrow = ObjectExtensions.<KRectangle>operator_doubleArrow(_createKRectangle, _function);
    return this.addToLibrary(_doubleArrow, "ren_director", library);
  }
  
  /**
   * Creates a rendering for a comment node.
   * 
   * @param node the node to create the rendering information for.
   * @return the rendering.
   */
  public KRendering createCommentNodeRendering(final KNode node) {
    KRectangle _createKRectangle = this.renderingFactory.createKRectangle();
    final Procedure1<KRectangle> _function = new Procedure1<KRectangle>() {
      public void apply(final KRectangle rec) {
        KColor _createKColor = KRenderingFigureProvider.this.renderingFactory.createKColor();
        final Procedure1<KColor> _function = new Procedure1<KColor>() {
          public void apply(final KColor col) {
            col.setRed(255);
            col.setGreen(255);
            col.setBlue(204);
          }
        };
        KColor _doubleArrow = ObjectExtensions.<KColor>operator_doubleArrow(_createKColor, _function);
        KRenderingFigureProvider.this._kRenderingExtensions.setBackground(rec, _doubleArrow);
        KRenderingFigureProvider.this._kRenderingExtensions.<KRectangle>setLineWidth(rec, 0);
      }
    };
    final KRectangle rectangle = ObjectExtensions.<KRectangle>operator_doubleArrow(_createKRectangle, _function);
    EList<KRendering> _children = rectangle.getChildren();
    KText _createKText = this.renderingFactory.createKText();
    final Procedure1<KText> _function_1 = new Procedure1<KText>() {
      public void apply(final KText text) {
        KLayoutData _layout = KRenderingFigureProvider.this._annotationExtensions.getLayout(node);
        String _property = _layout.<String>getProperty(PtolemyProperties.COMMENT_TEXT);
        text.setText(_property);
        KRenderingFigureProvider.this._kRenderingExtensions.<KText>setSurroundingSpace(text, 5, 0);
      }
    };
    KText _doubleArrow = ObjectExtensions.<KText>operator_doubleArrow(_createKText, _function_1);
    _children.add(_doubleArrow);
    return rectangle;
  }
  
  /**
   * Creates a rendering for a parameter node. Parameter nodes display model parameters in a grid-like
   * fashion.
   * 
   * @param node the node to create the rendering information for.
   * @return the rendering.
   */
  public KRendering createParameterNodeRendering(final KNode node) {
    KRectangle _createKRectangle = this.renderingFactory.createKRectangle();
    final Procedure1<KRectangle> _function = new Procedure1<KRectangle>() {
      public void apply(final KRectangle rec) {
        KRenderingFigureProvider.this._kRenderingExtensions.setForegroundInvisible(rec, true);
        KGridPlacement _createKGridPlacement = KRenderingFigureProvider.this.renderingFactory.createKGridPlacement();
        final Procedure1<KGridPlacement> _function = new Procedure1<KGridPlacement>() {
          public void apply(final KGridPlacement grid) {
            grid.setNumColumns(3);
          }
        };
        KGridPlacement _doubleArrow = ObjectExtensions.<KGridPlacement>operator_doubleArrow(_createKGridPlacement, _function);
        rec.setChildPlacement(_doubleArrow);
      }
    };
    final KRectangle rectangle = ObjectExtensions.<KRectangle>operator_doubleArrow(_createKRectangle, _function);
    KLayoutData _layout = this._annotationExtensions.getLayout(node);
    final List<Pair<String, String>> parameters = _layout.<List<Pair<String, String>>>getProperty(PtolemyProperties.PT_PARAMETERS);
    for (final Pair<String, String> parameter : parameters) {
      {
        KEllipse _createKEllipse = this.renderingFactory.createKEllipse();
        final Procedure1<KEllipse> _function_1 = new Procedure1<KEllipse>() {
          public void apply(final KEllipse ell) {
            KColor _lookupColor = GraphicsUtils.lookupColor("blue");
            KRenderingFigureProvider.this._kRenderingExtensions.setBackground(ell, _lookupColor);
            KPosition _createKPosition = KRenderingFigureProvider.this._kRenderingExtensions.createKPosition(KRenderingFigureProvider.this._kRenderingExtensions.LEFT, (-4), 0.5f, KRenderingFigureProvider.this._kRenderingExtensions.TOP, (-4), 0.5f);
            KPosition _createKPosition_1 = KRenderingFigureProvider.this._kRenderingExtensions.createKPosition(KRenderingFigureProvider.this._kRenderingExtensions.RIGHT, (-4), 0.5f, KRenderingFigureProvider.this._kRenderingExtensions.BOTTOM, (-4), 0.5f);
            KRenderingFigureProvider.this._kRenderingExtensions.setGridPlacementData(ell, 
              15, 
              15, _createKPosition, _createKPosition_1);
            KRenderingFigureProvider.this._kRenderingExtensions.setLineWidth(ell, 1);
          }
        };
        final KEllipse circle = ObjectExtensions.<KEllipse>operator_doubleArrow(_createKEllipse, _function_1);
        EList<KRendering> _children = rectangle.getChildren();
        _children.add(circle);
        KText _createKText = this.renderingFactory.createKText();
        final Procedure1<KText> _function_2 = new Procedure1<KText>() {
          public void apply(final KText name) {
            String _first = parameter.getFirst();
            String _plus = (_first + ":");
            name.setText(_plus);
            KRenderingFigureProvider.this._kRenderingExtensions.setHorizontalAlignment(name, KRenderingFigureProvider.this._kRenderingExtensions.H_LEFT);
            KRenderingFigureProvider.this._kRenderingExtensions.<KText>setFontSize(name, (KlighdConstants.DEFAULT_FONT_SIZE - 2));
            KPosition _createKPosition = KRenderingFigureProvider.this._kRenderingExtensions.createKPosition(KRenderingFigureProvider.this._kRenderingExtensions.LEFT, 0, 0, KRenderingFigureProvider.this._kRenderingExtensions.TOP, 3, 0);
            KPosition _createKPosition_1 = KRenderingFigureProvider.this._kRenderingExtensions.createKPosition(KRenderingFigureProvider.this._kRenderingExtensions.RIGHT, 5, 0, KRenderingFigureProvider.this._kRenderingExtensions.BOTTOM, 3, 0);
            KRenderingFigureProvider.this._kRenderingExtensions.setGridPlacementData(name, 
              0, 
              0, _createKPosition, _createKPosition_1);
          }
        };
        final KText nameText = ObjectExtensions.<KText>operator_doubleArrow(_createKText, _function_2);
        EList<KRendering> _children_1 = rectangle.getChildren();
        _children_1.add(nameText);
        KText _createKText_1 = this.renderingFactory.createKText();
        final Procedure1<KText> _function_3 = new Procedure1<KText>() {
          public void apply(final KText value) {
            String _xifexpression = null;
            String _second = parameter.getSecond();
            int _length = _second.length();
            boolean _greaterThan = (_length > 150);
            if (_greaterThan) {
              String _second_1 = parameter.getSecond();
              String _substring = _second_1.substring(0, 150);
              _xifexpression = (_substring + "...");
            } else {
              _xifexpression = parameter.getSecond();
            }
            value.setText(_xifexpression);
            KRenderingFigureProvider.this._kRenderingExtensions.setHorizontalAlignment(value, KRenderingFigureProvider.this._kRenderingExtensions.H_LEFT);
            KRenderingFigureProvider.this._kRenderingExtensions.<KText>setFontSize(value, (KlighdConstants.DEFAULT_FONT_SIZE - 2));
            KPosition _createKPosition = KRenderingFigureProvider.this._kRenderingExtensions.createKPosition(KRenderingFigureProvider.this._kRenderingExtensions.LEFT, 5, 0, KRenderingFigureProvider.this._kRenderingExtensions.TOP, 3, 0);
            KPosition _createKPosition_1 = KRenderingFigureProvider.this._kRenderingExtensions.createKPosition(KRenderingFigureProvider.this._kRenderingExtensions.RIGHT, 5, 0, KRenderingFigureProvider.this._kRenderingExtensions.BOTTOM, 3, 0);
            KRenderingFigureProvider.this._kRenderingExtensions.setGridPlacementData(value, 
              0, 
              0, _createKPosition, _createKPosition_1);
          }
        };
        final KText valueText = ObjectExtensions.<KText>operator_doubleArrow(_createKText_1, _function_3);
        EList<KRendering> _children_2 = rectangle.getChildren();
        _children_2.add(valueText);
      }
    }
    return rectangle;
  }
  
  /**
   * Creates a rendering for a state node.
   * 
   * @param node the node to create the rendering information for.
   * @return the rendering.
   */
  public KRendering createStateNodeRendering(final KNode node) {
    final boolean isFinal = this._annotationExtensions.getAnnotationBooleanValue(node, "isFinalState");
    final boolean isInitial = this._annotationExtensions.getAnnotationBooleanValue(node, "isInitialState");
    int _xifexpression = (int) 0;
    if (isInitial) {
      _xifexpression = 3;
    } else {
      _xifexpression = 1;
    }
    final int lineWidth = _xifexpression;
    int _xifexpression_1 = (int) 0;
    if (isInitial) {
      _xifexpression_1 = 4;
    } else {
      _xifexpression_1 = 3;
    }
    final int initialFinalInset = _xifexpression_1;
    KColor _xifexpression_2 = null;
    boolean _isMarkedAsHavingRefinement = this._markerExtensions.isMarkedAsHavingRefinement(node);
    if (_isMarkedAsHavingRefinement) {
      KColor _createKColor = this.renderingFactory.createKColor();
      final Procedure1<KColor> _function = new Procedure1<KColor>() {
        public void apply(final KColor col) {
          col.setRed(204);
          col.setGreen(255);
          col.setBlue(204);
        }
      };
      _xifexpression_2 = ObjectExtensions.<KColor>operator_doubleArrow(_createKColor, _function);
    } else {
      _xifexpression_2 = GraphicsUtils.lookupColor("white");
    }
    final KColor bgColor = _xifexpression_2;
    KText _createKText = this.renderingFactory.createKText();
    final Procedure1<KText> _function_1 = new Procedure1<KText>() {
      public void apply(final KText text) {
        String _name = KRenderingFigureProvider.this._labelExtensions.getName(node);
        text.setText(_name);
        KPosition _createKPosition = KRenderingFigureProvider.this._kRenderingExtensions.createKPosition(KRenderingFigureProvider.this._kRenderingExtensions.LEFT, 14, 0, KRenderingFigureProvider.this._kRenderingExtensions.TOP, 8, 0);
        KPosition _createKPosition_1 = KRenderingFigureProvider.this._kRenderingExtensions.createKPosition(KRenderingFigureProvider.this._kRenderingExtensions.RIGHT, 14, 0, KRenderingFigureProvider.this._kRenderingExtensions.BOTTOM, 8, 0);
        KRenderingFigureProvider.this._kRenderingExtensions.<KText>setAreaPlacementData(text, _createKPosition, _createKPosition_1);
      }
    };
    final KText label = ObjectExtensions.<KText>operator_doubleArrow(_createKText, _function_1);
    this._labelExtensions.setName(node, "");
    KRoundedRectangle _createKRoundedRectangle = this.renderingFactory.createKRoundedRectangle();
    final Procedure1<KRoundedRectangle> _function_2 = new Procedure1<KRoundedRectangle>() {
      public void apply(final KRoundedRectangle rec) {
        rec.setCornerHeight(30);
        rec.setCornerWidth(15);
        KPosition _createKPosition = KRenderingFigureProvider.this._kRenderingExtensions.createKPosition(KRenderingFigureProvider.this._kRenderingExtensions.LEFT, 0, 0, KRenderingFigureProvider.this._kRenderingExtensions.TOP, 0, 0);
        KPosition _createKPosition_1 = KRenderingFigureProvider.this._kRenderingExtensions.createKPosition(KRenderingFigureProvider.this._kRenderingExtensions.RIGHT, 0, 0, KRenderingFigureProvider.this._kRenderingExtensions.BOTTOM, 0, 0);
        KRenderingFigureProvider.this._kRenderingExtensions.<KRoundedRectangle>setAreaPlacementData(rec, _createKPosition, _createKPosition_1);
        KRenderingFigureProvider.this._kRenderingExtensions.setLineWidth(rec, lineWidth);
        KRenderingFigureProvider.this._kRenderingExtensions.setBackground(rec, bgColor);
      }
    };
    final KRoundedRectangle outerCircle = ObjectExtensions.<KRoundedRectangle>operator_doubleArrow(_createKRoundedRectangle, _function_2);
    if (isFinal) {
      KRoundedRectangle _createKRoundedRectangle_1 = this.renderingFactory.createKRoundedRectangle();
      final Procedure1<KRoundedRectangle> _function_3 = new Procedure1<KRoundedRectangle>() {
        public void apply(final KRoundedRectangle rec) {
          rec.setCornerHeight(22);
          rec.setCornerWidth(12);
          KPosition _createKPosition = KRenderingFigureProvider.this._kRenderingExtensions.createKPosition(KRenderingFigureProvider.this._kRenderingExtensions.LEFT, initialFinalInset, 0, KRenderingFigureProvider.this._kRenderingExtensions.TOP, initialFinalInset, 0);
          KPosition _createKPosition_1 = KRenderingFigureProvider.this._kRenderingExtensions.createKPosition(KRenderingFigureProvider.this._kRenderingExtensions.RIGHT, initialFinalInset, 0, KRenderingFigureProvider.this._kRenderingExtensions.BOTTOM, initialFinalInset, 0);
          KRenderingFigureProvider.this._kRenderingExtensions.<KRoundedRectangle>setAreaPlacementData(rec, _createKPosition, _createKPosition_1);
          KRenderingFigureProvider.this._kRenderingExtensions.setLineWidth(rec, lineWidth);
        }
      };
      final KRoundedRectangle innerCircle = ObjectExtensions.<KRoundedRectangle>operator_doubleArrow(_createKRoundedRectangle_1, _function_3);
      EList<KRendering> _children = innerCircle.getChildren();
      _children.add(label);
      EList<KRendering> _children_1 = outerCircle.getChildren();
      _children_1.add(innerCircle);
    } else {
      EList<KRendering> _children_2 = outerCircle.getChildren();
      _children_2.add(label);
    }
    return outerCircle;
  }
  
  /**
   * Creates a rendering for a node that displays a value.
   * 
   * @param node the node to create the rendering information for.
   * @param value the value the node should display.
   * @return the rendering.
   */
  public KRendering createValueDisplayingNodeRendering(final KNode node, final String value) {
    KRendering _createDefaultRendering = this.createDefaultRendering(node);
    final KContainerRendering nodeRendering = ((KContainerRendering) _createDefaultRendering);
    EList<KRendering> _children = nodeRendering.getChildren();
    KText _createKText = this.renderingFactory.createKText();
    final Procedure1<KText> _function = new Procedure1<KText>() {
      public void apply(final KText text) {
        text.setText(value);
        KRenderingFigureProvider.this._kRenderingExtensions.<KText>setSurroundingSpace(text, 5, 0);
        KColor _lookupColor = GraphicsUtils.lookupColor("darkgrey");
        KRenderingFigureProvider.this._kRenderingExtensions.<KText>setForeground(text, _lookupColor);
      }
    };
    KText _doubleArrow = ObjectExtensions.<KText>operator_doubleArrow(_createKText, _function);
    _children.add(_doubleArrow);
    return nodeRendering;
  }
  
  /**
   * Creates a rendering for a node that represents a modal model port.
   * 
   * @param node the node to create the rendering information for.
   * @return the rendering.
   */
  public KRendering createModalModelPortRendering(final KNode node) {
    final boolean input = this._markerExtensions.isMarkedAsInputPort(node);
    final boolean output = this._markerExtensions.isMarkedAsOutputPort(node);
    final boolean multiport = this._annotationExtensions.hasAnnotation(node, "multiport");
    String _xifexpression = null;
    if (multiport) {
      _xifexpression = "multi";
    } else {
      _xifexpression = "";
    }
    String _plus = ("ren_" + _xifexpression);
    String _xifexpression_1 = null;
    if (input) {
      _xifexpression_1 = "in";
    } else {
      _xifexpression_1 = "";
    }
    String _plus_1 = (_plus + _xifexpression_1);
    String _xifexpression_2 = null;
    if (output) {
      _xifexpression_2 = "out";
    } else {
      _xifexpression_2 = "";
    }
    String _plus_2 = (_plus_1 + _xifexpression_2);
    final String id = (_plus_2 + "mmport");
    final KRenderingLibrary library = this.getLibrary(node);
    final KRenderingRef rendering = this.getFromLibrary(id, library);
    boolean _notEquals = (!Objects.equal(rendering, null));
    if (_notEquals) {
      return rendering;
    }
    final KPolygon polygon = this.renderingFactory.createKPolygon();
    if (multiport) {
      this._kRenderingExtensions.<KPolygon>setBackgroundColor(polygon, 255, 255, 255);
      boolean _and = false;
      if (!input) {
        _and = false;
      } else {
        _and = (!output);
      }
      if (_and) {
        EList<KPosition> _points = polygon.getPoints();
        KPosition _createKPosition = this._miscellaneousExtensions.createKPosition(0, 5);
        _points.add(_createKPosition);
        EList<KPosition> _points_1 = polygon.getPoints();
        KPosition _createKPosition_1 = this._miscellaneousExtensions.createKPosition(5, 5);
        _points_1.add(_createKPosition_1);
        EList<KPosition> _points_2 = polygon.getPoints();
        KPosition _createKPosition_2 = this._miscellaneousExtensions.createKPosition(5, 0);
        _points_2.add(_createKPosition_2);
        EList<KPosition> _points_3 = polygon.getPoints();
        KPosition _createKPosition_3 = this._miscellaneousExtensions.createKPosition(10, 5);
        _points_3.add(_createKPosition_3);
        EList<KPosition> _points_4 = polygon.getPoints();
        KPosition _createKPosition_4 = this._miscellaneousExtensions.createKPosition(10, 0);
        _points_4.add(_createKPosition_4);
        EList<KPosition> _points_5 = polygon.getPoints();
        KPosition _createKPosition_5 = this._miscellaneousExtensions.createKPosition(20, 10);
        _points_5.add(_createKPosition_5);
        EList<KPosition> _points_6 = polygon.getPoints();
        KPosition _createKPosition_6 = this._miscellaneousExtensions.createKPosition(10, 20);
        _points_6.add(_createKPosition_6);
        EList<KPosition> _points_7 = polygon.getPoints();
        KPosition _createKPosition_7 = this._miscellaneousExtensions.createKPosition(10, 15);
        _points_7.add(_createKPosition_7);
        EList<KPosition> _points_8 = polygon.getPoints();
        KPosition _createKPosition_8 = this._miscellaneousExtensions.createKPosition(5, 20);
        _points_8.add(_createKPosition_8);
        EList<KPosition> _points_9 = polygon.getPoints();
        KPosition _createKPosition_9 = this._miscellaneousExtensions.createKPosition(5, 15);
        _points_9.add(_createKPosition_9);
        EList<KPosition> _points_10 = polygon.getPoints();
        KPosition _createKPosition_10 = this._miscellaneousExtensions.createKPosition(0, 15);
        _points_10.add(_createKPosition_10);
        EList<KPosition> _points_11 = polygon.getPoints();
        KPosition _createKPosition_11 = this._miscellaneousExtensions.createKPosition(0, 5);
        _points_11.add(_createKPosition_11);
      } else {
        boolean _and_1 = false;
        if (!(!input)) {
          _and_1 = false;
        } else {
          _and_1 = output;
        }
        if (_and_1) {
          EList<KPosition> _points_12 = polygon.getPoints();
          KPosition _createKPosition_12 = this._miscellaneousExtensions.createKPosition(0, 0);
          _points_12.add(_createKPosition_12);
          EList<KPosition> _points_13 = polygon.getPoints();
          KPosition _createKPosition_13 = this._miscellaneousExtensions.createKPosition(5, 5);
          _points_13.add(_createKPosition_13);
          EList<KPosition> _points_14 = polygon.getPoints();
          KPosition _createKPosition_14 = this._miscellaneousExtensions.createKPosition(5, 0);
          _points_14.add(_createKPosition_14);
          EList<KPosition> _points_15 = polygon.getPoints();
          KPosition _createKPosition_15 = this._miscellaneousExtensions.createKPosition(10, 5);
          _points_15.add(_createKPosition_15);
          EList<KPosition> _points_16 = polygon.getPoints();
          KPosition _createKPosition_16 = this._miscellaneousExtensions.createKPosition(20, 5);
          _points_16.add(_createKPosition_16);
          EList<KPosition> _points_17 = polygon.getPoints();
          KPosition _createKPosition_17 = this._miscellaneousExtensions.createKPosition(20, 15);
          _points_17.add(_createKPosition_17);
          EList<KPosition> _points_18 = polygon.getPoints();
          KPosition _createKPosition_18 = this._miscellaneousExtensions.createKPosition(10, 15);
          _points_18.add(_createKPosition_18);
          EList<KPosition> _points_19 = polygon.getPoints();
          KPosition _createKPosition_19 = this._miscellaneousExtensions.createKPosition(5, 20);
          _points_19.add(_createKPosition_19);
          EList<KPosition> _points_20 = polygon.getPoints();
          KPosition _createKPosition_20 = this._miscellaneousExtensions.createKPosition(5, 15);
          _points_20.add(_createKPosition_20);
          EList<KPosition> _points_21 = polygon.getPoints();
          KPosition _createKPosition_21 = this._miscellaneousExtensions.createKPosition(0, 20);
          _points_21.add(_createKPosition_21);
          EList<KPosition> _points_22 = polygon.getPoints();
          KPosition _createKPosition_22 = this._miscellaneousExtensions.createKPosition(0, 0);
          _points_22.add(_createKPosition_22);
        } else {
          boolean _and_2 = false;
          if (!input) {
            _and_2 = false;
          } else {
            _and_2 = output;
          }
          if (_and_2) {
            EList<KPosition> _points_23 = polygon.getPoints();
            KPosition _createKPosition_23 = this._miscellaneousExtensions.createKPosition(0, 5);
            _points_23.add(_createKPosition_23);
            EList<KPosition> _points_24 = polygon.getPoints();
            KPosition _createKPosition_24 = this._miscellaneousExtensions.createKPosition(5, 5);
            _points_24.add(_createKPosition_24);
            EList<KPosition> _points_25 = polygon.getPoints();
            KPosition _createKPosition_25 = this._miscellaneousExtensions.createKPosition(5, 0);
            _points_25.add(_createKPosition_25);
            EList<KPosition> _points_26 = polygon.getPoints();
            KPosition _createKPosition_26 = this._miscellaneousExtensions.createKPosition(10, 5);
            _points_26.add(_createKPosition_26);
            EList<KPosition> _points_27 = polygon.getPoints();
            KPosition _createKPosition_27 = this._miscellaneousExtensions.createKPosition(10, 0);
            _points_27.add(_createKPosition_27);
            EList<KPosition> _points_28 = polygon.getPoints();
            KPosition _createKPosition_28 = this._miscellaneousExtensions.createKPosition(15, 5);
            _points_28.add(_createKPosition_28);
            EList<KPosition> _points_29 = polygon.getPoints();
            KPosition _createKPosition_29 = this._miscellaneousExtensions.createKPosition(20, 5);
            _points_29.add(_createKPosition_29);
            EList<KPosition> _points_30 = polygon.getPoints();
            KPosition _createKPosition_30 = this._miscellaneousExtensions.createKPosition(20, 15);
            _points_30.add(_createKPosition_30);
            EList<KPosition> _points_31 = polygon.getPoints();
            KPosition _createKPosition_31 = this._miscellaneousExtensions.createKPosition(15, 15);
            _points_31.add(_createKPosition_31);
            EList<KPosition> _points_32 = polygon.getPoints();
            KPosition _createKPosition_32 = this._miscellaneousExtensions.createKPosition(10, 20);
            _points_32.add(_createKPosition_32);
            EList<KPosition> _points_33 = polygon.getPoints();
            KPosition _createKPosition_33 = this._miscellaneousExtensions.createKPosition(10, 15);
            _points_33.add(_createKPosition_33);
            EList<KPosition> _points_34 = polygon.getPoints();
            KPosition _createKPosition_34 = this._miscellaneousExtensions.createKPosition(5, 20);
            _points_34.add(_createKPosition_34);
            EList<KPosition> _points_35 = polygon.getPoints();
            KPosition _createKPosition_35 = this._miscellaneousExtensions.createKPosition(5, 15);
            _points_35.add(_createKPosition_35);
            EList<KPosition> _points_36 = polygon.getPoints();
            KPosition _createKPosition_36 = this._miscellaneousExtensions.createKPosition(0, 15);
            _points_36.add(_createKPosition_36);
            EList<KPosition> _points_37 = polygon.getPoints();
            KPosition _createKPosition_37 = this._miscellaneousExtensions.createKPosition(0, 5);
            _points_37.add(_createKPosition_37);
          } else {
            EList<KPosition> _points_38 = polygon.getPoints();
            KPosition _createKPosition_38 = this._miscellaneousExtensions.createKPosition(0, 5);
            _points_38.add(_createKPosition_38);
            EList<KPosition> _points_39 = polygon.getPoints();
            KPosition _createKPosition_39 = this._miscellaneousExtensions.createKPosition(20, 5);
            _points_39.add(_createKPosition_39);
            EList<KPosition> _points_40 = polygon.getPoints();
            KPosition _createKPosition_40 = this._miscellaneousExtensions.createKPosition(20, 15);
            _points_40.add(_createKPosition_40);
            EList<KPosition> _points_41 = polygon.getPoints();
            KPosition _createKPosition_41 = this._miscellaneousExtensions.createKPosition(0, 15);
            _points_41.add(_createKPosition_41);
            EList<KPosition> _points_42 = polygon.getPoints();
            KPosition _createKPosition_42 = this._miscellaneousExtensions.createKPosition(0, 5);
            _points_42.add(_createKPosition_42);
          }
        }
      }
    } else {
      this._kRenderingExtensions.<KPolygon>setBackgroundColor(polygon, 0, 0, 0);
      boolean _and_3 = false;
      if (!input) {
        _and_3 = false;
      } else {
        _and_3 = (!output);
      }
      if (_and_3) {
        EList<KPosition> _points_43 = polygon.getPoints();
        KPosition _createKPosition_43 = this._miscellaneousExtensions.createKPosition(10, 0);
        _points_43.add(_createKPosition_43);
        EList<KPosition> _points_44 = polygon.getPoints();
        KPosition _createKPosition_44 = this._miscellaneousExtensions.createKPosition(20, 10);
        _points_44.add(_createKPosition_44);
        EList<KPosition> _points_45 = polygon.getPoints();
        KPosition _createKPosition_45 = this._miscellaneousExtensions.createKPosition(10, 20);
        _points_45.add(_createKPosition_45);
        EList<KPosition> _points_46 = polygon.getPoints();
        KPosition _createKPosition_46 = this._miscellaneousExtensions.createKPosition(10, 15);
        _points_46.add(_createKPosition_46);
        EList<KPosition> _points_47 = polygon.getPoints();
        KPosition _createKPosition_47 = this._miscellaneousExtensions.createKPosition(0, 15);
        _points_47.add(_createKPosition_47);
        EList<KPosition> _points_48 = polygon.getPoints();
        KPosition _createKPosition_48 = this._miscellaneousExtensions.createKPosition(0, 5);
        _points_48.add(_createKPosition_48);
        EList<KPosition> _points_49 = polygon.getPoints();
        KPosition _createKPosition_49 = this._miscellaneousExtensions.createKPosition(10, 5);
        _points_49.add(_createKPosition_49);
        EList<KPosition> _points_50 = polygon.getPoints();
        KPosition _createKPosition_50 = this._miscellaneousExtensions.createKPosition(10, 0);
        _points_50.add(_createKPosition_50);
      } else {
        boolean _and_4 = false;
        if (!(!input)) {
          _and_4 = false;
        } else {
          _and_4 = output;
        }
        if (_and_4) {
          EList<KPosition> _points_51 = polygon.getPoints();
          KPosition _createKPosition_51 = this._miscellaneousExtensions.createKPosition(0, 0);
          _points_51.add(_createKPosition_51);
          EList<KPosition> _points_52 = polygon.getPoints();
          KPosition _createKPosition_52 = this._miscellaneousExtensions.createKPosition(5, 5);
          _points_52.add(_createKPosition_52);
          EList<KPosition> _points_53 = polygon.getPoints();
          KPosition _createKPosition_53 = this._miscellaneousExtensions.createKPosition(20, 5);
          _points_53.add(_createKPosition_53);
          EList<KPosition> _points_54 = polygon.getPoints();
          KPosition _createKPosition_54 = this._miscellaneousExtensions.createKPosition(20, 15);
          _points_54.add(_createKPosition_54);
          EList<KPosition> _points_55 = polygon.getPoints();
          KPosition _createKPosition_55 = this._miscellaneousExtensions.createKPosition(5, 15);
          _points_55.add(_createKPosition_55);
          EList<KPosition> _points_56 = polygon.getPoints();
          KPosition _createKPosition_56 = this._miscellaneousExtensions.createKPosition(0, 20);
          _points_56.add(_createKPosition_56);
          EList<KPosition> _points_57 = polygon.getPoints();
          KPosition _createKPosition_57 = this._miscellaneousExtensions.createKPosition(0, 0);
          _points_57.add(_createKPosition_57);
        } else {
          boolean _and_5 = false;
          if (!input) {
            _and_5 = false;
          } else {
            _and_5 = output;
          }
          if (_and_5) {
            EList<KPosition> _points_58 = polygon.getPoints();
            KPosition _createKPosition_58 = this._miscellaneousExtensions.createKPosition(0, 5);
            _points_58.add(_createKPosition_58);
            EList<KPosition> _points_59 = polygon.getPoints();
            KPosition _createKPosition_59 = this._miscellaneousExtensions.createKPosition(7, 5);
            _points_59.add(_createKPosition_59);
            EList<KPosition> _points_60 = polygon.getPoints();
            KPosition _createKPosition_60 = this._miscellaneousExtensions.createKPosition(7, 0);
            _points_60.add(_createKPosition_60);
            EList<KPosition> _points_61 = polygon.getPoints();
            KPosition _createKPosition_61 = this._miscellaneousExtensions.createKPosition(12, 5);
            _points_61.add(_createKPosition_61);
            EList<KPosition> _points_62 = polygon.getPoints();
            KPosition _createKPosition_62 = this._miscellaneousExtensions.createKPosition(20, 5);
            _points_62.add(_createKPosition_62);
            EList<KPosition> _points_63 = polygon.getPoints();
            KPosition _createKPosition_63 = this._miscellaneousExtensions.createKPosition(20, 15);
            _points_63.add(_createKPosition_63);
            EList<KPosition> _points_64 = polygon.getPoints();
            KPosition _createKPosition_64 = this._miscellaneousExtensions.createKPosition(12, 15);
            _points_64.add(_createKPosition_64);
            EList<KPosition> _points_65 = polygon.getPoints();
            KPosition _createKPosition_65 = this._miscellaneousExtensions.createKPosition(7, 20);
            _points_65.add(_createKPosition_65);
            EList<KPosition> _points_66 = polygon.getPoints();
            KPosition _createKPosition_66 = this._miscellaneousExtensions.createKPosition(7, 15);
            _points_66.add(_createKPosition_66);
            EList<KPosition> _points_67 = polygon.getPoints();
            KPosition _createKPosition_67 = this._miscellaneousExtensions.createKPosition(0, 15);
            _points_67.add(_createKPosition_67);
            EList<KPosition> _points_68 = polygon.getPoints();
            KPosition _createKPosition_68 = this._miscellaneousExtensions.createKPosition(0, 5);
            _points_68.add(_createKPosition_68);
          } else {
            EList<KPosition> _points_69 = polygon.getPoints();
            KPosition _createKPosition_69 = this._miscellaneousExtensions.createKPosition(0, 5);
            _points_69.add(_createKPosition_69);
            EList<KPosition> _points_70 = polygon.getPoints();
            KPosition _createKPosition_70 = this._miscellaneousExtensions.createKPosition(20, 5);
            _points_70.add(_createKPosition_70);
            EList<KPosition> _points_71 = polygon.getPoints();
            KPosition _createKPosition_71 = this._miscellaneousExtensions.createKPosition(20, 15);
            _points_71.add(_createKPosition_71);
            EList<KPosition> _points_72 = polygon.getPoints();
            KPosition _createKPosition_72 = this._miscellaneousExtensions.createKPosition(0, 15);
            _points_72.add(_createKPosition_72);
            EList<KPosition> _points_73 = polygon.getPoints();
            KPosition _createKPosition_73 = this._miscellaneousExtensions.createKPosition(0, 5);
            _points_73.add(_createKPosition_73);
          }
        }
      }
    }
    return this.addToLibrary(polygon, id, library);
  }
  
  /**
   * Creates a rendering for a regular node.
   * 
   * @param node the node to create the rendering information for.
   * @return the rendering.
   */
  public KRendering createRegularNodeRendering(final KNode node) {
    String _annotationValue = this._annotationExtensions.getAnnotationValue(node, TransformationConstants.ANNOTATION_PTOLEMY_CLASS);
    String _replace = _annotationValue.replace(".", "");
    final String id = ("ren_" + _replace);
    final KRenderingLibrary library = this.getLibrary(node);
    final KRenderingRef rendering = this.getFromLibrary(id, library);
    boolean _notEquals = (!Objects.equal(rendering, null));
    if (_notEquals) {
      return rendering;
    }
    String _annotationValue_1 = this._annotationExtensions.getAnnotationValue(node, TransformationConstants.ANNOTATION_PTOLEMY_CLASS);
    final KRendering ptRendering = this._ptolemyFigureInterface.createPtolemyFigureRendering(_annotationValue_1);
    boolean _notEquals_1 = (!Objects.equal(ptRendering, null));
    if (_notEquals_1) {
      return this.addToLibrary(ptRendering, id, library);
    }
    return this.createDefaultRendering(node);
  }
  
  /**
   * Creates a rendering for an accumulator node. This needs to be a separate case because the SVG
   * description for accumulator nodes in Ptolemy is broken.
   * 
   * @param node the node to create the rendering information for.
   * @return the rendering.
   */
  public KRendering createAccumulatorNodeRendering(final KNode node) {
    final KRenderingLibrary library = this.getLibrary(node);
    final KRenderingRef rendering = this.getFromLibrary("ren_accumulator", library);
    boolean _notEquals = (!Objects.equal(rendering, null));
    if (_notEquals) {
      return rendering;
    }
    final String accumulatorSvg = (((((("<svg height=\"41\" width=\"41\" >" + "<rect height=\"40\" style=\"fill:white;stroke:black;stroke-width:1\" ") + "width=\"40\" x=\"0.0\" y=\"0.0\" />") + "<path d=\"m 29.126953,6.2304687 0,3.0410157 -13.833984,0 8.789062,9.3515626 ") + "-8.789062,10.335937 14.554687,0 0,3.041016 -18.246093,0 ") + "0,-3.550781 8.419921,-9.826172 -8.419921,-8.9648439 0,-3.4277344 z\" />") + "</svg>");
    KRendering _createFigureFromSvg = GraphicsUtils.createFigureFromSvg(accumulatorSvg);
    return this.addToLibrary(_createFigureFromSvg, 
      "ren_accumulator", library);
  }
  
  /**
   * Creates a rendering for an edge that represents a transition in a state machine.
   */
  public KRendering createTransitionRendering(final KEdge edge) {
    KSpline _createKSpline = this.renderingFactory.createKSpline();
    final Procedure1<KSpline> _function = new Procedure1<KSpline>() {
      public void apply(final KSpline spline) {
        KRenderingFigureProvider.this._kRenderingExtensions.setLineWidth(spline, 1.6f);
        boolean _annotationBooleanValue = KRenderingFigureProvider.this._annotationExtensions.getAnnotationBooleanValue(edge, TransformationConstants.ANNOTATION_NONDETERMINISTIC_TRANSITION);
        if (_annotationBooleanValue) {
          KColor _lookupColor = GraphicsUtils.lookupColor("red");
          KRenderingFigureProvider.this._kRenderingExtensions.setForeground(spline, _lookupColor);
        }
        boolean _annotationBooleanValue_1 = KRenderingFigureProvider.this._annotationExtensions.getAnnotationBooleanValue(edge, TransformationConstants.ANNOTATION_DEFAULT_TRANSITION);
        if (_annotationBooleanValue_1) {
          KRenderingFigureProvider.this._kRenderingExtensions.setLineStyle(spline, LineStyle.DASH);
        }
        KRendering _addHeadArrowDecorator = KRenderingFigureProvider.this._kPolylineExtensions.addHeadArrowDecorator(spline);
        final Procedure1<KRendering> _function = new Procedure1<KRendering>() {
          public void apply(final KRendering it) {
            boolean _annotationBooleanValue = KRenderingFigureProvider.this._annotationExtensions.getAnnotationBooleanValue(edge, TransformationConstants.ANNOTATION_RESET_TRANSITION);
            boolean _not = (!_annotationBooleanValue);
            if (_not) {
              KPlacementData _placementData = it.getPlacementData();
              ((KDecoratorPlacementData) _placementData).setAbsolute((-12));
              KRenderingFigureProvider.this.addHistoryDecorator(spline);
            }
          }
        };
        ObjectExtensions.<KRendering>operator_doubleArrow(_addHeadArrowDecorator, _function);
      }
    };
    return ObjectExtensions.<KSpline>operator_doubleArrow(_createKSpline, _function);
  }
  
  private KRendering addHistoryDecorator(final KPolyline line) {
    KEllipse _addEllipse = this._kContainerRenderingExtensions.addEllipse(line);
    final Procedure1<KEllipse> _function = new Procedure1<KEllipse>() {
      public void apply(final KEllipse it) {
        KRenderingFigureProvider.this._kRenderingExtensions.setLineWidth(it, 0.5f);
        KColor _color = KRenderingFigureProvider.this._kColorExtensions.getColor("gray");
        KRenderingFigureProvider.this._kRenderingExtensions.setBackground(it, _color);
        KRenderingFigureProvider.this._kRenderingExtensions.<KEllipse>setDecoratorPlacementData(it, 12, 12, (-3), 1, false);
        KPolyline _addPolyline = KRenderingFigureProvider.this._kContainerRenderingExtensions.addPolyline(it, 1);
        final Procedure1<KPolyline> _function = new Procedure1<KPolyline>() {
          public void apply(final KPolyline it) {
            EList<KPosition> _points = it.getPoints();
            KPosition _createKPosition = KRenderingFigureProvider.this._kRenderingExtensions.createKPosition(KRenderingFigureProvider.this._kRenderingExtensions.LEFT, 3.5f, 0, KRenderingFigureProvider.this._kRenderingExtensions.TOP, 2.5f, 0);
            _points.add(_createKPosition);
            EList<KPosition> _points_1 = it.getPoints();
            KPosition _createKPosition_1 = KRenderingFigureProvider.this._kRenderingExtensions.createKPosition(KRenderingFigureProvider.this._kRenderingExtensions.LEFT, 3.5f, 0, KRenderingFigureProvider.this._kRenderingExtensions.BOTTOM, 2.5f, 0);
            _points_1.add(_createKPosition_1);
            EList<KPosition> _points_2 = it.getPoints();
            KPosition _createKPosition_2 = KRenderingFigureProvider.this._kRenderingExtensions.createKPosition(KRenderingFigureProvider.this._kRenderingExtensions.LEFT, 3.5f, 0, KRenderingFigureProvider.this._kRenderingExtensions.TOP, 0, 0.5f);
            _points_2.add(_createKPosition_2);
            EList<KPosition> _points_3 = it.getPoints();
            KPosition _createKPosition_3 = KRenderingFigureProvider.this._kRenderingExtensions.createKPosition(KRenderingFigureProvider.this._kRenderingExtensions.RIGHT, 3.5f, 0, KRenderingFigureProvider.this._kRenderingExtensions.TOP, 0, 0.5f);
            _points_3.add(_createKPosition_3);
            EList<KPosition> _points_4 = it.getPoints();
            KPosition _createKPosition_4 = KRenderingFigureProvider.this._kRenderingExtensions.createKPosition(KRenderingFigureProvider.this._kRenderingExtensions.RIGHT, 3.5f, 0, KRenderingFigureProvider.this._kRenderingExtensions.BOTTOM, 2.5f, 0);
            _points_4.add(_createKPosition_4);
            EList<KPosition> _points_5 = it.getPoints();
            KPosition _createKPosition_5 = KRenderingFigureProvider.this._kRenderingExtensions.createKPosition(KRenderingFigureProvider.this._kRenderingExtensions.RIGHT, 3.5f, 0, KRenderingFigureProvider.this._kRenderingExtensions.TOP, 2.5f, 0);
            _points_5.add(_createKPosition_5);
          }
        };
        ObjectExtensions.<KPolyline>operator_doubleArrow(_addPolyline, _function);
      }
    };
    return ObjectExtensions.<KEllipse>operator_doubleArrow(_addEllipse, _function);
  }
  
  /**
   * Creates a rendering for an edge that represents a data flow buffer.
   */
  public KRendering createDataFlowRendering(final KEdge edge) {
    KNode _source = edge.getSource();
    final KRenderingLibrary library = this.getLibrary(_source);
    KRenderingRef junction = this.getFromLibrary("ren_junction", library);
    boolean _equals = Objects.equal(junction, null);
    if (_equals) {
      KPolygon _createKPolygon = this.renderingFactory.createKPolygon();
      final Procedure1<KPolygon> _function = new Procedure1<KPolygon>() {
        public void apply(final KPolygon poly) {
          EList<KPosition> _points = poly.getPoints();
          KPosition _createKPosition = KRenderingFigureProvider.this._miscellaneousExtensions.createKPosition(4, 0);
          _points.add(_createKPosition);
          EList<KPosition> _points_1 = poly.getPoints();
          KPosition _createKPosition_1 = KRenderingFigureProvider.this._miscellaneousExtensions.createKPosition(8, 4);
          _points_1.add(_createKPosition_1);
          EList<KPosition> _points_2 = poly.getPoints();
          KPosition _createKPosition_2 = KRenderingFigureProvider.this._miscellaneousExtensions.createKPosition(4, 8);
          _points_2.add(_createKPosition_2);
          EList<KPosition> _points_3 = poly.getPoints();
          KPosition _createKPosition_3 = KRenderingFigureProvider.this._miscellaneousExtensions.createKPosition(0, 4);
          _points_3.add(_createKPosition_3);
          EList<KPosition> _points_4 = poly.getPoints();
          KPosition _createKPosition_4 = KRenderingFigureProvider.this._miscellaneousExtensions.createKPosition(4, 0);
          _points_4.add(_createKPosition_4);
          KRenderingFigureProvider.this._kRenderingExtensions.<KPolygon>setBackgroundColor(poly, 0, 0, 0);
          KPointPlacementData _createKPointPlacementData = KRenderingFigureProvider.this.renderingFactory.createKPointPlacementData();
          final Procedure1<KPointPlacementData> _function = new Procedure1<KPointPlacementData>() {
            public void apply(final KPointPlacementData ppd) {
              ppd.setHorizontalAlignment(HorizontalAlignment.CENTER);
              ppd.setVerticalAlignment(VerticalAlignment.CENTER);
              ppd.setMinWidth(8);
              ppd.setMinHeight(8);
            }
          };
          KPointPlacementData _doubleArrow = ObjectExtensions.<KPointPlacementData>operator_doubleArrow(_createKPointPlacementData, _function);
          poly.setPlacementData(_doubleArrow);
        }
      };
      KPolygon _doubleArrow = ObjectExtensions.<KPolygon>operator_doubleArrow(_createKPolygon, _function);
      KRenderingRef _addToLibrary = this.addToLibrary(_doubleArrow, "ren_junction", library);
      junction = _addToLibrary;
    }
    KRoundedBendsPolyline _createKRoundedBendsPolyline = this.renderingFactory.createKRoundedBendsPolyline();
    final Procedure1<KRoundedBendsPolyline> _function_1 = new Procedure1<KRoundedBendsPolyline>() {
      public void apply(final KRoundedBendsPolyline polyLine) {
        polyLine.setBendRadius(5f);
        KRenderingFigureProvider.this._kRenderingExtensions.setLineWidth(polyLine, 2f);
      }
    };
    final KRoundedBendsPolyline rendering = ObjectExtensions.<KRoundedBendsPolyline>operator_doubleArrow(_createKRoundedBendsPolyline, _function_1);
    rendering.setJunctionPointRendering(junction);
    return rendering;
  }
  
  /**
   * Creates a rendering for an edge that attaches a comment node to a commented node.
   * 
   * @param edge the edge to create the rendering information for.
   * @return the rendering.
   */
  public KRendering createCommentEdgeRendering(final KEdge edge) {
    KPolyline _createKPolyline = this.renderingFactory.createKPolyline();
    final Procedure1<KPolyline> _function = new Procedure1<KPolyline>() {
      public void apply(final KPolyline line) {
        KRenderingFigureProvider.this._kRenderingExtensions.setLineStyle(line, LineStyle.DASH);
        KRenderingFigureProvider.this._kRenderingExtensions.setLineWidth(line, 1);
        KColor _lookupColor = GraphicsUtils.lookupColor("grey");
        KRenderingFigureProvider.this._kRenderingExtensions.setForeground(line, _lookupColor);
      }
    };
    final KPolyline polyline = ObjectExtensions.<KPolyline>operator_doubleArrow(_createKPolyline, _function);
    return polyline;
  }
  
  /**
   * Width / height of a port.
   */
  private final float PORT_SIZE = 7f;
  
  /**
   * Half the width / height of a port.
   */
  private final float PORT_SIZE_HALF = 3.5f;
  
  /**
   * The id of the modifier that rotates the port renderings.
   */
  private final String PORT_STYLE_MODIFIER_ID = "de.cau.cs.kieler.ptolemy.klighd.ptolemyPortStyleModifier";
  
  /**
   * Creates a rendering for a port.
   * 
   * @param port the port to create the rendering information for.
   * @return the rendering.
   */
  public KRendering createPortRendering(final KPort port) {
    KColor _xifexpression = null;
    boolean _hasAnnotation = this._annotationExtensions.hasAnnotation(port, TransformationConstants.IS_PARAMETER_PORT);
    if (_hasAnnotation) {
      _xifexpression = GraphicsUtils.lookupColor("gray");
    } else {
      KColor _xifexpression_1 = null;
      boolean _and = false;
      boolean _hasAnnotation_1 = this._annotationExtensions.hasAnnotation(port, TransformationConstants.IS_IO_PORT);
      if (!_hasAnnotation_1) {
        _and = false;
      } else {
        boolean _hasAnnotation_2 = this._annotationExtensions.hasAnnotation(port, TransformationConstants.IS_MULTIPORT);
        _and = _hasAnnotation_2;
      }
      if (_and) {
        _xifexpression_1 = GraphicsUtils.lookupColor("white");
      } else {
        _xifexpression_1 = GraphicsUtils.lookupColor("black");
      }
      _xifexpression = _xifexpression_1;
    }
    final KColor portFillColor = _xifexpression;
    final KPolygon polygon = this._kRenderingExtensions.addPolygon(port);
    EList<KPosition> _points = polygon.getPoints();
    KPosition _createKPosition = this._kRenderingExtensions.createKPosition(
      this._kRenderingExtensions.LEFT, 0, 0, this._kRenderingExtensions.TOP, this.PORT_SIZE, 0);
    _points.add(_createKPosition);
    EList<KPosition> _points_1 = polygon.getPoints();
    KPosition _createKPosition_1 = this._kRenderingExtensions.createKPosition(
      this._kRenderingExtensions.LEFT, 0, 0, this._kRenderingExtensions.TOP, 0, 0);
    _points_1.add(_createKPosition_1);
    EList<KPosition> _points_2 = polygon.getPoints();
    KPosition _createKPosition_2 = this._kRenderingExtensions.createKPosition(
      this._kRenderingExtensions.LEFT, this.PORT_SIZE, 0, this._kRenderingExtensions.TOP, this.PORT_SIZE_HALF, 0);
    _points_2.add(_createKPosition_2);
    EList<KPosition> _points_3 = polygon.getPoints();
    KPosition _createKPosition_3 = this._kRenderingExtensions.createKPosition(
      this._kRenderingExtensions.LEFT, 0, 0, this._kRenderingExtensions.TOP, this.PORT_SIZE, 0);
    _points_3.add(_createKPosition_3);
    this._kRenderingExtensions.setRotation(polygon, Float.valueOf(0f));
    KRotation _rotation = this._kRenderingExtensions.getRotation(polygon);
    KPosition _createKPosition_4 = this._miscellaneousExtensions.createKPosition(this.PORT_SIZE_HALF, this.PORT_SIZE_HALF);
    _rotation.setRotationAnchor(_createKPosition_4);
    KRotation _rotation_1 = this._kRenderingExtensions.getRotation(polygon);
    _rotation_1.setModifierId(this.PORT_STYLE_MODIFIER_ID);
    this._kRenderingExtensions.<KPolygon>setBackground(polygon, portFillColor);
    KColor _lookupColor = GraphicsUtils.lookupColor("black");
    this._kRenderingExtensions.<KPolygon>setForeground(polygon, _lookupColor);
    return polygon;
  }
}
