/**
 * KIELER - Kiel Integrated Environment for Layout Eclipse RichClient
 * 
 * http://www.informatik.uni-kiel.de/rtsys/kieler/
 * 
 * Copyright 2012 by
 * + Christian-Albrechts-University of Kiel
 *   + Department of Computer Science
 *     + Real-Time and Embedded Systems Group
 * 
 * This code is provided under the terms of the Eclipse Public License (EPL).
 * See the file epl-v10.html for the license text.
 */
package de.cau.cs.kieler.ptolemy.klighd.transformation;

import com.google.common.base.Objects;
import com.google.common.collect.Iterables;
import com.google.inject.Inject;
import de.cau.cs.kieler.core.kgraph.KEdge;
import de.cau.cs.kieler.core.kgraph.KNode;
import de.cau.cs.kieler.core.kgraph.KPort;
import de.cau.cs.kieler.core.util.Pair;
import de.cau.cs.kieler.kiml.klayoutdata.KLayoutData;
import de.cau.cs.kieler.kiml.util.KimlUtil;
import de.cau.cs.kieler.klighd.syntheses.AbstractDiagramSynthesis;
import de.cau.cs.kieler.ptolemy.klighd.PtolemyProperties;
import de.cau.cs.kieler.ptolemy.klighd.transformation.CommentsExtractor;
import de.cau.cs.kieler.ptolemy.klighd.transformation.Flattener;
import de.cau.cs.kieler.ptolemy.klighd.transformation.extensions.AnnotationExtensions;
import de.cau.cs.kieler.ptolemy.klighd.transformation.extensions.LabelExtensions;
import de.cau.cs.kieler.ptolemy.klighd.transformation.extensions.MarkerExtensions;
import de.cau.cs.kieler.ptolemy.klighd.transformation.extensions.MiscellaneousExtensions;
import de.cau.cs.kieler.ptolemy.klighd.transformation.extensions.PortExtensions;
import de.cau.cs.kieler.ptolemy.klighd.transformation.util.HyperedgeGatherer;
import de.cau.cs.kieler.ptolemy.klighd.transformation.util.TransformationConstants;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import org.eclipse.emf.common.util.EList;
import org.eclipse.xtext.xbase.lib.CollectionLiterals;
import org.eclipse.xtext.xbase.lib.Extension;
import org.eclipse.xtext.xbase.lib.Functions.Function1;
import org.eclipse.xtext.xbase.lib.IterableExtensions;
import org.eclipse.xtext.xbase.lib.StringExtensions;
import org.ptolemy.moml.PropertyType;

/**
 * Optimizes a KGraph model freshly transformed from a Ptolemy2 model. This is step two of the Ptolemy
 * model import process.
 * 
 * <p>In this part, edge directions are computed, relations that only connect two ports are replaced
 * by a single edge, special annotations are replaced by nodes (the Ptolemy director, ...), and
 * edges that connect states are changed.</p>
 * 
 * @author cds
 * @author haf
 * @kieler.rating yellow 2012-06-14 KI-12 cmot, grh
 */
@SuppressWarnings("all")
public class Ptolemy2KGraphOptimization {
  /**
   * Handling annotations.
   */
  @Inject
  @Extension
  private AnnotationExtensions _annotationExtensions;
  
  /**
   * Handling labels.
   */
  @Inject
  @Extension
  private LabelExtensions _labelExtensions;
  
  /**
   * Marking nodes.
   */
  @Inject
  @Extension
  private MarkerExtensions _markerExtensions;
  
  /**
   * Miscellaneous stuff to make my life easier.
   */
  @Inject
  @Extension
  private MiscellaneousExtensions _miscellaneousExtensions;
  
  /**
   * Handling ports.
   */
  @Inject
  @Extension
  private PortExtensions _portExtensions;
  
  /**
   * Flattening.
   */
  @Inject
  private Flattener flattener;
  
  /**
   * Optimizes the given KGraph model.
   * 
   * <p>The order in which we do that is partly important. We can remove a relation if it has only
   * one incoming and one outgoing edge. In order to determine this, we need to have inferred the
   * edge directions. When we remove ports from states or convert annotations to nodes is less
   * important.</p>
   * 
   * @param kGraph the model to optimize.
   * @param hideRelations {@code true} if relations should be replaced by hyperedges; otherwise, we
   *                      only try and hide as many relations as possible.
   * @param flatten {@code true} if the graph should be flattened by removing all hierarchy.
   * @param commentsExtractor the extractor to use for extracting comments, or {@code null} if no
   *                          comments should be extracted.
   * @param diagramSynthesis the diagram synthesis that uses this classes. Used to map Ptolemy model
   *                         objects to KGraph model objects.
   */
  public void optimize(final KNode kGraph, final boolean hideRelations, final boolean flatten, final CommentsExtractor commentsExtractor, final AbstractDiagramSynthesis<?> diagramSynthesis) {
    this.inferEdgeDirections(kGraph);
    if (hideRelations) {
      this.removeAllRelations(kGraph);
    } else {
      this.removeUnnecessaryRelations(kGraph);
    }
    this.makeStatesPortless(kGraph);
    this.convertAnnotationsToNodes(kGraph);
    boolean _notEquals = (!Objects.equal(commentsExtractor, null));
    if (_notEquals) {
      commentsExtractor.extractComments(kGraph, diagramSynthesis);
    }
    if (flatten) {
      this.flattener.flatten(kGraph);
    }
  }
  
  /**
   * Infers the direction of the edges in the model tree rooted at the given node.
   * 
   * <p>First, a list of all ports of known and unknown type, all edge of unknown direction, and all
   * relations are collected. In that process, the direction of edges connected to ports of known type
   * is inferred.</p>
   * 
   * <p>Second, an attempt is made to infer the type of ports of yet unknown type. This succeeds if
   * the port is connected to an edge whose direction is known. The direction of all edges connected
   * to such a port is inferred in that process.</p>
   * 
   * <p>Third, all relations are traversed, looking for a relation with only incoming or only outgoing
   * edges and just one edge of unknown direction. If such a relation is found, the direction of that
   * one edge is inferred. If no such relation is found, an undirected edge's direction is randomly
   * fixed.</p>
   * 
   * <p>Repeat at step two.</p>
   * 
   * <p>By this time, either all edge directions have been inferred or there are some edges left whose
   * direction cannot be determined with the information available. These edges are left untouched,
   * since we might as well leave them in the direction they currently have.</p>
   * 
   * @param kGraph root of the graph.
   */
  private void inferEdgeDirections(final KNode kGraph) {
    final List<KPort> knownPorts = CollectionLiterals.<KPort>newArrayList();
    final List<KPort> unknownPorts = CollectionLiterals.<KPort>newArrayList();
    final List<KEdge> unknownEdges = CollectionLiterals.<KEdge>newArrayList();
    final List<KNode> unknownRelations = CollectionLiterals.<KNode>newArrayList();
    this.gatherModelElements(kGraph, knownPorts, unknownPorts, unknownEdges, unknownRelations);
    for (final KPort knownPort : knownPorts) {
      this.propagatePortTypeToIncidentEdges(knownPort, unknownEdges);
    }
    boolean portTypesChanged = false;
    boolean relationsChanged = false;
    boolean randomEdgeFixed = false;
    do {
      {
        boolean _inferPortTypes = this.inferPortTypes(unknownPorts, unknownEdges);
        portTypesChanged = _inferPortTypes;
        boolean _traverseRelations = this.traverseRelations(unknownRelations, unknownEdges, true);
        relationsChanged = _traverseRelations;
        if ((!(portTypesChanged || relationsChanged))) {
          boolean _traverseRelations_1 = this.traverseRelations(unknownRelations, unknownEdges, false);
          randomEdgeFixed = _traverseRelations_1;
        }
      }
    } while(((portTypesChanged || relationsChanged) || randomEdgeFixed));
  }
  
  /**
   * Traverses the model, filling the given lists with interesting elements: ports of known and
   * unknown type, edges of unknown direction, and relations with incident edges of unknown
   * direction. The lists can then be iterated over instead of always having to iterate over the
   * whole model again and again.
   * 
   * @param root the root of the model tree.
   * @param knownPorts list to which ports of known type are added.
   * @param unknownPorts list to which ports of unknown type are added.
   * @param unknownLinks list to which edges of unknown direction are added.
   * @param unknownRelations list to which relations with incident edges of unknown direction
   *                         are added.
   */
  private void gatherModelElements(final KNode root, final List<KPort> knownPorts, final List<KPort> unknownPorts, final List<KEdge> unknownEdges, final List<KNode> unknownRelations) {
    boolean _isMarkedAsHypernode = this._markerExtensions.isMarkedAsHypernode(root);
    if (_isMarkedAsHypernode) {
      unknownRelations.add(root);
    }
    EList<KPort> _ports = root.getPorts();
    for (final KPort port : _ports) {
      boolean _or = false;
      boolean _isMarkedAsInputPort = this._markerExtensions.isMarkedAsInputPort(port);
      if (_isMarkedAsInputPort) {
        _or = true;
      } else {
        boolean _isMarkedAsOutputPort = this._markerExtensions.isMarkedAsOutputPort(port);
        _or = _isMarkedAsOutputPort;
      }
      if (_or) {
        knownPorts.add(port);
      } else {
        unknownPorts.add(port);
      }
    }
    EList<KEdge> _outgoingEdges = root.getOutgoingEdges();
    for (final KEdge edge : _outgoingEdges) {
      boolean _isMarkedAsUndirected = this._markerExtensions.isMarkedAsUndirected(edge);
      if (_isMarkedAsUndirected) {
        unknownEdges.add(edge);
      }
    }
    EList<KNode> _children = root.getChildren();
    for (final KNode child : _children) {
      this.gatherModelElements(child, knownPorts, unknownPorts, unknownEdges, unknownRelations);
    }
  }
  
  /**
   * For a port of known type, sets the directions of its incident unknown edges accordingly. This
   * succeeds if the port is marked as being either an input port or an output port, not both.
   * 
   * <p>Note that for hierarchical ports, the port type is only valid for connections to its node's
   * environment. Children of the node must treat a hierarchical input port as being an output port,
   * and the other way round for hierarchical output ports.</p>
   * 
   * @param port the port.
   * @param unknownEdges list of edges with unknown direction. Edges whose direction is set are
   *                     removed from this list.
   * @return {@code true} if at least one edge's direction is fixed or reversed.
   */
  private boolean propagatePortTypeToIncidentEdges(final KPort port, final List<KEdge> unknownEdges) {
    final List<KEdge> edgesToBeReversed = CollectionLiterals.<KEdge>newArrayList();
    final List<KEdge> edgesToBeKept = CollectionLiterals.<KEdge>newArrayList();
    boolean result = false;
    boolean _and = false;
    boolean _isMarkedAsInputPort = this._markerExtensions.isMarkedAsInputPort(port);
    if (!_isMarkedAsInputPort) {
      _and = false;
    } else {
      boolean _isMarkedAsOutputPort = this._markerExtensions.isMarkedAsOutputPort(port);
      boolean _not = (!_isMarkedAsOutputPort);
      _and = _not;
    }
    if (_and) {
      EList<KEdge> _edges = port.getEdges();
      for (final KEdge edge : _edges) {
        KNode _node = port.getNode();
        EList<KNode> _children = _node.getChildren();
        KNode _source = edge.getSource();
        boolean _contains = _children.contains(_source);
        if (_contains) {
          edgesToBeReversed.add(edge);
        } else {
          boolean _and_1 = false;
          boolean _and_2 = false;
          KPort _sourcePort = edge.getSourcePort();
          boolean _equals = Objects.equal(_sourcePort, port);
          if (!_equals) {
            _and_2 = false;
          } else {
            KNode _node_1 = port.getNode();
            KNode _parent = _node_1.getParent();
            boolean _notEquals = (!Objects.equal(_parent, null));
            _and_2 = _notEquals;
          }
          if (!_and_2) {
            _and_1 = false;
          } else {
            KNode _node_2 = port.getNode();
            KNode _parent_1 = _node_2.getParent();
            EList<KNode> _children_1 = _parent_1.getChildren();
            KNode _target = edge.getTarget();
            boolean _contains_1 = _children_1.contains(_target);
            _and_1 = _contains_1;
          }
          if (_and_1) {
            edgesToBeReversed.add(edge);
          } else {
            edgesToBeKept.add(edge);
          }
        }
      }
    } else {
      boolean _and_3 = false;
      boolean _isMarkedAsOutputPort_1 = this._markerExtensions.isMarkedAsOutputPort(port);
      if (!_isMarkedAsOutputPort_1) {
        _and_3 = false;
      } else {
        boolean _isMarkedAsInputPort_1 = this._markerExtensions.isMarkedAsInputPort(port);
        boolean _not_1 = (!_isMarkedAsInputPort_1);
        _and_3 = _not_1;
      }
      if (_and_3) {
        EList<KEdge> _edges_1 = port.getEdges();
        for (final KEdge edge_1 : _edges_1) {
          KNode _node_3 = port.getNode();
          EList<KNode> _children_2 = _node_3.getChildren();
          KNode _target_1 = edge_1.getTarget();
          boolean _contains_2 = _children_2.contains(_target_1);
          if (_contains_2) {
            edgesToBeReversed.add(edge_1);
          } else {
            boolean _and_4 = false;
            boolean _and_5 = false;
            KPort _targetPort = edge_1.getTargetPort();
            boolean _equals_1 = Objects.equal(_targetPort, port);
            if (!_equals_1) {
              _and_5 = false;
            } else {
              KNode _node_4 = port.getNode();
              KNode _parent_2 = _node_4.getParent();
              boolean _notEquals_1 = (!Objects.equal(_parent_2, null));
              _and_5 = _notEquals_1;
            }
            if (!_and_5) {
              _and_4 = false;
            } else {
              KNode _node_5 = port.getNode();
              KNode _parent_3 = _node_5.getParent();
              EList<KNode> _children_3 = _parent_3.getChildren();
              KNode _source_1 = edge_1.getSource();
              boolean _contains_3 = _children_3.contains(_source_1);
              _and_4 = _contains_3;
            }
            if (_and_4) {
              edgesToBeReversed.add(edge_1);
            } else {
              edgesToBeKept.add(edge_1);
            }
          }
        }
      }
    }
    for (final KEdge edge_2 : edgesToBeReversed) {
      {
        this._miscellaneousExtensions.reverseEdge(edge_2);
        result = true;
        boolean _isMarkedAsUndirected = this._markerExtensions.isMarkedAsUndirected(edge_2);
        if (_isMarkedAsUndirected) {
          this._markerExtensions.markAsUndirected(edge_2, false);
          unknownEdges.remove(edge_2);
        }
      }
    }
    for (final KEdge edge_3 : edgesToBeKept) {
      boolean _isMarkedAsUndirected = this._markerExtensions.isMarkedAsUndirected(edge_3);
      if (_isMarkedAsUndirected) {
        this._markerExtensions.markAsUndirected(edge_3, false);
        unknownEdges.remove(edge_3);
        result = true;
      }
    }
    return result;
  }
  
  /**
   * Iterates over the given list of ports of unknown type and tries to infer the type of as many
   * ports as possible. For ports whose type is inferred, the incident edges are fixed accordingly.
   * 
   * @param unknownPorts list of unknown ports. Ports whose type is set are removed from this list.
   * @param unknownEdges list of edges with unknown direction. Edges whose direction is set are
   *                     removed from this list.
   * @return {@code true} if at least one edge's direction is fixed.
   */
  private boolean inferPortTypes(final List<KPort> unknownPorts, final List<KEdge> unknownEdges) {
    boolean result = false;
    final ListIterator<KPort> unknownPortsIterator = unknownPorts.listIterator();
    while (unknownPortsIterator.hasNext()) {
      {
        final KPort unknownPort = unknownPortsIterator.next();
        EList<KEdge> _edges = unknownPort.getEdges();
        final Function1<KEdge, Boolean> _function = new Function1<KEdge, Boolean>() {
          public Boolean apply(final KEdge e) {
            KPort _targetPort = e.getTargetPort();
            return Boolean.valueOf(Objects.equal(_targetPort, unknownPort));
          }
        };
        Iterable<KEdge> _filter = IterableExtensions.<KEdge>filter(_edges, _function);
        final KEdge directedIncomingEdge = this._miscellaneousExtensions.getFirstDirectedEdge(_filter);
        EList<KEdge> _edges_1 = unknownPort.getEdges();
        final Function1<KEdge, Boolean> _function_1 = new Function1<KEdge, Boolean>() {
          public Boolean apply(final KEdge e) {
            KPort _sourcePort = e.getSourcePort();
            return Boolean.valueOf(Objects.equal(_sourcePort, unknownPort));
          }
        };
        Iterable<KEdge> _filter_1 = IterableExtensions.<KEdge>filter(_edges_1, _function_1);
        final KEdge directedOutgoingEdge = this._miscellaneousExtensions.getFirstDirectedEdge(_filter_1);
        boolean _notEquals = (!Objects.equal(directedIncomingEdge, null));
        if (_notEquals) {
          KNode _node = unknownPort.getNode();
          EList<KNode> _children = _node.getChildren();
          KNode _source = directedIncomingEdge.getSource();
          boolean _contains = _children.contains(_source);
          if (_contains) {
            this._markerExtensions.markAsOutputPort(unknownPort, true);
          } else {
            this._markerExtensions.markAsInputPort(unknownPort, true);
          }
        } else {
          boolean _notEquals_1 = (!Objects.equal(directedOutgoingEdge, null));
          if (_notEquals_1) {
            KNode _node_1 = unknownPort.getNode();
            EList<KNode> _children_1 = _node_1.getChildren();
            KNode _target = directedOutgoingEdge.getTarget();
            boolean _contains_1 = _children_1.contains(_target);
            if (_contains_1) {
              this._markerExtensions.markAsInputPort(unknownPort, true);
            } else {
              this._markerExtensions.markAsOutputPort(unknownPort, true);
            }
          } else {
            String _name = this._labelExtensions.getName(unknownPort);
            boolean _isInputPortName = this._portExtensions.isInputPortName(_name);
            if (_isInputPortName) {
              this._markerExtensions.markAsInputPort(unknownPort, true);
            } else {
              String _name_1 = this._labelExtensions.getName(unknownPort);
              boolean _isOutputPortName = this._portExtensions.isOutputPortName(_name_1);
              if (_isOutputPortName) {
                this._markerExtensions.markAsOutputPort(unknownPort, true);
              }
            }
          }
        }
        boolean _or = false;
        boolean _isMarkedAsInputPort = this._markerExtensions.isMarkedAsInputPort(unknownPort);
        if (_isMarkedAsInputPort) {
          _or = true;
        } else {
          boolean _isMarkedAsOutputPort = this._markerExtensions.isMarkedAsOutputPort(unknownPort);
          _or = _isMarkedAsOutputPort;
        }
        if (_or) {
          boolean _or_1 = false;
          boolean _propagatePortTypeToIncidentEdges = this.propagatePortTypeToIncidentEdges(unknownPort, unknownEdges);
          if (_propagatePortTypeToIncidentEdges) {
            _or_1 = true;
          } else {
            _or_1 = result;
          }
          result = _or_1;
          unknownPortsIterator.remove();
        }
      }
    }
    return result;
  }
  
  /**
   * Iterates over the list of relations with unknown edges and tries to infer edge directions. This is
   * done by looking for relations with only incoming or only outgoing edges whose direction is known,
   * and edges whose direction is unknown. This method can operate in two different modes regarding
   * the number of edges with unknown direction.
   * 
   * <p>The first mode is conservative. For each undirected edge it only fixes its direction if it
   * is the only undirected edge incident to a relation. This is the safe mode of operation.</p>
   * 
   * <p>The second mode also accepts relations with more than one incident undirected edge. It takes
   * the first of them, sets its direction and returns, thereby only fixing the direction of at most
   * one edge in the whole model. This mode is used to fix an edge's direction if the other methods
   * cannot infer any more edge directions.</p>
   * 
   * @param unknownRelation list of relations with unknown incident edges. Relations whose incident
   *                        edges are all fixed are removed from this list.
   * @param unknownEdges list of edges with unknown direction. Edges whose direction is set are
   *                     removed from this list.
   * @param conservative if {@code false}, operates in a mode that fixes a single edge more or less
   *                     at random, if possible.
   * @return {@code true} if at least one edge's direction is fixed or reversed.
   */
  private boolean traverseRelations(final List<KNode> unknownRelations, final List<KEdge> unknownEdges, final boolean conservative) {
    boolean result = false;
    final ListIterator<KNode> unknownRelationsIterator = unknownRelations.listIterator();
    while (unknownRelationsIterator.hasNext()) {
      {
        final KNode unknownRelation = unknownRelationsIterator.next();
        boolean fixedEdgeInThisIteration = false;
        EList<KEdge> _incomingEdges = unknownRelation.getIncomingEdges();
        final Function1<KEdge, Boolean> _function = new Function1<KEdge, Boolean>() {
          public Boolean apply(final KEdge l) {
            boolean _isMarkedAsUndirected = Ptolemy2KGraphOptimization.this._markerExtensions.isMarkedAsUndirected(l);
            return Boolean.valueOf((!_isMarkedAsUndirected));
          }
        };
        final Iterable<KEdge> fixedIncomingEdges = IterableExtensions.<KEdge>filter(_incomingEdges, _function);
        EList<KEdge> _outgoingEdges = unknownRelation.getOutgoingEdges();
        final Function1<KEdge, Boolean> _function_1 = new Function1<KEdge, Boolean>() {
          public Boolean apply(final KEdge l) {
            boolean _isMarkedAsUndirected = Ptolemy2KGraphOptimization.this._markerExtensions.isMarkedAsUndirected(l);
            return Boolean.valueOf((!_isMarkedAsUndirected));
          }
        };
        final Iterable<KEdge> fixedOutgoingEdges = IterableExtensions.<KEdge>filter(_outgoingEdges, _function_1);
        List<KEdge> _incidentEdges = this._miscellaneousExtensions.getIncidentEdges(unknownRelation);
        final Function1<KEdge, Boolean> _function_2 = new Function1<KEdge, Boolean>() {
          public Boolean apply(final KEdge l) {
            return Boolean.valueOf(Ptolemy2KGraphOptimization.this._markerExtensions.isMarkedAsUndirected(l));
          }
        };
        final Iterable<KEdge> undirectedIncidentEdges = IterableExtensions.<KEdge>filter(_incidentEdges, _function_2);
        final int undirectedIncidentEdgesSize = IterableExtensions.size(undirectedIncidentEdges);
        if (((conservative && (undirectedIncidentEdgesSize == 1)) || ((!conservative) && (undirectedIncidentEdgesSize > 0)))) {
          Iterator<KEdge> _iterator = undirectedIncidentEdges.iterator();
          final KEdge undirectedEdge = _iterator.next();
          boolean _and = false;
          int _size = IterableExtensions.size(fixedIncomingEdges);
          boolean _greaterThan = (_size > 0);
          if (!_greaterThan) {
            _and = false;
          } else {
            int _size_1 = IterableExtensions.size(fixedOutgoingEdges);
            boolean _equals = (_size_1 == 0);
            _and = _equals;
          }
          if (_and) {
            KNode _source = undirectedEdge.getSource();
            boolean _notEquals = (!Objects.equal(_source, unknownRelation));
            if (_notEquals) {
              this._miscellaneousExtensions.reverseEdge(undirectedEdge);
            }
            this._markerExtensions.markAsUndirected(undirectedEdge, false);
            unknownEdges.remove(undirectedEdge);
            fixedEdgeInThisIteration = true;
          } else {
            boolean _and_1 = false;
            int _size_2 = IterableExtensions.size(fixedOutgoingEdges);
            boolean _greaterThan_1 = (_size_2 > 0);
            if (!_greaterThan_1) {
              _and_1 = false;
            } else {
              int _size_3 = IterableExtensions.size(fixedIncomingEdges);
              boolean _equals_1 = (_size_3 == 0);
              _and_1 = _equals_1;
            }
            if (_and_1) {
              KNode _target = undirectedEdge.getTarget();
              boolean _notEquals_1 = (!Objects.equal(_target, unknownRelation));
              if (_notEquals_1) {
                this._miscellaneousExtensions.reverseEdge(undirectedEdge);
              }
              this._markerExtensions.markAsUndirected(undirectedEdge, false);
              unknownEdges.remove(undirectedEdge);
              fixedEdgeInThisIteration = true;
            } else {
              int _size_4 = IterableExtensions.size(fixedIncomingEdges);
              boolean _greaterThan_2 = (_size_4 > 0);
              if (_greaterThan_2) {
                for (final KEdge edge : undirectedIncidentEdges) {
                  {
                    KNode _source_1 = edge.getSource();
                    boolean _notEquals_2 = (!Objects.equal(_source_1, unknownRelation));
                    if (_notEquals_2) {
                      this._miscellaneousExtensions.reverseEdge(edge);
                    }
                    this._markerExtensions.markAsUndirected(edge, false);
                    unknownEdges.remove(edge);
                    fixedEdgeInThisIteration = true;
                  }
                }
              }
            }
          }
        }
        if (fixedEdgeInThisIteration) {
          result = true;
          if ((undirectedIncidentEdgesSize == 1)) {
            unknownRelationsIterator.remove();
          }
          if ((!conservative)) {
            return result;
          }
        }
      }
    }
    return result;
  }
  
  /**
   * Recursively removes unnecessary relations in the model rooted at the given node. Unnecessary
   * relations are those that have one incoming and one outgoing edge.
   * 
   * @param root the model's root node.
   */
  private void removeUnnecessaryRelations(final KNode root) {
    EList<KNode> _children = root.getChildren();
    final Function1<KNode, Boolean> _function = new Function1<KNode, Boolean>() {
      public Boolean apply(final KNode c) {
        return Boolean.valueOf(Ptolemy2KGraphOptimization.this._markerExtensions.isMarkedAsHypernode(c));
      }
    };
    Iterable<KNode> _filter = IterableExtensions.<KNode>filter(_children, _function);
    final Iterator<KNode> relationsIterator = _filter.iterator();
    final List<KNode> nodesToBeRemoved = CollectionLiterals.<KNode>newArrayList();
    while (relationsIterator.hasNext()) {
      {
        final KNode relation = relationsIterator.next();
        boolean _and = false;
        EList<KEdge> _incomingEdges = relation.getIncomingEdges();
        int _size = _incomingEdges.size();
        boolean _equals = (_size == 1);
        if (!_equals) {
          _and = false;
        } else {
          EList<KEdge> _outgoingEdges = relation.getOutgoingEdges();
          int _size_1 = _outgoingEdges.size();
          boolean _equals_1 = (_size_1 == 1);
          _and = _equals_1;
        }
        if (_and) {
          EList<KEdge> _incomingEdges_1 = relation.getIncomingEdges();
          final KEdge inEdge = _incomingEdges_1.get(0);
          EList<KEdge> _outgoingEdges_1 = relation.getOutgoingEdges();
          final KEdge outEdge = _outgoingEdges_1.get(0);
          KNode _target = outEdge.getTarget();
          inEdge.setTarget(_target);
          KPort _targetPort = outEdge.getTargetPort();
          inEdge.setTargetPort(_targetPort);
          outEdge.setSource(null);
          outEdge.setTarget(null);
          outEdge.setSourcePort(null);
          outEdge.setTargetPort(null);
          List<PropertyType> _annotations = this._annotationExtensions.getAnnotations(inEdge);
          List<PropertyType> _annotations_1 = this._annotationExtensions.getAnnotations(relation);
          Iterables.<PropertyType>addAll(_annotations, _annotations_1);
          nodesToBeRemoved.add(relation);
        }
      }
    }
    for (final KNode node : nodesToBeRemoved) {
      EList<KNode> _children_1 = root.getChildren();
      _children_1.remove(node);
    }
    EList<KNode> _children_2 = root.getChildren();
    for (final KNode child : _children_2) {
      this.removeUnnecessaryRelations(child);
    }
  }
  
  /**
   * Recursively removes unnecessary relations in the model rooted at the given node. Unnecessary
   * relations are those that have one incoming and one outgoing edge.
   * 
   * @param root the model's root node.
   */
  private void removeAllRelations(final KNode root) {
    final HyperedgeGatherer magic = new HyperedgeGatherer(root);
    magic.gatherHyperedges();
    List<HyperedgeGatherer.Hyperedge> _hyperedges = magic.getHyperedges();
    for (final HyperedgeGatherer.Hyperedge hyperedge : _hyperedges) {
      {
        boolean _and = false;
        boolean _and_1 = false;
        int _size = hyperedge.relations.size();
        boolean _equals = (_size == 1);
        if (!_equals) {
          _and_1 = false;
        } else {
          int _size_1 = hyperedge.sourceNodes.size();
          int _size_2 = hyperedge.sourcePorts.size();
          int _plus = (_size_1 + _size_2);
          boolean _equals_1 = (_plus == 1);
          _and_1 = _equals_1;
        }
        if (!_and_1) {
          _and = false;
        } else {
          int _size_3 = hyperedge.targetNodes.size();
          int _size_4 = hyperedge.targetPorts.size();
          int _plus_1 = (_size_3 + _size_4);
          boolean _equals_2 = (_plus_1 == 1);
          _and = _equals_2;
        }
        final boolean preserveRelationAnnotations = _and;
        final List<PropertyType> relationAnnotations = CollectionLiterals.<PropertyType>newArrayList();
        if (preserveRelationAnnotations) {
          KNode _get = hyperedge.relations.get(0);
          List<PropertyType> _annotations = this._annotationExtensions.getAnnotations(_get);
          Iterables.<PropertyType>addAll(relationAnnotations, _annotations);
        }
        boolean _and_2 = false;
        boolean _and_3 = false;
        boolean _isEmpty = hyperedge.sourceNodes.isEmpty();
        if (!_isEmpty) {
          _and_3 = false;
        } else {
          boolean _isEmpty_1 = hyperedge.sourcePorts.isEmpty();
          _and_3 = _isEmpty_1;
        }
        boolean _not = (!_and_3);
        if (!_not) {
          _and_2 = false;
        } else {
          boolean _and_4 = false;
          boolean _isEmpty_2 = hyperedge.targetNodes.isEmpty();
          if (!_isEmpty_2) {
            _and_4 = false;
          } else {
            boolean _isEmpty_3 = hyperedge.targetPorts.isEmpty();
            _and_4 = _isEmpty_3;
          }
          boolean _not_1 = (!_and_4);
          _and_2 = _not_1;
        }
        if (_and_2) {
          for (final KNode relation : hyperedge.relations) {
            {
              EList<KNode> _children = root.getChildren();
              _children.remove(relation);
              while ((!relation.getIncomingEdges().isEmpty())) {
                {
                  EList<KEdge> _incomingEdges = relation.getIncomingEdges();
                  final KEdge edge = _incomingEdges.get(0);
                  edge.setSourcePort(null);
                  edge.setSource(null);
                  edge.setTargetPort(null);
                  edge.setTarget(null);
                }
              }
              while ((!relation.getOutgoingEdges().isEmpty())) {
                {
                  EList<KEdge> _outgoingEdges = relation.getOutgoingEdges();
                  final KEdge edge = _outgoingEdges.get(0);
                  edge.setSourcePort(null);
                  edge.setSource(null);
                  edge.setTargetPort(null);
                  edge.setTarget(null);
                }
              }
            }
          }
          for (final KNode sourceNode : hyperedge.sourceNodes) {
            {
              for (final KNode targetNode : hyperedge.targetNodes) {
                {
                  final KEdge newEdge = KimlUtil.createInitializedEdge();
                  List<PropertyType> _annotations_1 = this._annotationExtensions.getAnnotations(newEdge);
                  Iterables.<PropertyType>addAll(_annotations_1, relationAnnotations);
                  newEdge.setSource(sourceNode);
                  newEdge.setTarget(targetNode);
                }
              }
              for (final KPort targetPort : hyperedge.targetPorts) {
                {
                  final KEdge newEdge = KimlUtil.createInitializedEdge();
                  List<PropertyType> _annotations_1 = this._annotationExtensions.getAnnotations(newEdge);
                  Iterables.<PropertyType>addAll(_annotations_1, relationAnnotations);
                  newEdge.setSource(sourceNode);
                  KNode _node = targetPort.getNode();
                  newEdge.setTarget(_node);
                  newEdge.setTargetPort(targetPort);
                }
              }
            }
          }
          for (final KPort sourcePort : hyperedge.sourcePorts) {
            {
              for (final KNode targetNode : hyperedge.targetNodes) {
                {
                  final KEdge newEdge = KimlUtil.createInitializedEdge();
                  List<PropertyType> _annotations_1 = this._annotationExtensions.getAnnotations(newEdge);
                  Iterables.<PropertyType>addAll(_annotations_1, relationAnnotations);
                  KNode _node = sourcePort.getNode();
                  newEdge.setSource(_node);
                  newEdge.setSourcePort(sourcePort);
                  newEdge.setTarget(targetNode);
                }
              }
              for (final KPort targetPort : hyperedge.targetPorts) {
                {
                  final KEdge newEdge = KimlUtil.createInitializedEdge();
                  List<PropertyType> _annotations_1 = this._annotationExtensions.getAnnotations(newEdge);
                  Iterables.<PropertyType>addAll(_annotations_1, relationAnnotations);
                  KNode _node = sourcePort.getNode();
                  newEdge.setSource(_node);
                  newEdge.setSourcePort(sourcePort);
                  KNode _node_1 = targetPort.getNode();
                  newEdge.setTarget(_node_1);
                  newEdge.setTargetPort(targetPort);
                }
              }
            }
          }
        }
      }
    }
    EList<KNode> _children = root.getChildren();
    for (final KNode node : _children) {
      EList<KNode> _children_1 = node.getChildren();
      boolean _isEmpty = _children_1.isEmpty();
      boolean _not = (!_isEmpty);
      if (_not) {
        this.removeAllRelations(node);
      }
    }
  }
  
  /**
   * Removes ports from nodes representing modal model states in the model rooted at the given root
   * entity. Also, nodes containing such states are annotated for the layout algorithm to know to
   * use a layout algorithm for state machines.
   * 
   * @param root the model tree's root entity.
   */
  private void makeStatesPortless(final KNode root) {
    boolean _isMarkedAsState = this._markerExtensions.isMarkedAsState(root);
    if (_isMarkedAsState) {
      EList<KPort> _ports = root.getPorts();
      for (final KPort port : _ports) {
        while ((!port.getEdges().isEmpty())) {
          {
            EList<KEdge> _edges = port.getEdges();
            final KEdge edge = _edges.get(0);
            edge.setSourcePort(null);
            edge.setTargetPort(null);
          }
        }
      }
      EList<KPort> _ports_1 = root.getPorts();
      _ports_1.clear();
    }
    EList<KNode> _children = root.getChildren();
    for (final KNode child : _children) {
      this.makeStatesPortless(child);
    }
  }
  
  /**
   * Converts certain annotations into nodes of their own right. In particular, Ptolemy directors
   * are persisted as annotations in Ptolemy models, but need to be nodes in KGraph models to be
   * correctly displayed.
   * 
   * @param root root element of the model to look for convertible annotations in.
   */
  private void convertAnnotationsToNodes(final KNode root) {
    boolean _isMarkedAsFormerAnnotationNode = this._markerExtensions.isMarkedAsFormerAnnotationNode(root);
    if (_isMarkedAsFormerAnnotationNode) {
      return;
    }
    final List<PropertyType> parameterList = CollectionLiterals.<PropertyType>newLinkedList();
    List<PropertyType> _annotations = this._annotationExtensions.getAnnotations(root);
    final ListIterator<PropertyType> annotationsIterator = _annotations.listIterator();
    while (annotationsIterator.hasNext()) {
      {
        final PropertyType annotation = annotationsIterator.next();
        boolean _and = false;
        String _class_ = annotation.getClass_();
        boolean _notEquals = (!Objects.equal(_class_, null));
        if (!_notEquals) {
          _and = false;
        } else {
          String _class__1 = annotation.getClass_();
          boolean _endsWith = _class__1.endsWith("Director");
          _and = _endsWith;
        }
        if (_and) {
          final KNode directorNode = KimlUtil.createInitializedNode();
          String _name = annotation.getName();
          this._labelExtensions.setName(directorNode, _name);
          this._markerExtensions.markAsPtolemyElement(directorNode);
          this._markerExtensions.markAsDirector(directorNode);
          this._markerExtensions.markAsFormerAnnotationNode(directorNode);
          EList<KNode> _children = root.getChildren();
          _children.add(directorNode);
        } else {
          boolean _and_1 = false;
          String _class__2 = annotation.getClass_();
          boolean _equals = Objects.equal(_class__2, TransformationConstants.ANNOTATION_TYPE_PARAMETER);
          if (!_equals) {
            _and_1 = false;
          } else {
            String _value = annotation.getValue();
            boolean _isNullOrEmpty = StringExtensions.isNullOrEmpty(_value);
            boolean _not = (!_isNullOrEmpty);
            _and_1 = _not;
          }
          if (_and_1) {
            parameterList.add(annotation);
          }
        }
      }
    }
    boolean _and = false;
    EList<KNode> _children = root.getChildren();
    boolean _isEmpty = _children.isEmpty();
    boolean _not = (!_isEmpty);
    if (!_not) {
      _and = false;
    } else {
      boolean _isEmpty_1 = parameterList.isEmpty();
      boolean _not_1 = (!_isEmpty_1);
      _and = _not_1;
    }
    if (_and) {
      final List<Pair<String, String>> parameterPairs = CollectionLiterals.<Pair<String, String>>newLinkedList();
      for (final PropertyType parameter : parameterList) {
        String _name = parameter.getName();
        String _value = parameter.getValue();
        Pair<String, String> _pair = new Pair<String, String>(_name, _value);
        parameterPairs.add(_pair);
      }
      final KNode parameterNode = KimlUtil.createInitializedNode();
      this._markerExtensions.markAsPtolemyElement(parameterNode);
      this._markerExtensions.markAsFormerAnnotationNode(parameterNode);
      this._markerExtensions.markAsParameterNode(parameterNode);
      KLayoutData _layout = this._annotationExtensions.getLayout(parameterNode);
      _layout.<List<Pair<String, String>>>setProperty(PtolemyProperties.PT_PARAMETERS, parameterPairs);
      EList<KNode> _children_1 = root.getChildren();
      _children_1.add(parameterNode);
    }
    EList<KNode> _children_2 = root.getChildren();
    for (final KNode child : _children_2) {
      this.convertAnnotationsToNodes(child);
    }
  }
}
