/**
 * KIELER - Kiel Integrated Environment for Layout Eclipse RichClient
 * 
 * http://www.informatik.uni-kiel.de/rtsys/kieler/
 * 
 * Copyright 2012 by
 * + Christian-Albrechts-University of Kiel
 *   + Department of Computer Science
 *     + Real-Time and Embedded Systems Group
 * 
 * This code is provided under the terms of the Eclipse Public License (EPL).
 * See the file epl-v10.html for the license text.
 */
package de.cau.cs.kieler.ptolemy.klighd.transformation;

import com.google.common.base.Objects;
import com.google.common.base.Strings;
import com.google.common.collect.Iterables;
import com.google.inject.Inject;
import de.cau.cs.kieler.core.kgraph.KEdge;
import de.cau.cs.kieler.core.kgraph.KGraphElement;
import de.cau.cs.kieler.core.kgraph.KNode;
import de.cau.cs.kieler.core.kgraph.KPort;
import de.cau.cs.kieler.kiml.klayoutdata.KLayoutData;
import de.cau.cs.kieler.kiml.options.LayoutOptions;
import de.cau.cs.kieler.kiml.util.KimlUtil;
import de.cau.cs.kieler.klighd.syntheses.AbstractDiagramSynthesis;
import de.cau.cs.kieler.ptolemy.klighd.PluginConstants;
import de.cau.cs.kieler.ptolemy.klighd.transformation.PtolemyInterface;
import de.cau.cs.kieler.ptolemy.klighd.transformation.extensions.AnnotationExtensions;
import de.cau.cs.kieler.ptolemy.klighd.transformation.extensions.LabelExtensions;
import de.cau.cs.kieler.ptolemy.klighd.transformation.extensions.MarkerExtensions;
import de.cau.cs.kieler.ptolemy.klighd.transformation.extensions.MiscellaneousExtensions;
import de.cau.cs.kieler.ptolemy.klighd.transformation.util.TransformationConstants;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.xtext.xbase.lib.CollectionLiterals;
import org.eclipse.xtext.xbase.lib.Conversions;
import org.eclipse.xtext.xbase.lib.Exceptions;
import org.eclipse.xtext.xbase.lib.Extension;
import org.eclipse.xtext.xbase.lib.Functions.Function1;
import org.eclipse.xtext.xbase.lib.IterableExtensions;
import org.eclipse.xtext.xbase.lib.ListExtensions;
import org.ptolemy.moml.ClassType;
import org.ptolemy.moml.DocumentRoot;
import org.ptolemy.moml.EntityType;
import org.ptolemy.moml.LinkType;
import org.ptolemy.moml.PortType;
import org.ptolemy.moml.PropertyType;
import org.ptolemy.moml.RelationType;

/**
 * Transforms a Ptolemy2 model to a KGraph. This is step one of the Ptolemy model transformation
 * process, with step two consisting of the optimization of the transformed model as defined in
 * {@code Ptolemy2KGraphOptimization}, and step three consisting of providing the KRendering information
 * necessary to display the model as defined in {@code Ptolemy2KGraphVisualization}.
 * 
 * <p>In this part, all Ptolemy entities, relations and links are transformed to their KGraph
 * counterparts. The transformed objects are annotated to originate from a Ptolemy2 model. This
 * is fairly straightforward, except for the ports of actors. During the transformation, actors
 * must be instantiated using the KIELER Ptolemy library to find their ports. The problem is that
 * if an actor is not part of the official Ptolemy actor library, the instantiation often fails,
 * leaving us without properly defined ports. In particular, we then cannot determine if a port
 * is an input port or an output port, which leads to problems when the directions of links are
 * inferred in the second part of the import process. (contrary to KGraph edges, Ptolemy links are
 * undirected) The optimization part following this transformation part tries to solve the port
 * problem by heuristically inferring the direction of links.</p>
 * 
 * <p>The transformation works by recursively iterating over all entities defined in the model. For
 * each entity, we transform its properties to annotations, transform its child ports, its child
 * entities, and finally relations and links defined in the entity.</p>
 * 
 * <p><b>Note:</b> This transformation cannot simply be reused due to the way Xtend handles create
 * methods. To keep things simple, always use a new instance for each model to be transformed.</p>
 * 
 * @author cds
 * @author haf
 * @kieler.rating yellow 2012-06-15 KI-12 cmot, grh
 */
@SuppressWarnings("all")
public class Ptolemy2KGraphTransformation {
  /**
   * Accessing annotations.
   */
  @Inject
  @Extension
  private AnnotationExtensions _annotationExtensions;
  
  /**
   * Accessing labels for elements.
   */
  @Inject
  @Extension
  private LabelExtensions _labelExtensions;
  
  /**
   * Marking elements.
   */
  @Inject
  @Extension
  private MarkerExtensions _markerExtensions;
  
  /**
   * Further utility stuff.
   */
  @Inject
  @Extension
  private MiscellaneousExtensions _miscellaneousExtensions;
  
  /**
   * Interface to the Ptolemy library.
   */
  @Inject
  private PtolemyInterface ptolemy;
  
  /**
   * Flag indicating whether an instance of this transformation has already transformed something.
   * If so, it cannot be reused due to Xtend restrictions.
   */
  private boolean alreadyUsed = false;
  
  /**
   * List of warnings collected during the transformation. These will usually only be warnings about
   * actors that couldn't be instantiated.
   */
  private List<IStatus> warnings = new ArrayList<IStatus>();
  
  /**
   * The diagram synthesis from which this transformation is called. Used to save the mapping from
   * Ptolemy objects to view model objects.
   */
  private AbstractDiagramSynthesis<?> diagramSynthesis = null;
  
  /**
   * List of locally defined classes, which can be reused multiple times.
   */
  private List<KNode> actorClasses = new ArrayList<KNode>();
  
  /**
   * Entry point for the whole transformation business. Takes a Ptolemy MOML document's
   * root element and returns a KGraph node representing the converted model.
   * 
   * @param ptDocumentRoot the Ptolemy MOML document's root element.
   * @param synthesis the diagram synthesis calling this transformation. Used to map Ptolemy model
   *                  objects to KGraph model objects.
   * @return the transformed KGraph node.
   * @throws IllegalStateException if this class's instance has already been used.
   */
  public KNode transform(final DocumentRoot ptDocumentRoot, final AbstractDiagramSynthesis<?> synthesis) {
    if (this.alreadyUsed) {
      throw new IllegalStateException("Transformations cannot be reused.");
    }
    this.diagramSynthesis = synthesis;
    EntityType _entity = ptDocumentRoot.getEntity();
    KNode _transform = null;
    if (_entity!=null) {
      _transform=this.transform(_entity);
    }
    final KNode kEntityNode = _transform;
    boolean _notEquals = (!Objects.equal(kEntityNode, null));
    if (_notEquals) {
      return kEntityNode;
    }
    ClassType _class_ = ptDocumentRoot.getClass_();
    KNode _transform_1 = null;
    if (_class_!=null) {
      _transform_1=this.transform(_class_);
    }
    final KNode kClassNode = _transform_1;
    boolean _notEquals_1 = (!Objects.equal(kClassNode, null));
    if (_notEquals_1) {
      return kClassNode;
    }
    return KimlUtil.createInitializedNode();
  }
  
  /**
   * Transforms the given Ptolemy entity and its children into a KNode.
   * 
   * @param ptEntity the Ptolemy entity to transform.
   * @return the KGraph node.
   */
  private KNode transform(final EntityType ptEntity) {
    final ArrayList<?> _cacheKey = CollectionLiterals.newArrayList(ptEntity);
    final KNode _result;
    synchronized (_createCache_transform) {
      if (_createCache_transform.containsKey(_cacheKey)) {
        return _createCache_transform.get(_cacheKey);
      }
      KNode _createInitializedNode = KimlUtil.createInitializedNode();
      _result = _createInitializedNode;
      _createCache_transform.put(_cacheKey, _result);
    }
    _init_transform(_result, ptEntity);
    return _result;
  }
  
  private final HashMap<ArrayList<?>, KNode> _createCache_transform = CollectionLiterals.newHashMap();
  
  private void _init_transform(final KNode kNode, final EntityType ptEntity) {
    this.diagramSynthesis.<KNode>associateWith(kNode, ptEntity);
    String _name = ptEntity.getName();
    this._labelExtensions.setName(kNode, _name);
    this._markerExtensions.markAsPtolemyElement(kNode);
    String _class1 = ptEntity.getClass1();
    this._annotationExtensions.addAnnotation(kNode, TransformationConstants.ANNOTATION_PTOLEMY_CLASS, _class1);
    EList<PropertyType> _property = ptEntity.getProperty();
    this.addProperties(kNode, _property);
    this.addChildPorts(kNode, ptEntity);
    String _class1_1 = ptEntity.getClass1();
    boolean _equals = _class1_1.equals(TransformationConstants.ENTITY_CLASS_FSM);
    if (_equals) {
      this._markerExtensions.markAsStateMachineContainer(kNode);
      EList<EntityType> _entity = ptEntity.getEntity();
      this.addChildEntities(kNode, ((EntityType[])Conversions.unwrapArray(_entity, EntityType.class)));
      EList<RelationType> _relation = ptEntity.getRelation();
      this.addChildRelations(kNode, ((RelationType[])Conversions.unwrapArray(_relation, RelationType.class)));
      EList<LinkType> _link = ptEntity.getLink();
      this.addChildLinks(kNode, ((LinkType[])Conversions.unwrapArray(_link, LinkType.class)));
    } else {
      boolean _or = false;
      String _class1_2 = ptEntity.getClass1();
      boolean _equals_1 = _class1_2.equals(TransformationConstants.ENTITY_CLASS_MODAL_MODEL);
      if (_equals_1) {
        _or = true;
      } else {
        String _class1_3 = ptEntity.getClass1();
        boolean _equals_2 = _class1_3.equals(TransformationConstants.ENTITY_CLASS_FSM_MODAL_MODEL);
        _or = _equals_2;
      }
      if (_or) {
        this._markerExtensions.markAsStateMachineContainer(kNode);
        EList<EntityType> _entity_1 = ptEntity.getEntity();
        for (final EntityType child : _entity_1) {
          String _name_1 = child.getName();
          boolean _equals_3 = _name_1.equals(TransformationConstants.ENTITY_NAME_MODAL_CONTROLLER);
          if (_equals_3) {
            EList<EntityType> _entity_2 = child.getEntity();
            this.addChildEntities(kNode, ((EntityType[])Conversions.unwrapArray(_entity_2, EntityType.class)));
            EList<RelationType> _relation_1 = child.getRelation();
            this.addChildRelations(kNode, ((RelationType[])Conversions.unwrapArray(_relation_1, RelationType.class)));
            EList<LinkType> _link_1 = child.getLink();
            this.addChildLinks(kNode, ((LinkType[])Conversions.unwrapArray(_link_1, LinkType.class)));
          }
        }
      } else {
        boolean _or_1 = false;
        String _class1_4 = ptEntity.getClass1();
        boolean _equals_4 = _class1_4.equals(TransformationConstants.ENTITY_CLASS_STATE);
        if (_equals_4) {
          _or_1 = true;
        } else {
          String _class1_5 = ptEntity.getClass1();
          boolean _equals_5 = _class1_5.equals(TransformationConstants.ENTITY_CLASS_FSM_STATE);
          _or_1 = _equals_5;
        }
        if (_or_1) {
          final EntityType refinement = this._miscellaneousExtensions.findRefinement(ptEntity);
          boolean _notEquals = (!Objects.equal(refinement, null));
          if (_notEquals) {
            final KNode transformedRefinement = this.transform(refinement);
            EList<KNode> _children = kNode.getChildren();
            EList<KNode> _children_1 = transformedRefinement.getChildren();
            Iterables.<KNode>addAll(_children, _children_1);
            while ((!transformedRefinement.getPorts().isEmpty())) {
              EList<KNode> _children_2 = kNode.getChildren();
              EList<KPort> _ports = transformedRefinement.getPorts();
              KPort _get = _ports.get(0);
              KNode _transformRefinementPort = this.transformRefinementPort(_get);
              _children_2.add(_transformRefinementPort);
            }
            List<PropertyType> _annotations = this._annotationExtensions.getAnnotations(kNode);
            List<PropertyType> _annotations_1 = this._annotationExtensions.getAnnotations(transformedRefinement);
            Iterables.<PropertyType>addAll(_annotations, _annotations_1);
            String _annotationValue = this._annotationExtensions.getAnnotationValue(transformedRefinement, 
              TransformationConstants.ANNOTATION_PTOLEMY_CLASS);
            final String refinementClass = Strings.nullToEmpty(_annotationValue);
            boolean _or_2 = false;
            boolean _equals_6 = refinementClass.equals(TransformationConstants.ENTITY_CLASS_MODEL_CONTROLLER);
            if (_equals_6) {
              _or_2 = true;
            } else {
              boolean _equals_7 = refinementClass.equals(TransformationConstants.ENTITY_CLASS_FSM_MODEL_CONTROLLER);
              _or_2 = _equals_7;
            }
            if (_or_2) {
              this._markerExtensions.markAsStateMachineContainer(kNode);
            }
          }
        } else {
          EList<ClassType> _class_ = ptEntity.getClass_();
          final Function1<ClassType, KNode> _function = new Function1<ClassType, KNode>() {
            public KNode apply(final ClassType c) {
              return Ptolemy2KGraphTransformation.this.transform(c);
            }
          };
          List<KNode> _map = ListExtensions.<ClassType, KNode>map(_class_, _function);
          this.actorClasses.addAll(_map);
          EList<EntityType> _entity_3 = ptEntity.getEntity();
          this.addChildEntities(kNode, ((EntityType[])Conversions.unwrapArray(_entity_3, EntityType.class)));
          EList<RelationType> _relation_2 = ptEntity.getRelation();
          this.addChildRelations(kNode, ((RelationType[])Conversions.unwrapArray(_relation_2, RelationType.class)));
          EList<LinkType> _link_2 = ptEntity.getLink();
          this.addChildLinks(kNode, ((LinkType[])Conversions.unwrapArray(_link_2, LinkType.class)));
        }
      }
    }
  }
  
  /**
   * Transforms the given Ptolemy class and its children into a KNode.
   * 
   * <p>Assumption: a ClassType node can only occur at the top level. If this assumption turns out
   * not to be true, we need to overload quite a bunch of methods. Which we don't want to do. So
   * it better be true.</p>
   * 
   * @param ptClass the Ptolemy class to transform.
   * @return the KGraph node.
   */
  private KNode transform(final ClassType ptClass) {
    final ArrayList<?> _cacheKey = CollectionLiterals.newArrayList(ptClass);
    final KNode _result;
    synchronized (_createCache_transform_1) {
      if (_createCache_transform_1.containsKey(_cacheKey)) {
        return _createCache_transform_1.get(_cacheKey);
      }
      KNode _createInitializedNode = KimlUtil.createInitializedNode();
      _result = _createInitializedNode;
      _createCache_transform_1.put(_cacheKey, _result);
    }
    _init_transform_1(_result, ptClass);
    return _result;
  }
  
  private final HashMap<ArrayList<?>, KNode> _createCache_transform_1 = CollectionLiterals.newHashMap();
  
  private void _init_transform_1(final KNode kNode, final ClassType ptClass) {
    String _name = ptClass.getName();
    this._labelExtensions.setName(kNode, _name);
    this._markerExtensions.markAsPtolemyElement(kNode);
    String extendedClass = ptClass.getExtends();
    final Function1<KNode, Boolean> _function = new Function1<KNode, Boolean>() {
      public Boolean apply(final KNode c) {
        String _name = Ptolemy2KGraphTransformation.this._labelExtensions.getName(c);
        String _extends = ptClass.getExtends();
        return Boolean.valueOf(Objects.equal(_name, _extends));
      }
    };
    final KNode matchingLocalClass = IterableExtensions.<KNode>findFirst(this.actorClasses, _function);
    boolean _notEquals = (!Objects.equal(matchingLocalClass, null));
    if (_notEquals) {
      PropertyType _annotation = this._annotationExtensions.getAnnotation(matchingLocalClass, TransformationConstants.ANNOTATION_PTOLEMY_CLASS);
      String _value = _annotation.getValue();
      extendedClass = _value;
    }
    this._annotationExtensions.addAnnotation(kNode, TransformationConstants.ANNOTATION_PTOLEMY_CLASS, extendedClass);
    EList<PropertyType> _property = ptClass.getProperty();
    this.addProperties(kNode, _property);
    this.addChildPorts(kNode, ptClass);
    boolean _equals = extendedClass.equals(TransformationConstants.ENTITY_CLASS_FSM);
    if (_equals) {
      this._markerExtensions.markAsStateMachineContainer(kNode);
      EList<EntityType> _entity = ptClass.getEntity();
      this.addChildEntities(kNode, ((EntityType[])Conversions.unwrapArray(_entity, EntityType.class)));
      EList<RelationType> _relation = ptClass.getRelation();
      this.addChildRelations(kNode, ((RelationType[])Conversions.unwrapArray(_relation, RelationType.class)));
      EList<LinkType> _link = ptClass.getLink();
      this.addChildLinks(kNode, ((LinkType[])Conversions.unwrapArray(_link, LinkType.class)));
    } else {
      boolean _or = false;
      boolean _equals_1 = extendedClass.equals(TransformationConstants.ENTITY_CLASS_MODAL_MODEL);
      if (_equals_1) {
        _or = true;
      } else {
        boolean _equals_2 = extendedClass.equals(TransformationConstants.ENTITY_CLASS_FSM_MODAL_MODEL);
        _or = _equals_2;
      }
      if (_or) {
        this._markerExtensions.markAsStateMachineContainer(kNode);
        EList<EntityType> _entity_1 = ptClass.getEntity();
        for (final EntityType child : _entity_1) {
          String _name_1 = child.getName();
          boolean _equals_3 = _name_1.equals(TransformationConstants.ENTITY_NAME_MODAL_CONTROLLER);
          if (_equals_3) {
            EList<EntityType> _entity_2 = child.getEntity();
            this.addChildEntities(kNode, ((EntityType[])Conversions.unwrapArray(_entity_2, EntityType.class)));
            EList<RelationType> _relation_1 = child.getRelation();
            this.addChildRelations(kNode, ((RelationType[])Conversions.unwrapArray(_relation_1, RelationType.class)));
            EList<LinkType> _link_1 = child.getLink();
            this.addChildLinks(kNode, ((LinkType[])Conversions.unwrapArray(_link_1, LinkType.class)));
          }
        }
      } else {
        boolean _or_1 = false;
        boolean _equals_4 = extendedClass.equals(TransformationConstants.ENTITY_CLASS_MODAL_MODEL);
        if (_equals_4) {
          _or_1 = true;
        } else {
          boolean _equals_5 = extendedClass.equals(TransformationConstants.ENTITY_CLASS_FSM_MODAL_MODEL);
          _or_1 = _equals_5;
        }
        if (_or_1) {
          final EntityType refinement = this._miscellaneousExtensions.findRefinement(ptClass);
          boolean _notEquals_1 = (!Objects.equal(refinement, null));
          if (_notEquals_1) {
            final KNode transformedRefinement = this.transform(refinement);
            EList<KNode> _children = kNode.getChildren();
            EList<KNode> _children_1 = transformedRefinement.getChildren();
            Iterables.<KNode>addAll(_children, _children_1);
            while ((!transformedRefinement.getPorts().isEmpty())) {
              EList<KNode> _children_2 = kNode.getChildren();
              EList<KPort> _ports = transformedRefinement.getPorts();
              KPort _get = _ports.get(0);
              KNode _transformRefinementPort = this.transformRefinementPort(_get);
              _children_2.add(_transformRefinementPort);
            }
            List<PropertyType> _annotations = this._annotationExtensions.getAnnotations(kNode);
            List<PropertyType> _annotations_1 = this._annotationExtensions.getAnnotations(transformedRefinement);
            Iterables.<PropertyType>addAll(_annotations, _annotations_1);
            String _annotationValue = this._annotationExtensions.getAnnotationValue(transformedRefinement, 
              TransformationConstants.ANNOTATION_PTOLEMY_CLASS);
            final String refinementClass = Strings.nullToEmpty(_annotationValue);
            boolean _or_2 = false;
            boolean _equals_6 = refinementClass.equals(TransformationConstants.ENTITY_CLASS_MODEL_CONTROLLER);
            if (_equals_6) {
              _or_2 = true;
            } else {
              boolean _equals_7 = refinementClass.equals(TransformationConstants.ENTITY_CLASS_FSM_MODEL_CONTROLLER);
              _or_2 = _equals_7;
            }
            if (_or_2) {
              this._markerExtensions.markAsStateMachineContainer(kNode);
            }
          }
        } else {
          EList<ClassType> _class_ = ptClass.getClass_();
          final Function1<ClassType, KNode> _function_1 = new Function1<ClassType, KNode>() {
            public KNode apply(final ClassType c) {
              return Ptolemy2KGraphTransformation.this.transform(c);
            }
          };
          List<KNode> _map = ListExtensions.<ClassType, KNode>map(_class_, _function_1);
          this.actorClasses.addAll(_map);
          EList<EntityType> _entity_3 = ptClass.getEntity();
          this.addChildEntities(kNode, ((EntityType[])Conversions.unwrapArray(_entity_3, EntityType.class)));
          EList<RelationType> _relation_2 = ptClass.getRelation();
          this.addChildRelations(kNode, ((RelationType[])Conversions.unwrapArray(_relation_2, RelationType.class)));
          EList<LinkType> _link_2 = ptClass.getLink();
          this.addChildLinks(kNode, ((LinkType[])Conversions.unwrapArray(_link_2, LinkType.class)));
        }
      }
    }
  }
  
  /**
   * Transforms the given Ptolemy relation into a KNode and marks it as being a hypernode.
   * 
   * @param ptRelation the Ptolemy relation to transform.
   * @return the KGraph node.
   */
  private KNode transform(final RelationType ptRelation) {
    final ArrayList<?> _cacheKey = CollectionLiterals.newArrayList(ptRelation);
    final KNode _result;
    synchronized (_createCache_transform_2) {
      if (_createCache_transform_2.containsKey(_cacheKey)) {
        return _createCache_transform_2.get(_cacheKey);
      }
      KNode _createInitializedNode = KimlUtil.createInitializedNode();
      _result = _createInitializedNode;
      _createCache_transform_2.put(_cacheKey, _result);
    }
    _init_transform_2(_result, ptRelation);
    return _result;
  }
  
  private final HashMap<ArrayList<?>, KNode> _createCache_transform_2 = CollectionLiterals.newHashMap();
  
  private void _init_transform_2(final KNode kNode, final RelationType ptRelation) {
    this.diagramSynthesis.<KNode>associateWith(kNode, ptRelation);
    String _name = ptRelation.getName();
    this._labelExtensions.setName(kNode, _name);
    this._markerExtensions.markAsPtolemyElement(kNode);
    this._markerExtensions.markAsHypernode(kNode);
    EList<PropertyType> _property = ptRelation.getProperty();
    this.addProperties(kNode, _property);
  }
  
  /**
   * Transforms the given Ptolemy link into a KGraph edge. The problem here is that a Ptolemy link
   * is (1) undirected and (2) has multiple possible attributes for head and tail.
   * 
   * <p>The former is ignored at this point: the edge is simply annotated as not having had its
   * direction determined yet, which is one of the responsibilities of the second part of the
   * transformation.</p>
   * 
   * <p>The latter comes from the fact that an edge can connect either a port and a relation, or two
   * relations. We solve that by collecting all the elements incident to this link, hoping that there
   * are only two (if that is not the case, we leave the edge unconnected). The two are then
   * connected to the transformed edge. Their order is not relevant, since the edge's
   * direction is currently unknown anyway.</p>
   * 
   * <p>Note: This method expects relations of the {@code kParent} to already have been
   * transformed.</p>
   * 
   * @param ptLink the Ptolemy link to transform.
   * @param kaomParent the link's parent entity, with relations already transformed.
   * @return the transformed KAOM link.
   */
  private KEdge transform(final LinkType ptLink, final KNode kParent) {
    final ArrayList<?> _cacheKey = CollectionLiterals.newArrayList(ptLink, kParent);
    final KEdge _result;
    synchronized (_createCache_transform_3) {
      if (_createCache_transform_3.containsKey(_cacheKey)) {
        return _createCache_transform_3.get(_cacheKey);
      }
      KEdge _createInitializedEdge = KimlUtil.createInitializedEdge();
      _result = _createInitializedEdge;
      _createCache_transform_3.put(_cacheKey, _result);
    }
    _init_transform(_result, ptLink, kParent);
    return _result;
  }
  
  private final HashMap<ArrayList<?>, KEdge> _createCache_transform_3 = CollectionLiterals.newHashMap();
  
  private void _init_transform(final KEdge kEdge, final LinkType ptLink, final KNode kParent) {
    try {
      EList<KNode> _children = kParent.getChildren();
      final Function1<KNode, Boolean> _function = new Function1<KNode, Boolean>() {
        public Boolean apply(final KNode r) {
          String _name = Ptolemy2KGraphTransformation.this._labelExtensions.getName(r);
          String _relation = ptLink.getRelation();
          return Boolean.valueOf(_name.equals(_relation));
        }
      };
      final KNode kRelation = IterableExtensions.<KNode>findFirst(_children, _function);
      EList<KNode> _children_1 = kParent.getChildren();
      final Function1<KNode, Boolean> _function_1 = new Function1<KNode, Boolean>() {
        public Boolean apply(final KNode r) {
          String _name = Ptolemy2KGraphTransformation.this._labelExtensions.getName(r);
          String _relation1 = ptLink.getRelation1();
          return Boolean.valueOf(_name.equals(_relation1));
        }
      };
      final KNode kRelation1 = IterableExtensions.<KNode>findFirst(_children_1, _function_1);
      EList<KNode> _children_2 = kParent.getChildren();
      final Function1<KNode, Boolean> _function_2 = new Function1<KNode, Boolean>() {
        public Boolean apply(final KNode r) {
          String _name = Ptolemy2KGraphTransformation.this._labelExtensions.getName(r);
          String _relation2 = ptLink.getRelation2();
          return Boolean.valueOf(_name.equals(_relation2));
        }
      };
      final KNode kRelation2 = IterableExtensions.<KNode>findFirst(_children_2, _function_2);
      KPort _xifexpression = null;
      String _port = ptLink.getPort();
      boolean _equals = Objects.equal(_port, null);
      if (_equals) {
        _xifexpression = null;
      } else {
        String _port_1 = ptLink.getPort();
        _xifexpression = this.getOrCreatePortByName(kParent, _port_1);
      }
      final KPort kPort = _xifexpression;
      final ArrayList<KGraphElement> endpoints = new ArrayList<KGraphElement>();
      this._markerExtensions.markAsPtolemyElement(kEdge);
      boolean _notEquals = (!Objects.equal(kRelation, null));
      if (_notEquals) {
        endpoints.add(kRelation);
      }
      boolean _notEquals_1 = (!Objects.equal(kRelation1, null));
      if (_notEquals_1) {
        endpoints.add(kRelation1);
      }
      boolean _notEquals_2 = (!Objects.equal(kRelation2, null));
      if (_notEquals_2) {
        endpoints.add(kRelation2);
      }
      boolean _notEquals_3 = (!Objects.equal(kPort, null));
      if (_notEquals_3) {
        endpoints.add(kPort);
      }
      int _size = endpoints.size();
      boolean _equals_1 = (_size == 2);
      if (_equals_1) {
        KGraphElement _get = endpoints.get(0);
        if ((_get instanceof KNode)) {
          KGraphElement _get_1 = endpoints.get(0);
          kEdge.setSource(((KNode) _get_1));
        } else {
          KGraphElement _get_2 = endpoints.get(0);
          if ((_get_2 instanceof KPort)) {
            KGraphElement _get_3 = endpoints.get(0);
            KNode _node = ((KPort) _get_3).getNode();
            kEdge.setSource(_node);
            KGraphElement _get_4 = endpoints.get(0);
            kEdge.setSourcePort(((KPort) _get_4));
          }
        }
        KGraphElement _get_5 = endpoints.get(1);
        if ((_get_5 instanceof KNode)) {
          KGraphElement _get_6 = endpoints.get(1);
          kEdge.setTarget(((KNode) _get_6));
        } else {
          KGraphElement _get_7 = endpoints.get(1);
          if ((_get_7 instanceof KPort)) {
            KGraphElement _get_8 = endpoints.get(1);
            KNode _node_1 = ((KPort) _get_8).getNode();
            kEdge.setTarget(_node_1);
            KGraphElement _get_9 = endpoints.get(1);
            kEdge.setTargetPort(((KPort) _get_9));
          }
        }
      }
      this._markerExtensions.markAsUndirected(kEdge, true);
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  /**
   * Transforms the given Ptolemy port into a KPort.
   * 
   * @param ptPort the Ptolemy port to transform.
   * @return the KPort.
   */
  private KPort transform(final PortType ptPort) {
    final ArrayList<?> _cacheKey = CollectionLiterals.newArrayList(ptPort);
    final KPort _result;
    synchronized (_createCache_transform_4) {
      if (_createCache_transform_4.containsKey(_cacheKey)) {
        return _createCache_transform_4.get(_cacheKey);
      }
      KPort _createInitializedPort = KimlUtil.createInitializedPort();
      _result = _createInitializedPort;
      _createCache_transform_4.put(_cacheKey, _result);
    }
    _init_transform_3(_result, ptPort);
    return _result;
  }
  
  private final HashMap<ArrayList<?>, KPort> _createCache_transform_4 = CollectionLiterals.newHashMap();
  
  private void _init_transform_3(final KPort kPort, final PortType ptPort) {
    String _name = ptPort.getName();
    this._labelExtensions.setName(kPort, _name);
    this._markerExtensions.markAsPtolemyElement(kPort);
    String _class_ = ptPort.getClass_();
    this._annotationExtensions.addAnnotation(kPort, TransformationConstants.ANNOTATION_PTOLEMY_CLASS, _class_);
    EList<PropertyType> _property = ptPort.getProperty();
    this.addProperties(kPort, _property);
  }
  
  /**
   * Looks for a port with the given name under the given parent node. The port name may be of the
   * following two forms:
   * 
   * <ol>
   *   <li>{@code portName}<br/>
   *       In this simple form, the port is assumed to belong to the parent entity. If the port does
   *       not exist, it is created and added to the parent entity.
   *   </li>
   *   <li>{@code actorName.portName}<br/>
   *       In this extended form, the port is assumed to belong to an actor of the given name. The
   *       actor must be a direct child of the parent entity. If the actor doesn't have a port of the
   *       given name, it is created and added to the model. If an actor of the given name doesn't
   *       exist, the model is not modified.
   *   </li>
   * </ol>
   * 
   * @param kparent the node to look for ports and actors in.
   * @param name the name of the port to find or create.
   * @return the port.
   * @throws CoreException if the port name does not follow the expected format, or if the referenced
   *                       actor could not be found.
   */
  private KPort getOrCreatePortByName(final KNode kparent, final String name) throws CoreException {
    KPort _xblockexpression = null;
    {
      String[] _split = name.split(TransformationConstants.PORT_NAME_SEPARATOR_REGEX);
      final ArrayList<String> nameParts = CollectionLiterals.<String>newArrayList(_split);
      boolean _or = false;
      int _size = nameParts.size();
      boolean _lessThan = (_size < 1);
      if (_lessThan) {
        _or = true;
      } else {
        int _size_1 = nameParts.size();
        boolean _greaterThan = (_size_1 > 2);
        _or = _greaterThan;
      }
      if (_or) {
        String _replace = "Port name format not recognized: %1".replace("%1", name);
        Status _status = new Status(
          IStatus.ERROR, 
          PluginConstants.PLUGIN_ID, _replace, 
          null);
        throw new CoreException(_status);
      }
      KNode _switchResult = null;
      int _size_2 = nameParts.size();
      switch (_size_2) {
        case 1:
          _switchResult = kparent;
          break;
        case 2:
          EList<KNode> _children = kparent.getChildren();
          final Function1<KNode, Boolean> _function = new Function1<KNode, Boolean>() {
            public Boolean apply(final KNode a) {
              String _name = Ptolemy2KGraphTransformation.this._labelExtensions.getName(a);
              String _get = nameParts.get(0);
              return Boolean.valueOf(_name.equals(_get));
            }
          };
          _switchResult = IterableExtensions.<KNode>findFirst(_children, _function);
          break;
      }
      final KNode kActor = _switchResult;
      boolean _equals = Objects.equal(kActor, null);
      if (_equals) {
        String _replace_1 = "Port name references unknown actor: %1".replace("%1", name);
        Status _status_1 = new Status(
          IStatus.ERROR, 
          PluginConstants.PLUGIN_ID, _replace_1, 
          null);
        throw new CoreException(_status_1);
      }
      int _size_3 = nameParts.size();
      int _minus = (_size_3 - 1);
      final String portName = nameParts.get(_minus);
      EList<KPort> _ports = kActor.getPorts();
      final Function1<KPort, Boolean> _function_1 = new Function1<KPort, Boolean>() {
        public Boolean apply(final KPort p) {
          String _name = Ptolemy2KGraphTransformation.this._labelExtensions.getName(p);
          return Boolean.valueOf(_name.equals(portName));
        }
      };
      final KPort port = IterableExtensions.<KPort>findFirst(_ports, _function_1);
      KPort _xifexpression = null;
      boolean _equals_1 = Objects.equal(port, null);
      if (_equals_1) {
        EList<KPort> _ports_1 = kActor.getPorts();
        int _size_4 = _ports_1.size();
        _xifexpression = this.createPort(kActor, portName, _size_4);
      } else {
        _xifexpression = port;
      }
      _xblockexpression = _xifexpression;
    }
    return _xblockexpression;
  }
  
  /**
   * Creates a port with the given name and adds it to the given node.
   * 
   * @param kNode the node to create the port for.
   * @param name the name of the port to create.
   * @return the created port.
   */
  private KPort createPort(final KNode kNode, final String name, final int index) {
    KPort _xblockexpression = null;
    {
      final KPort result = KimlUtil.createInitializedPort();
      KLayoutData _layout = this._annotationExtensions.getLayout(result);
      _layout.<Integer>setProperty(LayoutOptions.PORT_INDEX, Integer.valueOf(index));
      this._labelExtensions.setName(result, name);
      this._markerExtensions.markAsPtolemyElement(result);
      result.setNode(kNode);
      _xblockexpression = result;
    }
    return _xblockexpression;
  }
  
  /**
   * Takes the given port of a state refinement and creates a node for it that the port is moved to.
   * This is necessary since states don't have hierarchical ports, but state refinements sometimes
   * need them. We thus display such ports as nodes with special rendering, just like in Ptolemy.
   * 
   * @param port the port to turn into a node.
   * @return the node with the port attached.
   */
  private KNode transformRefinementPort(final KPort port) {
    final KNode node = KimlUtil.createInitializedNode();
    EList<KPort> _ports = node.getPorts();
    _ports.add(port);
    this._markerExtensions.markAsModalModelPort(node);
    final boolean inputPort = this._annotationExtensions.hasAnnotation(port, "input");
    final boolean outputPort = this._annotationExtensions.hasAnnotation(port, "output");
    final boolean multiPort = this._annotationExtensions.hasAnnotation(port, "multiport");
    boolean _and = false;
    if (!inputPort) {
      _and = false;
    } else {
      _and = (!outputPort);
    }
    if (_and) {
      this._markerExtensions.markAsInputPort(node, true);
      this._markerExtensions.markAsInputPort(port, false);
      this._markerExtensions.markAsOutputPort(port, true);
    } else {
      boolean _and_1 = false;
      if (!(!inputPort)) {
        _and_1 = false;
      } else {
        _and_1 = outputPort;
      }
      if (_and_1) {
        this._markerExtensions.markAsOutputPort(node, true);
        this._markerExtensions.markAsOutputPort(port, false);
        this._markerExtensions.markAsInputPort(port, true);
      } else {
        boolean _and_2 = false;
        if (!inputPort) {
          _and_2 = false;
        } else {
          _and_2 = outputPort;
        }
        if (_and_2) {
          this._markerExtensions.markAsInputPort(node, true);
          this._markerExtensions.markAsOutputPort(node, true);
          this._markerExtensions.markAsInputPort(port, true);
          this._markerExtensions.markAsOutputPort(port, true);
        }
      }
    }
    if (multiPort) {
      this._annotationExtensions.addAnnotation(node, "multiport");
    }
    EList<KEdge> _edges = port.getEdges();
    for (final KEdge edge : _edges) {
      {
        KPort _sourcePort = edge.getSourcePort();
        boolean _equals = Objects.equal(_sourcePort, port);
        if (_equals) {
          edge.setSource(node);
        }
        KPort _targetPort = edge.getTargetPort();
        boolean _equals_1 = Objects.equal(_targetPort, port);
        if (_equals_1) {
          edge.setTarget(node);
        }
      }
    }
    return node;
  }
  
  /**
   * Recursively transforms the given list of properties and adds them to the given annotatable
   * object.
   * 
   * @param annotatable object to add the transformed properties to.
   * @param ptProperties list of properties to transform.
   */
  private boolean addProperties(final KGraphElement element, final List<PropertyType> ptProperties) {
    List<PropertyType> _annotations = this._annotationExtensions.getAnnotations(element);
    return _annotations.addAll(ptProperties);
  }
  
  /**
   * Transforms the given list of entities and adds them to the given parent entity.
   * 
   * @param parent the new parent node of the transformed entities.
   * @param ptEntities list of entities to transform
   */
  private void addChildEntities(final KNode parent, final EntityType[] ptEntities) {
    for (final EntityType ptEntity : ptEntities) {
      {
        final Function1<KNode, Boolean> _function = new Function1<KNode, Boolean>() {
          public Boolean apply(final KNode c) {
            String _name = Ptolemy2KGraphTransformation.this._labelExtensions.getName(c);
            String _class1 = ptEntity.getClass1();
            return Boolean.valueOf(Objects.equal(_name, _class1));
          }
        };
        final KNode matchingClass = IterableExtensions.<KNode>findFirst(this.actorClasses, _function);
        boolean _notEquals = (!Objects.equal(matchingClass, null));
        if (_notEquals) {
          final KNode copy = EcoreUtil.<KNode>copy(matchingClass);
          EList<KNode> _children = parent.getChildren();
          _children.add(copy);
          this.diagramSynthesis.<KNode>associateWith(copy, ptEntity);
          String _name = ptEntity.getName();
          this._labelExtensions.setName(copy, _name);
        } else {
          EList<KNode> _children_1 = parent.getChildren();
          KNode _transform = this.transform(ptEntity);
          _children_1.add(_transform);
        }
      }
    }
  }
  
  /**
   * Transforms the given list of relations and adds them to the given parent entity.
   * 
   * @param parent the new parent node of the transformed relations.
   * @param ptRelations list of relations to transform
   */
  private void addChildRelations(final KNode parent, final RelationType[] ptRelations) {
    for (final RelationType ptRelation : ptRelations) {
      EList<KNode> _children = parent.getChildren();
      KNode _transform = this.transform(ptRelation);
      _children.add(_transform);
    }
  }
  
  /**
   * Transforms the given list of links and adds them to the given parent node.
   * 
   * @param parent the node containing the relations and actors connected by the links once they
   *               are transformed.
   * @param ptLinks list of links to transform
   */
  private void addChildLinks(final KNode kParent, final LinkType[] ptLinks) {
    for (final LinkType ptLink : ptLinks) {
      this.transform(ptLink, kParent);
    }
  }
  
  /**
   * Adds ports to the given node created from the given Ptolemy entity or class.
   * 
   * <p>This is no trivial task. Ports are not necessarily explicitly defined objects in a Ptolemy
   * model. Usually, they are created implicitly in an actor's Java implementation. Hence, to see
   * what ports an entity has, we need to try to instantiate it in Ptolemy.</p>
   * 
   * <p>If the actor is not available in KIELER's Ptolemy library, instantiating an actor to check
   * for its ports will of course fail, leaving the entity without ports. To handle these cases, such
   * ports are created once they are referenced later on. However, their specific attributes (e.g.,
   * whether they are input ports or output ports) will then be unavailable, which in turn may cause
   * link directions to be incorrectly inferred.</p>
   * 
   * @param knode the node to add the ports to.
   * @param entityOrClass the entity or class that was transformed into the knode.
   */
  private void addChildPorts(final KNode knode, final EObject entityOrClass) {
    final ArrayList<KPort> ports = new ArrayList<KPort>();
    try {
      List<KPort> _portsFromImplementation = this.ptolemy.getPortsFromImplementation(entityOrClass);
      ports.addAll(_portsFromImplementation);
    } catch (final Throwable _t) {
      if (_t instanceof Exception) {
        final Exception e = (Exception)_t;
        String _message = e.getMessage();
        Status _status = new Status(
          IStatus.WARNING, 
          PluginConstants.PLUGIN_ID, _message, e);
        this.warnings.add(_status);
      } else {
        throw Exceptions.sneakyThrow(_t);
      }
    }
    List<KPort> _xifexpression = null;
    if ((entityOrClass instanceof EntityType)) {
      EList<PortType> _port = ((EntityType) entityOrClass).getPort();
      final Function1<PortType, KPort> _function = new Function1<PortType, KPort>() {
        public KPort apply(final PortType p) {
          return Ptolemy2KGraphTransformation.this.transform(p);
        }
      };
      _xifexpression = ListExtensions.<PortType, KPort>map(_port, _function);
    } else {
      List<KPort> _xifexpression_1 = null;
      if ((entityOrClass instanceof ClassType)) {
        EList<PortType> _port_1 = ((ClassType) entityOrClass).getPort();
        final Function1<PortType, KPort> _function_1 = new Function1<PortType, KPort>() {
          public KPort apply(final PortType p) {
            return Ptolemy2KGraphTransformation.this.transform(p);
          }
        };
        _xifexpression_1 = ListExtensions.<PortType, KPort>map(_port_1, _function_1);
      }
      _xifexpression = _xifexpression_1;
    }
    final List<KPort> modelPorts = _xifexpression;
    int index = ports.size();
    for (final KPort modelPort : modelPorts) {
      {
        KLayoutData _layout = this._annotationExtensions.getLayout(modelPort);
        _layout.<Integer>setProperty(LayoutOptions.PORT_INDEX, Integer.valueOf(index));
        final Function1<KPort, Boolean> _function_2 = new Function1<KPort, Boolean>() {
          public Boolean apply(final KPort p) {
            String _name = Ptolemy2KGraphTransformation.this._labelExtensions.getName(p);
            String _name_1 = Ptolemy2KGraphTransformation.this._labelExtensions.getName(modelPort);
            return Boolean.valueOf(_name.equals(_name_1));
          }
        };
        final KPort existingPort = IterableExtensions.<KPort>findFirst(ports, _function_2);
        boolean _notEquals = (!Objects.equal(existingPort, null));
        if (_notEquals) {
          List<PropertyType> _annotations = this._annotationExtensions.getAnnotations(existingPort);
          List<PropertyType> _annotations_1 = this._annotationExtensions.getAnnotations(modelPort);
          _annotations.addAll(_annotations_1);
        } else {
          ports.add(modelPort);
        }
        index = (index + 1);
      }
    }
    for (final KPort port : ports) {
      {
        port.setNode(knode);
        boolean _or = false;
        boolean _isMarkedAsInputPort = this._markerExtensions.isMarkedAsInputPort(port);
        if (_isMarkedAsInputPort) {
          _or = true;
        } else {
          boolean _isMarkedAsOutputPort = this._markerExtensions.isMarkedAsOutputPort(port);
          _or = _isMarkedAsOutputPort;
        }
        boolean _not = (!_or);
        if (_not) {
          boolean _or_1 = false;
          boolean _hasAnnotation = this._annotationExtensions.hasAnnotation(port, "input");
          if (_hasAnnotation) {
            _or_1 = true;
          } else {
            boolean _hasAnnotation_1 = this._annotationExtensions.hasAnnotation(port, "inputoutput");
            _or_1 = _hasAnnotation_1;
          }
          if (_or_1) {
            this._markerExtensions.markAsInputPort(port, true);
          }
          boolean _or_2 = false;
          boolean _hasAnnotation_2 = this._annotationExtensions.hasAnnotation(port, "output");
          if (_hasAnnotation_2) {
            _or_2 = true;
          } else {
            boolean _hasAnnotation_3 = this._annotationExtensions.hasAnnotation(port, "inputoutput");
            _or_2 = _hasAnnotation_3;
          }
          if (_or_2) {
            this._markerExtensions.markAsOutputPort(port, true);
          }
        }
      }
    }
  }
  
  /**
   * Returns the list of warnings produced by the transformation.
   */
  public List<IStatus> getWarnings() {
    return this.warnings;
  }
}
