/**
 * KIELER - Kiel Integrated Environment for Layout Eclipse RichClient
 * 
 * http://www.informatik.uni-kiel.de/rtsys/kieler/
 * 
 * Copyright 2013 by
 * + Christian-Albrechts-University of Kiel
 *   + Department of Computer Science
 *     + Real-Time and Embedded Systems Group
 * 
 * This code is provided under the terms of the Eclipse Public License (EPL).
 * See the file epl-v10.html for the license text.
 */
package de.cau.cs.kieler.ptolemy.klighd.transformation;

import com.google.common.base.Objects;
import com.google.common.base.Strings;
import com.google.inject.Inject;
import de.cau.cs.kieler.core.kgraph.KEdge;
import de.cau.cs.kieler.core.kgraph.KGraphData;
import de.cau.cs.kieler.core.kgraph.KGraphElement;
import de.cau.cs.kieler.core.kgraph.KLabel;
import de.cau.cs.kieler.core.kgraph.KLabeledGraphElement;
import de.cau.cs.kieler.core.kgraph.KNode;
import de.cau.cs.kieler.core.kgraph.KPort;
import de.cau.cs.kieler.core.krendering.KAreaPlacementData;
import de.cau.cs.kieler.core.krendering.KContainerRendering;
import de.cau.cs.kieler.core.krendering.KPlacementData;
import de.cau.cs.kieler.core.krendering.KPosition;
import de.cau.cs.kieler.core.krendering.KRendering;
import de.cau.cs.kieler.core.krendering.KRenderingRef;
import de.cau.cs.kieler.core.krendering.KText;
import de.cau.cs.kieler.core.krendering.KXPosition;
import de.cau.cs.kieler.core.krendering.KYPosition;
import de.cau.cs.kieler.core.krendering.Trigger;
import de.cau.cs.kieler.core.krendering.extensions.KContainerRenderingExtensions;
import de.cau.cs.kieler.core.krendering.extensions.KRenderingExtensions;
import de.cau.cs.kieler.core.math.KVector;
import de.cau.cs.kieler.kiml.graphviz.layouter.GraphvizTool;
import de.cau.cs.kieler.kiml.klayoutdata.KLayoutData;
import de.cau.cs.kieler.kiml.klayoutdata.KShapeLayout;
import de.cau.cs.kieler.kiml.options.Direction;
import de.cau.cs.kieler.kiml.options.EdgeLabelPlacement;
import de.cau.cs.kieler.kiml.options.EdgeRouting;
import de.cau.cs.kieler.kiml.options.LayoutOptions;
import de.cau.cs.kieler.kiml.options.NodeLabelPlacement;
import de.cau.cs.kieler.kiml.options.PortConstraints;
import de.cau.cs.kieler.kiml.options.PortLabelPlacement;
import de.cau.cs.kieler.kiml.options.PortSide;
import de.cau.cs.kieler.kiml.options.SizeConstraint;
import de.cau.cs.kieler.kiml.util.nodespacing.Spacing;
import de.cau.cs.kieler.klighd.KlighdConstants;
import de.cau.cs.kieler.klighd.microlayout.Bounds;
import de.cau.cs.kieler.klighd.microlayout.PlacementUtil;
import de.cau.cs.kieler.klighd.syntheses.DiagramSyntheses;
import de.cau.cs.kieler.klighd.util.ExpansionAwareLayoutOption;
import de.cau.cs.kieler.klighd.util.KlighdProperties;
import de.cau.cs.kieler.ptolemy.klighd.transformation.KRenderingFigureProvider;
import de.cau.cs.kieler.ptolemy.klighd.transformation.extensions.AnnotationExtensions;
import de.cau.cs.kieler.ptolemy.klighd.transformation.extensions.LabelExtensions;
import de.cau.cs.kieler.ptolemy.klighd.transformation.extensions.MarkerExtensions;
import de.cau.cs.kieler.ptolemy.klighd.transformation.extensions.MiscellaneousExtensions;
import de.cau.cs.kieler.ptolemy.klighd.transformation.util.TransformationConstants;
import java.util.EnumSet;
import java.util.List;
import org.eclipse.emf.common.util.EList;
import org.eclipse.xtext.xbase.lib.Exceptions;
import org.eclipse.xtext.xbase.lib.Extension;
import org.eclipse.xtext.xbase.lib.ObjectExtensions;
import org.eclipse.xtext.xbase.lib.Procedures.Procedure1;
import org.ptolemy.moml.PropertyType;

/**
 * Enriches a KGraph model freshly transformed from a Ptolemy2 model with the KRendering information
 * necessary to properly display the model. This is the final step of the Ptolemy model import process.
 * 
 * @author cds
 */
@SuppressWarnings("all")
public class Ptolemy2KGraphVisualization {
  /**
   * Access to annotations.
   */
  @Inject
  @Extension
  private AnnotationExtensions _annotationExtensions;
  
  /**
   * Access to labels.
   */
  @Inject
  @Extension
  private LabelExtensions _labelExtensions;
  
  /**
   * Access to marked nodes.
   */
  @Inject
  @Extension
  private MarkerExtensions _markerExtensions;
  
  /**
   * Extensions used during the transformation. To make things easier. And stuff.
   */
  @Inject
  @Extension
  private MiscellaneousExtensions _miscellaneousExtensions;
  
  /**
   * Utility class that provides renderings.
   */
  @Inject
  @Extension
  private KRenderingExtensions _kRenderingExtensions;
  
  /**
   * Utility class that provides renderings.
   */
  @Inject
  @Extension
  private KContainerRenderingExtensions _kContainerRenderingExtensions;
  
  /**
   * Utility class that provides renderings.
   */
  @Inject
  @Extension
  private KRenderingFigureProvider _kRenderingFigureProvider;
  
  /**
   * Whether Graphviz is available to be used or not.
   */
  private final boolean isGraphvizAvailable = (!Objects.equal(GraphvizTool.getDotExecutable(false), null));
  
  /**
   * alpha value of the background of expanded compound nodes.
   */
  private int compoundNodeAlpha = 10;
  
  /**
   * Annotates the given KGraph with the information necessary to render it as a Ptolemy model.
   * 
   * @param kGraph the KGraph created from a Ptolemy model.
   * @param compoundNodeAlpha alpha value of the background of expanded compound nodes. A higher value
   *                          translates to more intense backgrounds. INTENSE!
   */
  public void visualize(final KNode kGraph, final int compoundNodeAlpha) {
    this.compoundNodeAlpha = compoundNodeAlpha;
    this.setLayoutAlgorithm(kGraph);
    this.visualizeRecursively(kGraph);
  }
  
  /**
   * Annotates the children of the given node with the information necessary to render them and
   * calls itself recursively with each of the nodes.
   * 
   * @param node the node to visualize.
   * @param firstLevel {@code true} if the given node is the root of the graph. Used to auto-expand
   *                   compound nodes on the first level.
   */
  private void visualizeRecursively(final KNode node) {
    EList<KNode> _children = node.getChildren();
    for (final KNode child : _children) {
      {
        boolean _isMarkedAsState = this._markerExtensions.isMarkedAsState(child);
        if (_isMarkedAsState) {
          this.addStateNodeRendering(child);
          this.visualizeRecursively(child);
        } else {
          EList<KNode> _children_1 = child.getChildren();
          boolean _isEmpty = _children_1.isEmpty();
          boolean _not = (!_isEmpty);
          if (_not) {
            this.addCompoundNodeRendering(child);
            this.visualizeRecursively(child);
          } else {
            boolean _isMarkedAsHypernode = this._markerExtensions.isMarkedAsHypernode(child);
            if (_isMarkedAsHypernode) {
              this.addRelationNodeRendering(child);
            } else {
              boolean _isMarkedAsDirector = this._markerExtensions.isMarkedAsDirector(child);
              if (_isMarkedAsDirector) {
                this.addDirectorNodeRendering(child);
              } else {
                boolean _isMarkedAsComment = this._markerExtensions.isMarkedAsComment(child);
                if (_isMarkedAsComment) {
                  this.addCommentNodeRendering(child);
                } else {
                  boolean _isMarkedAsParameterNode = this._markerExtensions.isMarkedAsParameterNode(child);
                  if (_isMarkedAsParameterNode) {
                    this.addParameterNodeRendering(child);
                  } else {
                    boolean _isMarkedAsValueDisplayingActor = this._markerExtensions.isMarkedAsValueDisplayingActor(child);
                    if (_isMarkedAsValueDisplayingActor) {
                      this.addValueDisplayingNodeRendering(child);
                    } else {
                      boolean _isMarkedAsModalModelPort = this._markerExtensions.isMarkedAsModalModelPort(child);
                      if (_isMarkedAsModalModelPort) {
                        this.addModalModelPortRendering(child);
                      } else {
                        this.addRegularNodeRendering(child);
                      }
                    }
                  }
                }
              }
            }
          }
        }
        EList<KPort> _ports = child.getPorts();
        for (final KPort port : _ports) {
          {
            this.addPortRendering(port);
            this.addLabelRendering(port);
            this.addToolTip(port);
          }
        }
        EList<KEdge> _outgoingEdges = child.getOutgoingEdges();
        for (final KEdge edge : _outgoingEdges) {
          {
            this.addEdgeRendering(edge);
            this.addLabelRendering(edge);
          }
        }
        this.addLabelRendering(child);
        this.addToolTip(child);
      }
    }
  }
  
  /**
   * Renders the given node as a compound node.
   * 
   * @param node the node to attach the rendering information to.
   */
  private void addCompoundNodeRendering(final KNode node) {
    KLayoutData _layout = this._annotationExtensions.getLayout(node);
    final KShapeLayout layout = ((KShapeLayout) _layout);
    layout.<Boolean>setProperty(KlighdProperties.EXPAND, Boolean.valueOf(false));
    EnumSet<NodeLabelPlacement> _of = EnumSet.<NodeLabelPlacement>of(
      NodeLabelPlacement.OUTSIDE, NodeLabelPlacement.H_LEFT, NodeLabelPlacement.V_TOP);
    layout.<EnumSet<NodeLabelPlacement>>setProperty(LayoutOptions.NODE_LABEL_PLACEMENT, _of);
    ExpansionAwareLayoutOption.<PortConstraints>setProperty(layout, LayoutOptions.PORT_CONSTRAINTS, 
      PortConstraints.FIXED_ORDER, PortConstraints.FREE);
    EnumSet<SizeConstraint> _fixed = SizeConstraint.fixed();
    layout.<EnumSet<SizeConstraint>>setProperty(LayoutOptions.SIZE_CONSTRAINT, _fixed);
    Spacing.Margins _margins = new Spacing.Margins(5, 5, 5, 5);
    layout.<Spacing.Margins>setProperty(LayoutOptions.ADDITIONAL_PORT_SPACE, _margins);
    this.setLayoutAlgorithm(node);
    final KRendering collapsedRendering = this._kRenderingFigureProvider.createRegularNodeRendering(node);
    KContainerRendering _addRenderingWithStandardSelectionWrapper = DiagramSyntheses.addRenderingWithStandardSelectionWrapper(node, collapsedRendering);
    final Procedure1<KContainerRendering> _function = new Procedure1<KContainerRendering>() {
      public void apply(final KContainerRendering it) {
        it.<Boolean>setProperty(KlighdProperties.COLLAPSED_RENDERING, Boolean.valueOf(true));
        Ptolemy2KGraphVisualization.this._kRenderingExtensions.<KContainerRendering>addDoubleClickAction(it, KlighdConstants.ACTION_COLLAPSE_EXPAND);
      }
    };
    ObjectExtensions.<KContainerRendering>operator_doubleArrow(_addRenderingWithStandardSelectionWrapper, _function);
    final KRendering expandedRendering = this._kRenderingFigureProvider.createExpandedCompoundNodeRendering(node, this.compoundNodeAlpha);
    KContainerRendering _addRenderingWithStandardSelectionWrapper_1 = DiagramSyntheses.addRenderingWithStandardSelectionWrapper(node, expandedRendering);
    final Procedure1<KContainerRendering> _function_1 = new Procedure1<KContainerRendering>() {
      public void apply(final KContainerRendering it) {
        it.<Boolean>setProperty(KlighdProperties.EXPANDED_RENDERING, Boolean.valueOf(true));
        Ptolemy2KGraphVisualization.this._kRenderingExtensions.<KContainerRendering>addDoubleClickAction(it, KlighdConstants.ACTION_COLLAPSE_EXPAND);
      }
    };
    ObjectExtensions.<KContainerRendering>operator_doubleArrow(_addRenderingWithStandardSelectionWrapper_1, _function_1);
  }
  
  /**
   * Renders the given node as a state machine state.
   * 
   * @param node the node to attach the rendering information to.
   */
  private void addStateNodeRendering(final KNode node) {
    EList<KNode> _children = node.getChildren();
    boolean _isEmpty = _children.isEmpty();
    if (_isEmpty) {
      final KRendering rendering = this._kRenderingFigureProvider.createStateNodeRendering(node);
      EList<KGraphData> _data = node.getData();
      _data.add(rendering);
    } else {
      KLayoutData _layout = this._annotationExtensions.getLayout(node);
      final KShapeLayout layout = ((KShapeLayout) _layout);
      layout.<Boolean>setProperty(KlighdProperties.EXPAND, Boolean.valueOf(false));
      EnumSet<SizeConstraint> _fixed = SizeConstraint.fixed();
      layout.<EnumSet<SizeConstraint>>setProperty(LayoutOptions.SIZE_CONSTRAINT, _fixed);
      this.setLayoutAlgorithm(node);
      final KRendering collapsedRendering = this._kRenderingFigureProvider.createStateNodeRendering(node);
      collapsedRendering.<Boolean>setProperty(KlighdProperties.COLLAPSED_RENDERING, Boolean.valueOf(true));
      this._kRenderingExtensions.<KRendering>addAction(collapsedRendering, Trigger.DOUBLECLICK, KlighdConstants.ACTION_COLLAPSE_EXPAND);
      EList<KGraphData> _data_1 = node.getData();
      _data_1.add(collapsedRendering);
      final KRendering expandedRendering = this._kRenderingFigureProvider.createExpandedCompoundNodeRendering(node, this.compoundNodeAlpha);
      expandedRendering.<Boolean>setProperty(KlighdProperties.EXPANDED_RENDERING, Boolean.valueOf(true));
      this._kRenderingExtensions.<KRendering>addAction(expandedRendering, Trigger.DOUBLECLICK, KlighdConstants.ACTION_COLLAPSE_EXPAND);
      EList<KGraphData> _data_2 = node.getData();
      _data_2.add(expandedRendering);
    }
  }
  
  /**
   * Renders the given node as a relation node. Also removes the relation's labels.
   * 
   * @param node the node to attach the rendering information to.
   */
  private void addRelationNodeRendering(final KNode node) {
    KLayoutData _layout = this._annotationExtensions.getLayout(node);
    final KShapeLayout layout = ((KShapeLayout) _layout);
    EList<KLabel> _labels = node.getLabels();
    _labels.clear();
    EList<KGraphData> _data = node.getData();
    KRendering _createRelationNodeRendering = this._kRenderingFigureProvider.createRelationNodeRendering(node);
    _data.add(_createRelationNodeRendering);
    layout.setHeight(10);
    layout.setWidth(10);
  }
  
  /**
   * Renders the given node as a director.
   * 
   * @param node the node to attach the rendering information to.
   */
  private void addDirectorNodeRendering(final KNode node) {
    KLayoutData _layout = this._annotationExtensions.getLayout(node);
    final KShapeLayout layout = ((KShapeLayout) _layout);
    EnumSet<NodeLabelPlacement> _of = EnumSet.<NodeLabelPlacement>of(
      NodeLabelPlacement.OUTSIDE, NodeLabelPlacement.H_LEFT, NodeLabelPlacement.V_TOP);
    layout.<EnumSet<NodeLabelPlacement>>setProperty(LayoutOptions.NODE_LABEL_PLACEMENT, _of);
    layout.<Integer>setProperty(LayoutOptions.PRIORITY, Integer.valueOf(1000));
    final KRendering rendering = this._kRenderingFigureProvider.createDirectorNodeRendering(node);
    EList<KGraphData> _data = node.getData();
    _data.add(rendering);
    this.setLayoutSize(layout, rendering);
  }
  
  /**
   * Renders the given node as a comment node.
   * 
   * @param node the node to attach the rendering information to.
   */
  private void addCommentNodeRendering(final KNode node) {
    final KRendering rendering = this._kRenderingFigureProvider.createCommentNodeRendering(node);
    EList<KGraphData> _data = node.getData();
    _data.add(rendering);
    List<KEdge> _incidentEdges = this._miscellaneousExtensions.getIncidentEdges(node);
    for (final KEdge edge : _incidentEdges) {
      {
        final KRendering edgeRendering = this._kRenderingFigureProvider.createCommentEdgeRendering(edge);
        EList<KGraphData> _data_1 = edge.getData();
        _data_1.add(edgeRendering);
      }
    }
  }
  
  /**
   * Renders the given node as a parameter node.
   * 
   * @param node the node to attach the rendering information to.
   */
  private void addParameterNodeRendering(final KNode node) {
    KLayoutData _layout = this._annotationExtensions.getLayout(node);
    final KShapeLayout layout = ((KShapeLayout) _layout);
    layout.<Integer>setProperty(LayoutOptions.PRIORITY, Integer.valueOf(800));
    final KRendering rendering = this._kRenderingFigureProvider.createParameterNodeRendering(node);
    EList<KGraphData> _data = node.getData();
    _data.add(rendering);
  }
  
  /**
   * Renders the given node displaying a specific value, for instance a Const actor.
   * 
   * @param node the node to attach the rendering information to.
   */
  private void addValueDisplayingNodeRendering(final KNode node) {
    KLayoutData _layout = this._annotationExtensions.getLayout(node);
    final KShapeLayout layout = ((KShapeLayout) _layout);
    EnumSet<NodeLabelPlacement> _of = EnumSet.<NodeLabelPlacement>of(
      NodeLabelPlacement.OUTSIDE, NodeLabelPlacement.H_LEFT, NodeLabelPlacement.V_TOP);
    layout.<EnumSet<NodeLabelPlacement>>setProperty(LayoutOptions.NODE_LABEL_PLACEMENT, _of);
    layout.<PortLabelPlacement>setProperty(LayoutOptions.PORT_LABEL_PLACEMENT, PortLabelPlacement.OUTSIDE);
    layout.<PortConstraints>setProperty(LayoutOptions.PORT_CONSTRAINTS, PortConstraints.FIXED_ORDER);
    String _annotationValue = this._annotationExtensions.getAnnotationValue(node, TransformationConstants.ANNOTATION_PTOLEMY_CLASS);
    final String className = Strings.nullToEmpty(_annotationValue);
    String _get = TransformationConstants.VALUE_DISPLAY_MAP.get(className);
    final String value = this._annotationExtensions.getAnnotationValue(node, _get);
    String _elvis = null;
    if (value != null) {
      _elvis = value;
    } else {
      _elvis = "";
    }
    final KRendering rendering = this._kRenderingFigureProvider.createValueDisplayingNodeRendering(node, _elvis);
    EList<KGraphData> _data = node.getData();
    _data.add(rendering);
  }
  
  /**
   * Renders the given node as a modal model port.
   * 
   * @param node the node to attach the rendering information to.
   */
  private void addModalModelPortRendering(final KNode node) {
    KLayoutData _layout = this._annotationExtensions.getLayout(node);
    final KShapeLayout layout = ((KShapeLayout) _layout);
    layout.<PortConstraints>setProperty(LayoutOptions.PORT_CONSTRAINTS, PortConstraints.FIXED_SIDE);
    final KRendering rendering = this._kRenderingFigureProvider.createModalModelPortRendering(node);
    EList<KGraphData> _data = node.getData();
    _data.add(rendering);
    layout.setHeight(20);
    layout.setWidth(20);
  }
  
  /**
   * Renders the given node just like it would be rendered in Ptolemy, if possible.
   * 
   * @param node the node to attach the rendering information to.
   */
  private void addRegularNodeRendering(final KNode node) {
    KLayoutData _layout = this._annotationExtensions.getLayout(node);
    final KShapeLayout layout = ((KShapeLayout) _layout);
    EnumSet<NodeLabelPlacement> _of = EnumSet.<NodeLabelPlacement>of(
      NodeLabelPlacement.OUTSIDE, NodeLabelPlacement.H_LEFT, NodeLabelPlacement.V_TOP);
    layout.<EnumSet<NodeLabelPlacement>>setProperty(LayoutOptions.NODE_LABEL_PLACEMENT, _of);
    layout.<PortLabelPlacement>setProperty(LayoutOptions.PORT_LABEL_PLACEMENT, PortLabelPlacement.OUTSIDE);
    layout.<PortConstraints>setProperty(LayoutOptions.PORT_CONSTRAINTS, PortConstraints.FIXED_ORDER);
    Spacing.Margins _margins = new Spacing.Margins(5, 5, 5, 5);
    layout.<Spacing.Margins>setProperty(LayoutOptions.ADDITIONAL_PORT_SPACE, _margins);
    KRendering _switchResult = null;
    String _annotationValue = this._annotationExtensions.getAnnotationValue(node, TransformationConstants.ANNOTATION_PTOLEMY_CLASS);
    boolean _matched = false;
    if (!_matched) {
      if (Objects.equal(_annotationValue, "ptolemy.actor.lib.Accumulator")) {
        _matched=true;
        _switchResult = this._kRenderingFigureProvider.createAccumulatorNodeRendering(node);
      }
    }
    if (!_matched) {
      _switchResult = this._kRenderingFigureProvider.createRegularNodeRendering(node);
    }
    final KRendering rendering = _switchResult;
    DiagramSyntheses.addRenderingWithStandardSelectionWrapper(node, rendering);
    this.setLayoutSize(layout, rendering);
  }
  
  /**
   * Adds KRendering information to the given port.
   * 
   * @param port the port to add rendering information to.
   */
  private void addPortRendering(final KPort port) {
    KLayoutData _layout = this._annotationExtensions.getLayout(port);
    final KShapeLayout layout = ((KShapeLayout) _layout);
    final boolean inputPort = this._markerExtensions.isMarkedAsInputPort(port);
    final boolean outputPort = this._markerExtensions.isMarkedAsOutputPort(port);
    final String ptolemySpecifiedPortSide = this._annotationExtensions.getAnnotationValue(port, "_cardinal");
    PortSide portSide = PortSide.UNDEFINED;
    try {
      PortSide _valueOf = PortSide.valueOf(ptolemySpecifiedPortSide);
      portSide = _valueOf;
    } catch (final Throwable _t) {
      if (_t instanceof Exception) {
        final Exception e = (Exception)_t;
      } else {
        throw Exceptions.sneakyThrow(_t);
      }
    }
    boolean _equals = Objects.equal(portSide, PortSide.UNDEFINED);
    if (_equals) {
      boolean _and = false;
      if (!inputPort) {
        _and = false;
      } else {
        _and = outputPort;
      }
      if (_and) {
        portSide = PortSide.SOUTH;
      } else {
        if (inputPort) {
          portSide = PortSide.WEST;
        } else {
          portSide = PortSide.EAST;
        }
      }
    }
    layout.<PortSide>setProperty(LayoutOptions.PORT_SIDE, portSide);
    final Integer index = layout.<Integer>getProperty(LayoutOptions.PORT_INDEX);
    if (portSide != null) {
      switch (portSide) {
        case NORTH:
          layout.<Float>setProperty(LayoutOptions.OFFSET, Float.valueOf(0f));
          break;
        case SOUTH:
          layout.<Float>setProperty(LayoutOptions.OFFSET, Float.valueOf(0f));
          layout.<Integer>setProperty(LayoutOptions.PORT_INDEX, Integer.valueOf((-(index).intValue())));
          break;
        case EAST:
          layout.<Float>setProperty(LayoutOptions.OFFSET, Float.valueOf(0f));
          boolean _isMarkedAsModalModelPort = this._markerExtensions.isMarkedAsModalModelPort(port);
          boolean _not = (!_isMarkedAsModalModelPort);
          if (_not) {
            KVector _kVector = new KVector(7, 3.5);
            layout.<KVector>setProperty(LayoutOptions.PORT_ANCHOR, _kVector);
          }
          break;
        case WEST:
          layout.<Float>setProperty(LayoutOptions.OFFSET, Float.valueOf(0f));
          boolean _isMarkedAsModalModelPort_1 = this._markerExtensions.isMarkedAsModalModelPort(port);
          boolean _not_1 = (!_isMarkedAsModalModelPort_1);
          if (_not_1) {
            KVector _kVector_1 = new KVector(0, 3.5);
            layout.<KVector>setProperty(LayoutOptions.PORT_ANCHOR, _kVector_1);
          }
          layout.<Integer>setProperty(LayoutOptions.PORT_INDEX, Integer.valueOf((-(index).intValue())));
          break;
        default:
          break;
      }
    }
    KRendering rendering = null;
    boolean _isMarkedAsModalModelPort_2 = this._markerExtensions.isMarkedAsModalModelPort(port);
    boolean _not_2 = (!_isMarkedAsModalModelPort_2);
    if (_not_2) {
      KRendering _createPortRendering = this._kRenderingFigureProvider.createPortRendering(port);
      rendering = _createPortRendering;
      EList<KGraphData> _data = port.getData();
      _data.add(rendering);
      layout.setWidth(8);
      layout.setHeight(8);
    }
    String _name = this._labelExtensions.getName(port);
    int _length = _name.length();
    boolean _greaterThan = (_length > 0);
    if (_greaterThan) {
      boolean _isMarkedAsModalModelPort_3 = this._markerExtensions.isMarkedAsModalModelPort(port);
      if (_isMarkedAsModalModelPort_3) {
        KNode _node = port.getNode();
        String _name_1 = this._labelExtensions.getName(port);
        this._labelExtensions.setName(_node, _name_1);
      } else {
        String _name_2 = this._labelExtensions.getName(port);
        String _plus = ("Port: " + _name_2);
        rendering.<String>setProperty(KlighdProperties.TOOLTIP, _plus);
      }
    }
    EList<KLabel> _labels = port.getLabels();
    _labels.clear();
  }
  
  /**
   * Adds KRendering information to the given edge.
   * 
   * @param edge the edge to add rendering information to.
   */
  private void addEdgeRendering(final KEdge edge) {
    boolean _or = false;
    KNode _source = edge.getSource();
    boolean _isMarkedAsState = this._markerExtensions.isMarkedAsState(_source);
    if (_isMarkedAsState) {
      _or = true;
    } else {
      KNode _target = edge.getTarget();
      boolean _isMarkedAsState_1 = this._markerExtensions.isMarkedAsState(_target);
      _or = _isMarkedAsState_1;
    }
    if (_or) {
      final StringBuffer labelText = new StringBuffer();
      final String annotation = this._annotationExtensions.getAnnotationValue(edge, TransformationConstants.ANNOTATION_ANNOTATION);
      boolean _isNullOrEmpty = Strings.isNullOrEmpty(annotation);
      boolean _not = (!_isNullOrEmpty);
      if (_not) {
        labelText.append(("\n" + annotation));
      }
      final String guardExpression = this._annotationExtensions.getAnnotationValue(edge, TransformationConstants.ANNOTATION_GUARD_EXPRESSION);
      boolean _isNullOrEmpty_1 = Strings.isNullOrEmpty(guardExpression);
      boolean _not_1 = (!_isNullOrEmpty_1);
      if (_not_1) {
        labelText.append(("\nGuard: " + guardExpression));
      }
      final String outputActions = this._annotationExtensions.getAnnotationValue(edge, TransformationConstants.ANNOTATION_OUTPUT_ACTIONS);
      boolean _isNullOrEmpty_2 = Strings.isNullOrEmpty(outputActions);
      boolean _not_2 = (!_isNullOrEmpty_2);
      if (_not_2) {
        labelText.append(("\nOutput: " + outputActions));
      }
      final String setActions = this._annotationExtensions.getAnnotationValue(edge, TransformationConstants.ANNOTATION_SET_ACTIONS);
      boolean _isNullOrEmpty_3 = Strings.isNullOrEmpty(setActions);
      boolean _not_3 = (!_isNullOrEmpty_3);
      if (_not_3) {
        labelText.append(("\nSet: " + setActions));
      }
      int _length = labelText.length();
      boolean _greaterThan = (_length > 0);
      if (_greaterThan) {
        String _substring = labelText.substring(1);
        this._labelExtensions.setName(edge, _substring);
        EList<KLabel> _labels = edge.getLabels();
        KLabel _get = _labels.get(0);
        final KLayoutData layout = this._annotationExtensions.getLayout(_get);
        layout.<EdgeLabelPlacement>setProperty(LayoutOptions.EDGE_LABEL_PLACEMENT, EdgeLabelPlacement.CENTER);
      }
      final KRendering rendering = this._kRenderingFigureProvider.createTransitionRendering(edge);
      EList<KGraphData> _data = edge.getData();
      _data.add(rendering);
    } else {
      EList<KGraphData> _data_1 = edge.getData();
      KRendering _createDataFlowRendering = this._kRenderingFigureProvider.createDataFlowRendering(edge);
      _data_1.add(_createDataFlowRendering);
    }
  }
  
  /**
   * Attaches rendering information to all labels of the given element.
   * 
   * @param elemtent the element.
   */
  private void addLabelRendering(final KLabeledGraphElement element) {
    EList<KLabel> _labels = element.getLabels();
    for (final KLabel label : _labels) {
      {
        KContainerRendering _addRenderingWithStandardSelectionWrapper = DiagramSyntheses.addRenderingWithStandardSelectionWrapper(label, null);
        final KText ktext = this._kContainerRenderingExtensions.addText(_addRenderingWithStandardSelectionWrapper, "");
        boolean _isMarkedAsModalModelPort = this._markerExtensions.isMarkedAsModalModelPort(element);
        if (_isMarkedAsModalModelPort) {
          KLayoutData _layout = this._annotationExtensions.getLayout(label);
          final KShapeLayout layout = ((KShapeLayout) _layout);
          final Bounds bounds = PlacementUtil.estimateTextSize(label);
          layout.setXpos(0);
          float _height = bounds.getHeight();
          float _plus = (_height + 3.0f);
          float _minus = (-_plus);
          layout.setYpos(_minus);
        }
        if ((element instanceof KEdge)) {
          this._kRenderingExtensions.setFontSize(ktext, (KlighdConstants.DEFAULT_FONT_SIZE - 2));
        }
      }
    }
  }
  
  /**
   * Generates a tool tip for the given element based on its properties if it has rendering
   * information. (modal model ports don't have any)
   * 
   * @param element the element to generate the tooltip for.
   */
  private void addToolTip(final KGraphElement element) {
    final KRendering krendering = this._kRenderingExtensions.getKRendering(element);
    boolean _equals = Objects.equal(krendering, null);
    if (_equals) {
      return;
    }
    final String toolTip = krendering.<String>getProperty(KlighdProperties.TOOLTIP);
    final StringBuffer toolTipText = new StringBuffer();
    boolean _isNullOrEmpty = Strings.isNullOrEmpty(toolTip);
    boolean _not = (!_isNullOrEmpty);
    if (_not) {
      toolTipText.append(("\n" + toolTip));
    }
    List<PropertyType> _annotations = this._annotationExtensions.getAnnotations(element);
    for (final PropertyType property : _annotations) {
      {
        boolean _and = false;
        String _name = property.getName();
        boolean _startsWith = _name.startsWith("_");
        boolean _not_1 = (!_startsWith);
        if (!_not_1) {
          _and = false;
        } else {
          boolean _and_1 = false;
          boolean _and_2 = false;
          if (!(element instanceof KNode)) {
            _and_2 = false;
          } else {
            boolean _isMarkedAsComment = this._markerExtensions.isMarkedAsComment(((KNode) element));
            _and_2 = _isMarkedAsComment;
          }
          if (!_and_2) {
            _and_1 = false;
          } else {
            String _name_1 = property.getName();
            boolean _equals_1 = _name_1.equals(TransformationConstants.ANNOTATION_COMMENT_TEXT);
            _and_1 = _equals_1;
          }
          boolean _not_2 = (!_and_1);
          _and = _not_2;
        }
        final boolean includeAnnotation = _and;
        if (includeAnnotation) {
          StringBuffer _append = toolTipText.append("\n");
          String _name_2 = property.getName();
          _append.append(_name_2);
          String _value = property.getValue();
          boolean _isNullOrEmpty_1 = Strings.isNullOrEmpty(_value);
          boolean _not_3 = (!_isNullOrEmpty_1);
          if (_not_3) {
            StringBuffer _append_1 = toolTipText.append(": ");
            String _value_1 = property.getValue();
            _append_1.append(_value_1);
          }
        }
      }
    }
    int _length = toolTipText.length();
    boolean _greaterThan = (_length > 0);
    if (_greaterThan) {
      String _substring = toolTipText.substring(1);
      krendering.<String>setProperty(KlighdProperties.TOOLTIP, _substring);
    }
  }
  
  /**
   * Sets the layout size depending on the information in the given rendering information.
   * 
   * @param layout the layout object to set size information on.
   * @param rendering the rendering to infer the size information from. May be {@code null}, in which
   *                  case a default size is assumed.
   */
  private void setLayoutSize(final KShapeLayout layout, final KRendering rendering) {
    KRendering actualRendering = rendering;
    while ((actualRendering instanceof KRenderingRef)) {
      KRendering _rendering = ((KRenderingRef) actualRendering).getRendering();
      actualRendering = _rendering;
    }
    boolean _equals = Objects.equal(actualRendering, null);
    if (_equals) {
      layout.setHeight(50);
      layout.setWidth(50);
    } else {
      boolean _and = false;
      KPlacementData _placementData = actualRendering.getPlacementData();
      boolean _notEquals = (!Objects.equal(_placementData, null));
      if (!_notEquals) {
        _and = false;
      } else {
        KPlacementData _placementData_1 = actualRendering.getPlacementData();
        _and = (_placementData_1 instanceof KAreaPlacementData);
      }
      if (_and) {
        KPlacementData _placementData_2 = actualRendering.getPlacementData();
        final KAreaPlacementData placementData = ((KAreaPlacementData) _placementData_2);
        KPosition _bottomRight = placementData.getBottomRight();
        KYPosition<?> _y = _bottomRight.getY();
        float _absolute = _y.getAbsolute();
        layout.setHeight(_absolute);
        KPosition _bottomRight_1 = placementData.getBottomRight();
        KXPosition<?> _x = _bottomRight_1.getX();
        float _absolute_1 = _x.getAbsolute();
        layout.setWidth(_absolute_1);
      } else {
        KVector _kVector = new KVector(60, 40);
        layout.<KVector>setProperty(KlighdProperties.MINIMAL_NODE_SIZE, _kVector);
        EnumSet<SizeConstraint> _of = EnumSet.<SizeConstraint>of(SizeConstraint.MINIMUM_SIZE);
        layout.<EnumSet<SizeConstraint>>setProperty(LayoutOptions.SIZE_CONSTRAINT, _of);
      }
    }
  }
  
  /**
   * Sets the layout algorithm of the given node depending on which kind of diagram the node hosts.
   * 
   * @param node the node to set the layout algorithm information on.
   */
  private void setLayoutAlgorithm(final KNode node) {
    final KLayoutData layout = this._annotationExtensions.getLayout(node);
    boolean _and = false;
    boolean _isMarkedAsStateMachineContainer = this._markerExtensions.isMarkedAsStateMachineContainer(node);
    if (!_isMarkedAsStateMachineContainer) {
      _and = false;
    } else {
      _and = this.isGraphvizAvailable;
    }
    if (_and) {
      layout.<String>setProperty(LayoutOptions.ALGORITHM, "de.cau.cs.kieler.graphviz.dot");
      layout.<Direction>setProperty(LayoutOptions.DIRECTION, Direction.RIGHT);
    } else {
      layout.<String>setProperty(LayoutOptions.ALGORITHM, "de.cau.cs.kieler.klay.layered");
      layout.<EdgeRouting>setProperty(LayoutOptions.EDGE_ROUTING, EdgeRouting.ORTHOGONAL);
    }
  }
}
